package haspare

import (
	"context"
	"encoding/json"
	"fmt"
	log "github.com/sirupsen/logrus"
	appsv1 "k8s.io/api/apps/v1"
	corev1 "k8s.io/api/core/v1"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/apimachinery/pkg/labels"
	"k8s.io/apimachinery/pkg/types"
	"k8s.io/client-go/kubernetes"
	appslisters "k8s.io/client-go/listers/apps/v1"
	corelisters "k8s.io/client-go/listers/core/v1"
	quarkv1alpha1 "quark.netapp.io/volume-controller/pkg/apis/volumecontroller/v1alpha1"
	quarkclientset "quark.netapp.io/volume-controller/pkg/client/clientset/versioned"
	"quark.netapp.io/volume-controller/pkg/client/listers/volumecontroller/v1alpha1"
	"quark.netapp.io/volume-controller/pkg/quarkconfig"
	"quark.netapp.io/volume-controller/pkg/util"
)

const (
	haSpareSelector        = "app=ha-spare"
	quarkPodSelector       = quarkconfig.AppLabel + "=quark"
	migrationSpareSelector = quarkconfig.AppLabel + "=" +
		quarkconfig.MigrationSpare
)

type haSpareController struct {
	namespace      string
	haSpareImage   string
	pullPolicy     string
	regional       bool
	quarkVariant   string
	kubeClient     kubernetes.Interface
	quarkClient    quarkclientset.Interface
	volumes        v1alpha1.NetAppVolumeLister
	serviceLevels  v1alpha1.NetAppServiceLevelLister
	deployments    appslisters.DeploymentLister
	pods           corelisters.PodLister
	nodes          corelisters.NodeLister
	storagePools   v1alpha1.NetAppStoragePoolLister
	enableGcp1p    bool
	cloud          string
	ownerReference metav1.OwnerReference
}

func NewHASpareController(
	namespace,
	haSpareImage,
	pullPolicy string,
	regional bool,
	quarkVariant string,
	kubeClient kubernetes.Interface,
	quarkClient quarkclientset.Interface,
	volumes v1alpha1.NetAppVolumeLister,
	serviceLevels v1alpha1.NetAppServiceLevelLister,
	deployments appslisters.DeploymentLister,
	pods corelisters.PodLister,
	nodes corelisters.NodeLister,
	storagePools v1alpha1.NetAppStoragePoolLister,
	enableGcp1p bool,
	cloud string,
) (*haSpareController, error) {
	nvc, err := deployments.Deployments(namespace).Get(quarkconfig.NVCDeploymentName)
	if err != nil {
		return nil, err
	}
	return &haSpareController{
		namespace:     namespace,
		haSpareImage:  haSpareImage,
		pullPolicy:    pullPolicy,
		regional:      regional,
		quarkVariant:  quarkVariant,
		kubeClient:    kubeClient,
		quarkClient:   quarkClient,
		volumes:       volumes,
		serviceLevels: serviceLevels,
		deployments:   deployments,
		pods:          pods,
		nodes:         nodes,
		storagePools:  storagePools,
		enableGcp1p:   enableGcp1p,
		cloud:         cloud,
		ownerReference: *metav1.NewControllerRef(
			nvc, appsv1.SchemeGroupVersion.WithKind("Deployment")),
	}, nil
}

// Reconcile queries NetAppServiceLevels in the cluster to ensure one HA spare deployment
// exists for every QuarkPodBucket. HA spare deployments are initially scaled to zero.
//
// Reconcile then lists all volumes in the cluster and maps each volume to a QuarkPodBucket and
// NodeSelector to determine which HA spare deployments to scale up and/or scale down.
func (c *haSpareController) Reconcile() error {
	serviceLevels := make(map[string]*quarkv1alpha1.NetAppServiceLevel)
	slList, err := c.serviceLevels.List(labels.NewSelector())
	if err != nil {
		return fmt.Errorf("failed to list netappservicelevels: %s", err.Error())
	} else {
		for _, sl := range slList {
			// ha-spares need not be deployed when VMRS is disabled or with min configs when NHC isn't present
			if !sl.Spec.RightSize || c.quarkVariant == quarkconfig.MinVariant {
				log.WithFields(log.Fields{
					"serviceLevel": sl.Name,
				}).Info("VM rightsizing disabled. Skipping reconciliation.")
				return nil
			}
			serviceLevels[sl.Name] = sl
		}
	}
	if len(slList) == 0 {
		log.Debug("No netappservicelevels found. Skipping reconciliation.")
		return nil
	}
	if err := c.ensureDeploymentsExist(serviceLevels); err != nil {
		return fmt.Errorf("failed to reconcile HA spares: %s", err.Error())
	}
	if err := c.scaleDeployments(serviceLevels); err != nil {
		return fmt.Errorf("failed to scale HA spares: %s", err.Error())
	}
	return nil
}

// ensureDeploymentsExist ensures that an HA spare deployment
// exists for each QuarkPodBucket in a list of NetAppServiceLevels.
func (c *haSpareController) ensureDeploymentsExist(sl map[string]*quarkv1alpha1.NetAppServiceLevel) error {
	haSpares := make(map[string]*appsv1.Deployment)
	deploymentSelector, _ := labels.Parse(haSpareSelector)
	haSpareList, err := c.deployments.Deployments(c.namespace).List(deploymentSelector)
	if err != nil {
		return fmt.Errorf("failed to list HA spare deployments: %s", err.Error())
	}

	for _, ha := range haSpareList {
		// We do not use pod anti affinity anymore. Delete and recreate the deployment if found
		if ha.Spec.Template.Spec.Affinity != nil {
			if ha.Spec.Template.Spec.Affinity.PodAntiAffinity != nil {
				log.Errorf("antiAffinity block is not nil, delete ha-spare deployment and recreate")
				err = c.kubeClient.AppsV1().Deployments(c.namespace).Delete(context.TODO(), ha.Name, metav1.DeleteOptions{})
				if err != nil {
					log.WithFields(log.Fields{
						"haSpare": ha.Name,
					}).Errorf("Unable to delete HA spare deployment: %s", err.Error())
					return err
				}
				continue
			}
		}
		log.Debugf("adding to haSpares: %v", ha.Name)
		haSpares[ha.Name] = ha
	}

	nodes, err := c.nodes.List(labels.Everything())
	if err != nil {
		return fmt.Errorf("failed to list cluster nodes: %v", err)
	}
	zones := util.GetZonesForNodes(nodes)
	for _, serviceLevel := range sl {
		for _, qpb := range serviceLevel.Spec.QuarkPodBuckets {
			for _, zone := range zones {
				haSpareKey := getHASpareKey(qpb, zone)
				// haSpare does not exist so create it
				if _, ok := haSpares[haSpareKey]; !ok {
					log.WithFields(log.Fields{
						"haSpare":      haSpareKey,
						"nodeSelector": qpb.NodeSelector,
					}).Info("Creating haSpare with nodeSelector.")
					has := c.makeHAS(
						haSpareKey,
						util.GetResources(qpb),
						util.NodeSelectorStringToMap(qpb.NodeSelector),
						zone,
					)
					_, err := c.kubeClient.AppsV1().Deployments(c.namespace).Create(context.TODO(), has, metav1.CreateOptions{})
					if err != nil {
						log.WithFields(log.Fields{
							"haSpare":      haSpareKey,
							"nodeSelector": qpb.NodeSelector,
						}).Errorf("Failed to create HA spare: %s", err.Error())
					}
				}
			}
		}
	}
	return nil
}

// Returns a list of HA spare keys and whether the SL needs to be upgraded
// One of storagePool/volume should be non-nil; the other must be nil
func (c *haSpareController) getHASpareKeysNeedsUpgrade(
	storagePool *quarkv1alpha1.NetAppStoragePool,
	volume *quarkv1alpha1.NetAppVolume,
	zone string,
	sl map[string]*quarkv1alpha1.NetAppServiceLevel,
	failover bool,
) ([]string, bool, error) {
	var (
		serviceLevel *quarkv1alpha1.NetAppServiceLevel
		found        bool
		keys         = make([]string, 0)
	)
	if storagePool != nil {
		// Don't create HA spares for disabled SPs.
		if storagePool.Spec.Disabled {
			return nil, false, nil
		}
		serviceLevel, found = sl[storagePool.Spec.ServiceLevel]
		if !found {
			return nil, false, fmt.Errorf(
				"storagePool has a servicelevel that does not exist: %s",
				storagePool.Spec.ServiceLevel)
		}
	} else {
		serviceLevel, found = sl[volume.Spec.ServiceLevel]
		if !found {
			return nil, false, fmt.Errorf(
				"volume has a servicelevel that does not exist: %s",
				volume.Spec.ServiceLevel)
		}
	}
	specQuarkPodBucket, _, err := c.getQuarkPodBuckets(
		serviceLevel, volume, storagePool)
	if err != nil {
		return nil, false, fmt.Errorf(
			"failed to get quarkpodbuckets for  %s: %s", volume.Name,
			err.Error())
	}
	if zone != "" {
		// If the pod doesn't have a zone, we don't need HA spares for it
		if failover {
			// In the interest of cost savings, the goal is to create
			// ha-spare pods for failover zones only on small (n2-standard-4)
			// nodes.  But until we are prepared for migrating all storagepools
			// to small nodes, failover will be passed in as false
			keys = []string{fmt.Sprintf("ha-spare-small-%s", zone)}
		} else {
			keys = []string{getHASpareKey(*specQuarkPodBucket, zone)}
		}
	}
	needsUpgrade := false
	if (volume != nil &&
		volume.Status.QuarkVersion != serviceLevel.Spec.QuarkVersion) ||
		(storagePool != nil &&
			storagePool.Status.QuarkVersion != serviceLevel.Spec.QuarkVersion) {
		// not inlining this for clarity in the return value
		needsUpgrade = true
	}
	return keys, needsUpgrade, nil
}

// scaleDeployments ensures that there is one HA spare replica for each node selector
// in the cluster with one or more volumes.
func (c *haSpareController) scaleDeployments(sl map[string]*quarkv1alpha1.NetAppServiceLevel) error {
	var (
		needsUpgrade         bool
		haSpareCount         = make(map[string]int32)
		haSparePrimaryCount  = make(map[string]int32)
		haSpareFailoverCount = make(map[string]int32)
		zoneForNode          = make(map[string]string)
		zoneForSP            = make(map[string]string)
		zoneForVol           = make(map[string]string)
	)
	nodes, err := c.nodes.List(labels.NewSelector())
	if err != nil {
		return fmt.Errorf("failed to list nodes: %v", err)
	}
	for _, node := range nodes {
		if zone := node.Labels[quarkconfig.ZoneTopologyKey]; zone != "" {
			zoneForNode[node.Name] = zone
		}
	}

	podSelector, err := labels.Parse(quarkPodSelector)
	if err != nil {
		return fmt.Errorf("unable to parse podSelector %s: %v",
			quarkPodSelector, err)
	}
	pods, err := c.pods.Pods(c.namespace).List(podSelector)
	if err != nil {
		return fmt.Errorf("failed to list Quark pods for selector %s: %v",
			quarkPodSelector, err)
	}
	for _, pod := range pods {
		volumeName := pod.Labels[quarkconfig.VolumeLabel]
		storagePoolName := pod.Labels[quarkconfig.StoragePoolLabel]
		if volumeName == "" && storagePoolName == "" {
			return fmt.Errorf("quark pod %s does not have associated volume or storage pool labels",
				pod.Name)
		}
		if pod.Spec.NodeName == "" {
			continue
		}
		zone := zoneForNode[pod.Spec.NodeName]
		if zone == "" {
			log.WithFields(log.Fields{
				"volume":      volumeName,
				"storagePool": storagePoolName,
				"pod":         pod.Name,
				"node":        pod.Spec.NodeName,
			}).Warn("Found pod without current zone.")
			continue
		}

		if storagePoolName != "" {
			zoneForSP[storagePoolName] = zone
		} else {
			zoneForVol[volumeName] = zone
		}
	}
	// Check whether storage pool spec values indicate that we need an HA
	// spare in a zone due to an impending migration.
	storagePools, err := c.storagePools.NetAppStoragePools(c.namespace).List(
		labels.Everything())
	for _, storagePool := range storagePools {
		haSpareKeys := []string{}
		if storagePool.Spec.Zone != "" {
			// If spec.Zone isn't empty, go ahead and create an HA spare for
			// it, as it'll either be the SP's current zone or the destination
			// zone of a migration
			var newHaSpareKeys []string
			newHaSpareKeys, needsUpgrade, err = c.getHASpareKeysNeedsUpgrade(
				storagePool, nil, storagePool.Spec.Zone, sl, false)
			if err != nil {
				return err // error already formatted
			}
			haSpareKeys = append(haSpareKeys, newHaSpareKeys...)
			for _, k := range newHaSpareKeys {
				haSparePrimaryCount[k] = 1
			}
		}
		if storagePool.Spec.Zone != zoneForSP[storagePool.Name] {
			// Note that storagePool.Spec.Zone can be blank for zonal SPs
			// and for SPs that were created using the .10 UI, so we need to
			// check the current zone of the SP here. In addition, it's
			// probably good to keep the original HA spare around until the
			// pod has migrated, since migration can take up to 3-4 minutes.
			var newHaSpareKeys []string
			// Note that if the SP doesn't have a pod, the zone parameter will
			// be "", which getHASpareKeysNeedsUpgrade handles.
			newHaSpareKeys, needsUpgrade, err = c.getHASpareKeysNeedsUpgrade(
				storagePool, nil, zoneForSP[storagePool.Name], sl, false)
			if err != nil {
				return err // error already formatted
			}
			haSpareKeys = append(haSpareKeys, newHaSpareKeys...)
			for _, k := range newHaSpareKeys {
				haSparePrimaryCount[k] = 1
			}
		}
		// For 1P, also need haspare for the other available zones
		if c.enableGcp1p || c.cloud == quarkconfig.CloudAzure {
			for i, zone := range storagePool.Spec.AvailableZones {
				if i < 2 {
					// An Azure deployment has multiple available zones for each SP. Limit the number of failover
					// spares deployed to 1. This check needs to be removed to support multiple zone failures
					if zone != storagePool.Spec.Zone {
						// The goal is to create these ha-spare pods only on small (n2-standard-4)
						// nodes.  But until we are prepared for migrating all storagepools to small
						// nodes, pass in false for the failover bool
						newHaSpareKeys, _, err := c.getHASpareKeysNeedsUpgrade(
							storagePool, nil, zone, sl, false)
						if err != nil {
							return err // error already formatted
						}
						haSpareKeys = append(haSpareKeys, newHaSpareKeys...)
						for _, k := range newHaSpareKeys {
							haSpareFailoverCount[k]++
						}
					}
				}
			}
		}

		for _, k := range haSpareKeys {
			haSpareCount[k] = haSparePrimaryCount[k] + haSpareFailoverCount[k]
		}
	}

	deploymentSelector, _ := labels.Parse(haSpareSelector)
	haSpareList, err := c.deployments.Deployments(c.namespace).List(deploymentSelector)
	if err != nil {
		return fmt.Errorf("failed to list HA spare deployments: %s", err.Error())
	}

	// update serviceLevel status once all volumes have been upgraded
	for _, serviceLevel := range sl {
		if !needsUpgrade && serviceLevel.Spec.QuarkVersion != serviceLevel.Status.QuarkVersion {
			if err = c.updateServiceLevelStatus(serviceLevel); err != nil {
				return err
			}
		}
	}

	for _, haSpare := range haSpareList {
		// Update the haspare deployments to have the new topology key
		if zone, hasOldTopoKey := haSpare.Spec.Template.Spec.NodeSelector[quarkconfig.OldZoneTopologyKey]; hasOldTopoKey {
			// remove the failure-domain key and add the new key with the value from failure-domain
			haCopy := haSpare.DeepCopy()
			spec := &haCopy.Spec.Template.Spec

			delete(spec.NodeSelector, quarkconfig.OldZoneTopologyKey)
			spec.NodeSelector[quarkconfig.ZoneTopologyKey] = zone

			if spec.Affinity != nil && spec.Affinity.PodAntiAffinity != nil && spec.Affinity.PodAntiAffinity.RequiredDuringSchedulingIgnoredDuringExecution != nil {
				for index, _ := range spec.Affinity.PodAntiAffinity.RequiredDuringSchedulingIgnoredDuringExecution {
					if spec.Affinity.PodAntiAffinity.RequiredDuringSchedulingIgnoredDuringExecution[index].TopologyKey == quarkconfig.OldZoneTopologyKey {
						spec.Affinity.PodAntiAffinity.RequiredDuringSchedulingIgnoredDuringExecution[index].TopologyKey = quarkconfig.ZoneTopologyKey
					}
				}
			}

			log.WithFields(log.Fields{
				"haSpareName": haSpare.Name,
				"fromSpecNS":  haSpare.Spec.Template.Spec.NodeSelector,
				"toSpecNS":    haCopy.Spec.Template.Spec.NodeSelector,
			}).Debug("Updating HA spare deployment's nodeSelector.")

			if _, err = c.kubeClient.AppsV1().Deployments(c.namespace).Update(context.TODO(), haCopy, metav1.UpdateOptions{}); err != nil {
				return fmt.Errorf("failed to update the ha-spare deployment: %s, err: %v", haSpare.Name, err)
			}
		}

		var patchValue *int32
		if *haSpare.Spec.Replicas != haSpareCount[haSpare.Name] {
			patchValue = toIntPtr(haSpareCount[haSpare.Name])
		}
		if patchValue != nil {
			patch, _ := json.Marshal([]interface{}{
				map[string]interface{}{
					"op":    "replace",
					"path":  "/spec/replicas",
					"value": *patchValue,
				},
			})
			log.WithFields(log.Fields{
				"haSpareName":     haSpare.Name,
				"fromNumReplicas": *haSpare.Spec.Replicas,
				"toNumReplicas":   *patchValue,
			}).Info("Scaling HA spare deployment.")
			if _, err = c.kubeClient.AppsV1().Deployments(c.namespace).Patch(context.TODO(), haSpare.Name, types.JSONPatchType, patch, metav1.PatchOptions{}); err != nil {
				return fmt.Errorf("failed to reconcile deployment: %s, err: %v", haSpare.Name, err)
			}
		}
	}
	return nil
}

func (c *haSpareController) updateServiceLevelStatus(sl *quarkv1alpha1.NetAppServiceLevel) error {
	log.WithFields(log.Fields{
		"serviceLevel": sl.Name,
		"newVersion":   sl.Spec.QuarkVersion,
		"oldVersion":   sl.Status.QuarkVersion,
	}).Info("Updating NetAppServiceLevel status.")
	slCopy := sl.DeepCopy()
	slCopy.Status.QuarkPodBuckets = sl.Spec.QuarkPodBuckets
	slCopy.Status.Regional = sl.Spec.Regional
	slCopy.Status.RightSize = sl.Spec.RightSize
	slCopy.Status.QuarkVersion = sl.Spec.QuarkVersion
	_, err := c.quarkClient.QuarkV1alpha1().NetAppServiceLevels(sl.Namespace).UpdateStatus(context.TODO(), slCopy, metav1.UpdateOptions{})
	if err != nil {
		return fmt.Errorf("failed to update NetAppServiceLevel status on the server: %s", err.Error())
	}
	return nil
}

func (c *haSpareController) makeHAS(
	name string,
	resources corev1.ResourceList,
	nodeSelector map[string]string,
	zone string,
) *appsv1.Deployment {
	labels := map[string]string{
		"app":      quarkconfig.HASpareDeploymentName,
		"ha-spare": name,
	}
	nodeSelector[quarkconfig.ZoneTopologyKey] = zone
	return &appsv1.Deployment{
		ObjectMeta: metav1.ObjectMeta{
			Name:      name,
			Namespace: c.namespace,
			Labels:    labels,
			OwnerReferences: []metav1.OwnerReference{
				c.ownerReference,
			},
		},
		Spec: appsv1.DeploymentSpec{
			Replicas: toIntPtr(0),
			Selector: &metav1.LabelSelector{
				MatchLabels: labels,
			},
			Strategy: appsv1.DeploymentStrategy{
				Type: appsv1.RecreateDeploymentStrategyType,
			},
			Template: corev1.PodTemplateSpec{
				ObjectMeta: metav1.ObjectMeta{
					Labels: labels,
				},
				Spec: corev1.PodSpec{
					NodeSelector:                  nodeSelector,
					ServiceAccountName:            "netapp-ha-spare-account",
					PriorityClassName:             quarkconfig.LowPriorityClass,
					TerminationGracePeriodSeconds: toInt64Ptr(0),
					Containers: []corev1.Container{
						{
							Name:            quarkconfig.HASpareDeploymentName,
							Image:           c.haSpareImage,
							ImagePullPolicy: corev1.PullPolicy(c.pullPolicy),
							Resources: corev1.ResourceRequirements{
								Limits:   resources,
								Requests: resources,
							},
						},
					},
				},
			},
		},
	}
}

func (c *haSpareController) getQuarkPodBuckets(
	serviceLevel *quarkv1alpha1.NetAppServiceLevel,
	volume *quarkv1alpha1.NetAppVolume, storagePool *quarkv1alpha1.NetAppStoragePool) (specBucket *quarkv1alpha1.QuarkPodBucket,
	statusBucket *quarkv1alpha1.QuarkPodBucket,
	err error) {

	var size int64
	// If storage pool is present, use its spec size to determine the correct bucket
	if storagePool != nil {
		size = storagePool.Spec.Size.Value()
	} else {
		size = volume.Spec.Size.Value()
	}
	for _, qpb := range serviceLevel.Spec.QuarkPodBuckets {
		minVolSize := qpb.MinVolumeSize.Value()
		maxVolSize := qpb.MaxVolumeSize.Value()
		if size > minVolSize && size <= maxVolSize {
			specBucket = &qpb
			break
		}
	}
	if specBucket != nil {
		for _, qpb := range serviceLevel.Status.QuarkPodBuckets {
			minVolSize := qpb.MinVolumeSize.Value()
			maxVolSize := qpb.MaxVolumeSize.Value()
			if size > minVolSize && size <= maxVolSize {
				statusBucket = &qpb
				break
			}
		}
		return
	}
	err = fmt.Errorf("unable to determine quarkpodbucket for volume: %#v", volume.Name)
	return
}

func (c *haSpareController) findStoragePool(volume *quarkv1alpha1.NetAppVolume) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	if storagePoolUuid, ok := volume.Labels[quarkconfig.StoragePoolLabel]; ok {
		storagePool, err = c.storagePools.NetAppStoragePools(volume.Namespace).Get(storagePoolUuid)
		if err != nil {
			err = fmt.Errorf("failed to get storage pool '%s' for volume '%s': %s", storagePoolUuid, volume.Name, err.Error())
		}
	}
	return
}

func getHASpareKey(qpb quarkv1alpha1.QuarkPodBucket, zone string) string {
	return fmt.Sprintf("ha-spare-%s-%s", qpb.Name, zone)
}

func toIntPtr(in int32) *int32 {
	return &in
}

func toInt64Ptr(in int64) *int64 {
	return &in
}
