package v2

import (
	"context"
	"fmt"
	"strings"

	uuid "github.com/satori/go.uuid"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/apimachinery/pkg/labels"
	"k8s.io/apimachinery/pkg/runtime/schema"
	quarkv1alpha1 "quark.netapp.io/volume-controller/pkg/apis/volumecontroller/v1alpha1"
	"quark.netapp.io/volume-controller/pkg/quarkconfig"
)

// NetAppVolumeCloudBackupListUpdate is used to signal the most recent state of a NetAppVolumeCloudBackupListUpdate and the
// most recent changes.
type NetAppVolumeCloudBackupListUpdate struct {
	VolumeCloudBackupList *quarkv1alpha1.NetAppVolumeCloudBackupList
	UpdatedConditions     []quarkv1alpha1.NetAppVolumeCloudBackupListCondition
}

// constructNetAppVolumeReplication creates a NetAppVolumeReplication struct resource. This api will be called
// internally during first cloud backup create for the volume so this need not be exported to SDE.
func constructNetAppVolumeCloudBackupList(
	ownerID string,
	endpointUUID uuid.UUID, // this will be the name of cloudbackuplist
	storagePoolUUID uuid.UUID,
	volumeUUID uuid.UUID,
	objStoreUUID uuid.UUID,
	volumeReplicationUUID uuid.UUID,
) (cloudBackupList *quarkv1alpha1.NetAppVolumeCloudBackupList, err error) {
	return &quarkv1alpha1.NetAppVolumeCloudBackupList{
		TypeMeta: metav1.TypeMeta{},
		ObjectMeta: metav1.ObjectMeta{
			Name: endpointUUID.String(),
			Labels: map[string]string{
				quarkconfig.StoragePoolLabel:       storagePoolUUID.String(),
				quarkconfig.VolumeLabel:            volumeUUID.String(),
				quarkconfig.VolumeReplicationLabel: volumeReplicationUUID.String(),
				quarkconfig.ObjectStoreLabel:       objStoreUUID.String(),
				quarkconfig.OwnerIDLabel:           ownerID,
			},
		},
		Spec: quarkv1alpha1.NetAppVolumeCloudBackupListSpec{
			DeleteBackups: []string{},
		},
	}, nil
}

// netAppVolumeCloudBackupListCreate creates a NetAppVolumeCloudBackupList CR by making a REST API call to k8s api server.
// This api will be called internally during first cloud backup create for the volume so this need not be exported to SDE.
func (client *NVCClient) netAppVolumeCloudBackupListCreate(
	cloudBackupList *quarkv1alpha1.NetAppVolumeCloudBackupList,
) (*quarkv1alpha1.NetAppVolumeCloudBackupList, error) {
	// TODO-IB any validations here or should those be done in cloud backup create api ?

	volumeReplication, err := client.NetAppVolumeReplicationGet(cloudBackupList.Labels[quarkconfig.VolumeReplicationLabel])
	if err != nil {
		return nil, err
	}
	cloudBackupList.OwnerReferences = []metav1.OwnerReference{
		*metav1.NewControllerRef(volumeReplication, schema.GroupVersionKind{
			Group:   quarkv1alpha1.SchemeGroupVersion.Group,
			Version: quarkv1alpha1.SchemeGroupVersion.Version,
			Kind:    "NetAppVolumeReplication",
		}),
	}

	return client.quarkClient.QuarkV1alpha1().NetAppVolumeCloudBackupLists(client.nvcNamespace).Create(
		context.TODO(), cloudBackupList, metav1.CreateOptions{})
}

// cloudBackupListUUID is the endpointUUID
func (client *NVCClient) NetAppVolumeCloudBackupListGet(cloudBackupListUUID uuid.UUID) (
	*quarkv1alpha1.NetAppVolumeCloudBackupList, error) {
	return client.quarkClient.QuarkV1alpha1().NetAppVolumeCloudBackupLists(client.nvcNamespace).Get(
		context.TODO(), cloudBackupListUUID.String(), metav1.GetOptions{})
}

// NetAppVolumeCloudBackupListGetForSpAndVolume returns a NetAppVolumeCloudBackupList
// object filtered by StoragePool UUID and Volume UUID. It takes in two parameters,
// spUUID and volumeUUID, and returns a NetAppVolumeCloudBackupList object and an error.
// Parameters:
//
//	spUUID - the UUID of the StoragePool
//	volumeUUID - the UUID of the Volume
//
// Returns:
//
//	*quarkv1alpha1.NetAppVolumeCloudBackupList - a NetAppVolumeCloudBackupList object
//	error - error if none found, or any other api error
func (client *NVCClient) NetAppVolumeCloudBackupListGetForSpAndVolume(spUUID uuid.UUID, volumeUUID uuid.UUID) (
	*quarkv1alpha1.NetAppVolumeCloudBackupList, error) {
	matchingLabels := map[string]string{quarkconfig.StoragePoolLabel: spUUID.String(),
		quarkconfig.VolumeLabel: volumeUUID.String()}
	labelSelector := metav1.LabelSelector{MatchLabels: matchingLabels}

	nvbs, err := client.quarkClient.QuarkV1alpha1().NetAppVolumeCloudBackupLists(client.nvcNamespace).List(
		context.TODO(), metav1.ListOptions{LabelSelector: labels.Set(labelSelector.MatchLabels).String()})
	if err != nil {
		return nil, err
	}
	// If nothing matches, filter returns empty list and not an error
	if len(nvbs.Items) == 0 {
		// TODO: Would be good to define a Not-found error type so we can communicate the different
		// error conditions to the caller
		return nil, fmt.Errorf("Not found")
	}
	// There will be only 1 CloudBackupList for sp/vol combo
	return &(nvbs.Items[0]), nil
}

// NetAppVolumeCloudBackupListContainsBackup is a function that checks if a given backupUUID(snapshot) is
// present in a cloudBackupListUUID. It returns a boolean value indicating if the backup is present
// in the list and an error if any.
// Parameters:
//
//	cloudBackupListUUID - the UUID of the cloud backup list
//	backupUUID - the UUID of the backup
//
// Returns:
//
//	bool - true if the backup is present in the list, false otherwise
//	error - error if there is any
func (client *NVCClient) NetAppVolumeCloudBackupListContainsBackup(cloudBackupListUUID, backupUUID uuid.UUID) (bool, error) {
	backupList, err := client.NetAppVolumeCloudBackupListGet(cloudBackupListUUID)
	if err != nil {
		return false, err
	}

	backUUIDStr := backupUUID.String()
	for _, backup := range backupList.Status.Backups {
		if backup.Uuid == backUUIDStr {
			return true, nil
		}
	}
	return false, nil
}

// NetAppVolumeCloudBackupDelete is a function that deletes a specified backup for a given storage
// pool and volume. It takes three parameters: backupToDelete, volumeUUID, and spUUID, and returns
// an error if the delete operation fails.
// Parameters:
//
//	backupToDelete - the backup to delete
//	volumeUUID - the UUID of the volume
//	spUUID - the UUID of the storage pool
//
// Returns:
//
//	error - error if there is any, if the backupList is not found, or a backup delete is already in progress
func (client *NVCClient) NetAppVolumeCloudBackupDelete(backupToDelete uuid.UUID, volumeUUID uuid.UUID, spUUID uuid.UUID) error {

	cloudbackupList, err := client.NetAppVolumeCloudBackupListGetForSpAndVolume(spUUID, volumeUUID)
	if err != nil {
		return err
	}

	if client.backupDeleteInProgress(cloudbackupList) {
		return fmt.Errorf("a delete operation is already in progress for backup %v", cloudbackupList.Spec.DeleteBackups[0])
	}

	cloudbackupList.Spec.DeleteBackups = []string{backupToDelete.String()}

	_, err = client.netAppVolumeCloudBackupListUpdate(cloudbackupList)
	return err
}

func (client *NVCClient) NetAppVolumeCloudBackupEndpointDelete(volumeUUID uuid.UUID, spUUID uuid.UUID) error {
	cloudbackupList, err := client.NetAppVolumeCloudBackupListGetForSpAndVolume(spUUID, volumeUUID)
	if err != nil {
		return err
	}

	if client.backupDeleteInProgress(cloudbackupList) {
		return fmt.Errorf("a delete operation is already in progress for backup %v", cloudbackupList.Spec.DeleteBackups[0])
	}

	cloudbackupList.Spec.DeleteBackups = []string{"*"}

	_, err = client.netAppVolumeCloudBackupListUpdate(cloudbackupList)
	return err
}

func (client *NVCClient) netAppVolumeCloudBackupListUpdate(
	cloudBackupList *quarkv1alpha1.NetAppVolumeCloudBackupList,
) (*quarkv1alpha1.NetAppVolumeCloudBackupList, error) {
	newCloudBackupList, err := client.quarkClient.QuarkV1alpha1().NetAppVolumeCloudBackupLists(client.nvcNamespace).Update(
		context.TODO(), cloudBackupList, metav1.UpdateOptions{})
	if err != nil {
		if strings.Contains(err.Error(), quarkconfig.OldVersionErr) {
			return nil, StaleVersionError{}
		}
		return nil, err
	}
	return newCloudBackupList, nil
}

func (client *NVCClient) backupDeleteInProgress(cloudbackupList *quarkv1alpha1.NetAppVolumeCloudBackupList) bool {
	if len(cloudbackupList.Spec.DeleteBackups) == 1 &&
		(len(cloudbackupList.Status.DeleteBackups) == 0 ||
			cloudbackupList.Spec.DeleteBackups[0] != cloudbackupList.Status.DeleteBackups[0]) {
		return true
	}
	return false
}

func netAppVolumeCloudBackupDeleted(cloudBackupList *quarkv1alpha1.NetAppVolumeCloudBackupList) bool {
	return len(cloudBackupList.Spec.DeleteBackups) == 1 && len(cloudBackupList.Status.DeleteBackups) == 1 &&
		cloudBackupList.Spec.DeleteBackups[0] == cloudBackupList.Status.DeleteBackups[0]
}

func netAppVolumeCloudBackupEndpointDeletedByDMAP(cloudBackupList *quarkv1alpha1.NetAppVolumeCloudBackupList) bool {
	return len(cloudBackupList.Spec.DeleteBackups) == 1 && len(cloudBackupList.Status.DeleteBackups) == 1 &&
		cloudBackupList.Spec.DeleteBackups[0] == "*" && cloudBackupList.Status.DeleteBackups[0] == "*"
}
