package v2

import (
	"context"
	"fmt"
	"reflect"
	"regexp"
	"sort"
	"strconv"
	"strings"
	"time"

	uuid "github.com/satori/go.uuid"
	log "github.com/sirupsen/logrus"
	corev1 "k8s.io/api/core/v1"
	"k8s.io/apimachinery/pkg/api/errors"
	"k8s.io/apimachinery/pkg/api/resource"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/apimachinery/pkg/runtime/schema"
	utilruntime "k8s.io/apimachinery/pkg/util/runtime"

	apiextensionsclient "k8s.io/apiextensions-apiserver/pkg/client/clientset/clientset"
	"k8s.io/client-go/kubernetes"
	_ "k8s.io/client-go/plugin/pkg/client/auth"
	"k8s.io/client-go/rest"
	"k8s.io/client-go/tools/cache"
	"k8s.io/client-go/tools/clientcmd"

	quarkv1alpha1 "quark.netapp.io/volume-controller/pkg/apis/volumecontroller/v1alpha1"
	quarkclientset "quark.netapp.io/volume-controller/pkg/client/clientset/versioned"

	quarkinformers "quark.netapp.io/volume-controller/pkg/client/informers/externalversions"
	quarklisters "quark.netapp.io/volume-controller/pkg/client/listers/volumecontroller/v1alpha1"
	"quark.netapp.io/volume-controller/pkg/nvc"
	"quark.netapp.io/volume-controller/pkg/quarkconfig"
	"quark.netapp.io/volume-controller/pkg/util"
)

var (
	// enforcing 1 Gi volume increment for beta version and onwards
	incrementalSize resource.Quantity = resource.MustParse("1Gi")
	minVolSize      resource.Quantity = resource.MustParse("100Gi")
	// minimum volume size for storagepool is 1Gi
	minSpVolSize resource.Quantity = resource.MustParse("1Gi")
	maxVolSize   resource.Quantity = resource.MustParse("100Ti")
)

const (
	versionUpgradeErr = "Failed to upgrade NetAppVolume"
	upgradeAbortErr   = "Failed to abort NetAppVolumeUpgrade"
)

type NVCClientConflictError struct {
	err string
}

func (ce *NVCClientConflictError) Error() string {
	return ce.err
}

func NewNVCClientConflictError(errMsg string) *NVCClientConflictError {
	return &NVCClientConflictError{errMsg}
}

type StaleVersionError struct {
}

func (e StaleVersionError) Error() string {
	return "StaleObjectVersion"
}

type CloneVolumeSplitScanInProgressError struct {
	CloneList []quarkv1alpha1.NetAppVolume
}

func (e CloneVolumeSplitScanInProgressError) Error() string {
	return fmt.Sprintf("split scan in progress for clone volume associated with object %v", e.CloneList)
}

type CloneVolumeBaseSnapshotError struct {
}

func (e CloneVolumeBaseSnapshotError) Error() string {
	return fmt.Sprintf("snapshot is the base snapshot of a clone volume that has split scan ongoing")
}

type RestoreSnapshotIsCloneLockedError struct{}

func (e RestoreSnapshotIsCloneLockedError) Error() string {
	return "snapshot restore operation not allowed because this operation would destroy later snapshots that are required for cloned volumes"
}

type ClientConfig struct {
	K8sApiRetryInterval time.Duration
	K8sApiRetryCount    int
}

type ClientOption func(*ClientConfig)

func WithApiRetryInterval(timeout int) ClientOption {
	return func(config *ClientConfig) {
		config.K8sApiRetryInterval = time.Duration(timeout) * time.Second
	}
}

func WithApiRetryCount(retries int) ClientOption {
	return func(config *ClientConfig) {
		config.K8sApiRetryCount = retries
	}
}

type NVCClient struct {
	// NVCClient config
	Config ClientConfig

	// Kubernetes clientset
	kubeClient kubernetes.Interface
	// apiextensionsclient is the clientset for Crds
	crdClient apiextensionsclient.Interface
	// Quark clientset
	quarkClient quarkclientset.Interface
	// Kubernetes cluster config
	kubeConfig *rest.Config
	// NVC namespace (i.e., where NVC custom resources will be created)
	nvcNamespace string
	// Client event channels
	volumeDeleteChannel                chan<- *NetAppVolumeDelete
	volumeUpdateChannel                chan<- *NetAppVolumeUpdate
	snapshotDeleteChannel              chan<- *NetAppVolumeSnapshotDelete
	snapshotUpdateChannel              chan<- *NetAppVolumeSnapshotUpdate
	snapshotRestoreChannel             chan<- *NetAppVolumeSnapshotRestore
	snapshotsListChannel               chan<- *NetAppVolumeSnapshotsListUpdate
	cloudSnapshotUpdateChannel         chan<- *NetAppVolumeCloudSnapshotUpdate
	cloudSnapshotDeleteChannel         chan<- *NetAppVolumeCloudSnapshotDelete
	exportPolicyUpdateChannel          chan<- *NetAppVolumeExportPolicyUpdate
	snapshotPolicyUpdateChannel        chan<- *NetAppVolumeSnapshotPolicyUpdate
	quarkUpdateChannel                 chan<- *NetAppQuarkUpdate
	quarkDeleteChannel                 chan<- *NetAppQuarkDelete
	storagePoolDeleteChannel           chan<- *NetAppStoragePoolDelete
	storagePoolUpdateChannel           chan<- *NetAppStoragePoolUpdate
	volumeReplicationUpdateChannel     chan<- *NetAppVolumeReplicationUpdate
	volumeReplicationDeleteChannel     chan<- *NetAppVolumeReplicationDelete
	clusterPeeringUpdateChannel        chan<- *NetAppClusterPeeringUpdate
	clusterPeeringDeleteChannel        chan<- *NetAppClusterPeeringDelete
	objectStoreUpdateChannel           chan<- *NetAppObjectStoreUpdate
	objectStoreDeleteChannel           chan<- *NetAppObjectStoreDelete
	volumeCloudBackupListUpdateChannel chan<- *NetAppVolumeCloudBackupListUpdate

	// Controller parameters to monitor NetAppVolumes
	volumesLister               quarklisters.NetAppVolumeLister
	volumesSynced               cache.InformerSynced
	storagePoolLister           quarklisters.NetAppStoragePoolLister
	storagePoolSynced           cache.InformerSynced
	snapshotsLister             quarklisters.NetAppVolumeSnapshotLister
	snapshotsSynced             cache.InformerSynced
	snapshotsListLister         quarklisters.NetAppVolumeSnapshotsListLister
	snapshotsListSynced         cache.InformerSynced
	cloudSnapshotsLister        quarklisters.NetAppVolumeCloudSnapshotLister
	cloudSnapshotsSynced        cache.InformerSynced
	exportPolicyLister          quarklisters.NetAppExportPolicyLister
	exportPolicySynced          cache.InformerSynced
	snapshotPolicyLister        quarklisters.NetAppSnapshotPolicyLister
	snapshotPolicySynced        cache.InformerSynced
	quarksLister                quarklisters.QuarkLister
	quarksSynced                cache.InformerSynced
	volumeReplicationsLister    quarklisters.NetAppVolumeReplicationLister
	volumeReplicationSynced     cache.InformerSynced
	clusterPeeringLister        quarklisters.NetAppClusterPeeringLister
	clusterPeeringSynced        cache.InformerSynced
	objectStoreLister           quarklisters.NetAppObjectStoreLister
	objectStoreSynced           cache.InformerSynced
	volumeCloudBackupListLister quarklisters.NetAppVolumeCloudBackupListLister
	volumeCloudBackupListSynced cache.InformerSynced
	//workqueue              workqueue.RateLimitingInterface
	//workqueueSnapshot      workqueue.RateLimitingInterface
	//workqueueSnapshotsList workqueue.RateLimitingInterface
	//knownVolumes             map[string]bool
	//knownVolumeSnapshots     map[string]bool
	//knownVolumeSnapshotsList map[string]bool
	DarkContent               bool
	EnableCCPd                bool
	EnableCRR                 bool
	EnableLdap                bool
	EnableGcp1p               bool
	EnableGranularLiveness    bool
	EnableNfs41               bool
	EnableServiceMesh         bool
	EnableBackupSecurityUsers bool
	EnableIntegratedBackup    bool
}

type NetAppVolumeUsage quarkv1alpha1.NetAppVolumeUsageStatus

type ExportRule quarkv1alpha1.ExportRule

func newNVCClient(kubeConfig *rest.Config, namespace string, options ...ClientOption) (*NVCClient, error) {
	// Setting up clientsets
	kubeClient, err := kubernetes.NewForConfig(kubeConfig)
	if err != nil {
		return nil, err
	}
	crdClient, err := apiextensionsclient.NewForConfig(kubeConfig)
	if err != nil {
		return nil, err
	}
	quarkClient, err := quarkclientset.NewForConfig(kubeConfig)
	if err != nil {
		return nil, err
	}

	client := &NVCClient{
		kubeClient:   kubeClient,
		crdClient:    crdClient,
		quarkClient:  quarkClient,
		kubeConfig:   kubeConfig,
		nvcNamespace: namespace,
		Config: ClientConfig{
			K8sApiRetryInterval: 1 * time.Second,
			K8sApiRetryCount:    3,
		},
		//knownVolumes:             make(map[string]bool),
		//knownVolumeSnapshots:     make(map[string]bool),
		//knownVolumeSnapshotsList: make(map[string]bool),
	}

	quarkCRs, err := client.NetAppQuarkGetAll()
	if err != nil {
		if errors.IsNotFound(err) {
			//  len(quarkCRs) = 0, likely first install
			return client, nil
		}
		return nil, err
	}
	if len(quarkCRs) == 1 {
		// set feature flags looking at quark CR
		client.EnableCCPd = quarkCRs[0].Spec.EnableCcpd
		client.EnableCRR = quarkCRs[0].Spec.EnableCrr
		client.EnableLdap = quarkCRs[0].Spec.EnableLdap
		client.EnableGcp1p = quarkCRs[0].Spec.EnableGcp1p
		client.EnableServiceMesh = quarkCRs[0].Spec.EnableServiceMesh
		client.EnableNfs41 = quarkCRs[0].Spec.EnableNfs41
		client.EnableGranularLiveness = quarkCRs[0].Spec.EnableGranularLiveness
		client.EnableBackupSecurityUsers = quarkCRs[0].Spec.EnableBackupSecurityUsers
		client.EnableIntegratedBackup = quarkCRs[0].Spec.EnableIntegratedBackup
	} else if len(quarkCRs) > 1 {
		// do not allow the creation of nvcclient if more than one quark CR is present in the namespace
		return nil, fmt.Errorf("more than one quark CR found in %v namespace", client.nvcNamespace)
	}
	for _, option := range options {
		option(&client.Config)
	}
	return client, nil
}

// Creates and initializes NVCClient
// kubeConfigPath: Path to cluster config file
// apiServerIP: Api server IP, overrides kubeConfigPath if specified
// namespace: Namespace where NVC is deployed
func NewNVCClient(kubeConfigPath, apiServerIP, namespace string, options ...ClientOption) (*NVCClient, error) {
	kubeConfig, err := clientcmd.BuildConfigFromFlags(apiServerIP, kubeConfigPath)
	if err != nil {
		return nil, err
	}
	return newNVCClient(kubeConfig, namespace, options...)
}

// Creates and initializes NVCClient
// kubeConfigPath: Path to cluster config file
// apiServerIP: API server IP, overrides kubeConfigPath if specified
// namespace: Namespace where NVC is deployed
// minVolSize: Minimum supported volume size
func NewNVCClientWithSize(kubeConfigPath, apiServerIP, namespace string, minVolSizeString string, options ...ClientOption) (client *NVCClient, err error) {
	minVolSize, err = resource.ParseQuantity(minVolSizeString)
	if err != nil {
		return nil, fmt.Errorf("invalid value for size")
	}
	minSpVolSize, err = resource.ParseQuantity(minVolSizeString)
	if err != nil {
		return nil, fmt.Errorf("invalid value for size")
	}

	return NewNVCClient(kubeConfigPath, apiServerIP, namespace, options...)
}

// Creates and initializes NVCClient
// kubeConfigBytes: kubeconfig content
// namespace: Namespace where NVC is deployed
func NewNVCClientWithKubeConfigBytes(kubeConfigBytes []byte, namespace string, options ...ClientOption) (*NVCClient, error) {
	kubeConfig, err := clientcmd.RESTConfigFromKubeConfig(kubeConfigBytes)
	if err != nil {
		return nil, err
	}
	return newNVCClient(kubeConfig, namespace, options...)
}

// Creates and initializes NVCClient
// kubeConfig: kubeconfig content
// namespace: Namespace where NVC is deployed
func NewNVCClientWithRestConfig(kubeConfig *rest.Config, namespace string, options ...ClientOption) (*NVCClient, error) {
	return newNVCClient(kubeConfig, namespace, options...)
}

// Creates and initializes NVCClient
// kubeConfigPath: Path to cluster config file
// apiServerIP: Api server IP, overrides kubeConfigPath if specified
// namespace: Namespace where NVC is deployed
// darkContentFlag: flag to enable dark content
func NewNVCClientDark(kubeConfigPath, apiServerIP, namespace string, darkContentFlag bool, options ...ClientOption) (*NVCClient, error) {
	client, err := NewNVCClient(kubeConfigPath, apiServerIP, namespace, options...)
	if err == nil {
		client.SetDarkContent(darkContentFlag)
	}
	return client, err
}

// Creates and initializes NVCClient
// kubeConfigPath: Path to cluster config file
// apiServerIP: API server IP, overrides kubeConfigPath if specified
// namespace: Namespace where NVC is deployed
// minVolSize: Minimum supported volume size
// darkContentFlag: flag to enable dark content
func NewNVCClientWithSizeDark(kubeConfigPath, apiServerIP, namespace string, minVolSizeString string, darkContentFlag bool, options ...ClientOption) (*NVCClient, error) {
	client, err := NewNVCClientWithSize(kubeConfigPath, apiServerIP, namespace, minVolSizeString, options...)
	if err == nil {
		client.SetDarkContent(darkContentFlag)
	}
	return client, err
}

// Creates and initializes NVCClient
// kubeConfigBytes: kubeconfig content
// namespace: Namespace where NVC is deployed
// darkContentFlag: flag to enable dark content
func NewNVCClientWithKubeConfigBytesDark(kubeConfigBytes []byte, namespace string, darkContentFlag bool, options ...ClientOption) (*NVCClient, error) {
	client, err := NewNVCClientWithKubeConfigBytes(kubeConfigBytes, namespace, options...)
	if err == nil {
		client.SetDarkContent(darkContentFlag)
	}
	return client, err
}

// SetDarkContent sets the value for the "dark content" flag
func (client *NVCClient) SetDarkContent(val bool) {
	client.DarkContent = val
}

func (client *NVCClient) NetAppVolumeCreate(volume *quarkv1alpha1.NetAppVolume) (*quarkv1alpha1.NetAppVolume, error) {

	if _, ok := volume.Labels[quarkconfig.StoragePoolLabel]; ok {
		// Don't allow volume creation if the storagePool is being deleted
		storagePool, err := client.NetAppStoragePoolGet(volume.Labels[quarkconfig.StoragePoolLabel])
		if err != nil || storagePool == nil {
			return nil, fmt.Errorf("failed to get the NetappStoragePool: %v", err)
		}
		if storagePool.ObjectMeta.DeletionTimestamp != nil {
			return nil, fmt.Errorf("cannot create volume on, Storage pool, '%s', deletion in progress", volume.Labels[quarkconfig.StoragePoolLabel])
		}

		// Make sure the SMB configured in storagepool
		if volume.Spec.Protocol == "cifs" {
			smbConfigured, err := client.NetAppStoragePoolGetSMBStatus(volume.Labels[quarkconfig.StoragePoolLabel])
			if err != nil {
				return nil, fmt.Errorf("failed to get SMB server status for Storage pool: %s", err.Error())
			}
			if !smbConfigured {
				return nil, fmt.Errorf("SMB server not configured for Storage pool")
			}
		} else {
			// Don't allow nfs volume creation if ldap configuration is enabled and not ready
			if storagePool.Spec.LdapEnabled && storagePool.Spec.LdapEnabled != storagePool.Status.LdapEnabled {
				return nil, fmt.Errorf("cannot create volume on, Storage pool, '%s', Ldap configuration not ready", volume.Labels[quarkconfig.StoragePoolLabel])
			}

			var kerberosRulePresent bool
			if exportPolicy, err := client.NetAppExportPolicyGet(volume.Spec.ExportPolicy); err == nil {
				rules := make([]ExportRule, len(exportPolicy.Spec.Rules))
				for index, value := range exportPolicy.Spec.Rules {
					rules[index].RuleIndex = value.RuleIndex
					rules[index].SecurityType = value.SecurityType
				}
				kerberosRulePresent = isKerberosRulePresent(rules)
			}

			if volume.Spec.KerberosEnabled {
				// Don't allow nfs4.1 volume creation if kerberos is configured but not ready.
				isKerbConfiguredInSP, err := client.NetAppStoragePoolKerberosReady(storagePool)
				if !isKerbConfiguredInSP {
					return nil, fmt.Errorf("cannot create kerberos volume on, Storage pool: %s, Kerberos not ready: %w", storagePool.Name, err)
				}
				// Don't allow volume creation if "krb" SecurityType is not present in export.
				if kerberosRulePresent == false {
					return nil, fmt.Errorf("failed to create volume, Kerberos volume must have Kerberos Security in export policy. ExportPolicy: %s", volume.Spec.ExportPolicy)
				}
			} else {
				// Restricting volume creation when kerberos not enabled and "krb" SecurityType is used in export.
				if kerberosRulePresent {
					return nil, fmt.Errorf("failed to create volume, Non-Kerberos volume cannot have Kerberos Security in export policy. ExportPolicy: %s", volume.Spec.ExportPolicy)
				}

			}
		}
		availableSpace, err := client.NetAppStoragePoolGetAvailableSpace(volume.Labels[quarkconfig.StoragePoolLabel])
		if err != nil {
			return nil, fmt.Errorf("failed to get available space for Storage pool: %s", err.Error())
		}
		if availableSpace < volume.Spec.Size.Value() {
			return nil, fmt.Errorf("insufficient space left in storagePool, free space=%d, requested=%s", availableSpace, volume.Spec.Size.String())
		}

		// if CMEK is enabled in SP, then volume cannot be created until encryptionInitialized is true or if CMEKConfigurationError is seen
		// NetAppStoragePoolGetAvailableSpace fails if SP is not online
		if util.IsCMEKConfigured(storagePool) {
			errConditionPresent, _ := util.IsErrorConditionPresent(storagePool.Status.Conditions, quarkv1alpha1.NetAppStoragePoolCMEKConfigurationError)
			if errConditionPresent || !storagePool.Status.EncryptionInitialized {
				return nil, fmt.Errorf("cannot create volume as encryption is still not initialized on the storage pool, this storage pool has encryption enabled")
			}
		}
	}
	// operator allows creation of one quark CR only, only one element will be present in quarks list
	quarks, err := client.NetAppQuarkGetAll()
	if err != nil {
		return nil, fmt.Errorf("failed to get Quark: %s", err.Error())
	}
	if len(quarks) == 0 {
		return nil, fmt.Errorf("no quarks found")
	}
	if _, ok := volume.Labels[quarkconfig.StoragePoolLabel]; !ok {
		quark := quarks[0]
		// add quark version to create NetAppVolume only for "Production" & "TestVariant" configs
		switch strings.ToLower(quark.Spec.ControllerConfig.Variant) {
		case quarkconfig.ProductionVariant, quarkconfig.OnPremVariant:
			volume.Spec.QuarkVersion = quark.Status.Version
		}
	}

	if client.EnableNfs41 && (volume.Spec.Protocol == quarkconfig.DefaultNFS4Protocol || volume.Spec.Protocol == quarkconfig.DefaultNFS3Protocol) {
		isValidProtocol := client.ValidateProtocolMatch(volume.Spec.ExportPolicy, []string{volume.Spec.Protocol})
		if !isValidProtocol {
			return nil, fmt.Errorf("Protocol mismatch with export policy")
		}
	}

	result, err := client.createWithRetry(volume)
	if err != nil {
		return nil, err
	}
	return result.(*quarkv1alpha1.NetAppVolume), nil
}

// Label a quark volume with SDE's volume uuid and storage pool uuid.
// This function expects 3 arguments .
// volumeName is supposed to be name of the volume to label. sdeVolUUID and sdeSpUUID are
// volume and SP uuid assigned by SDE.
func (client *NVCClient) LabelNetAppVolume(volumeName, sdeVolUUID, sdeSpUUID string) error {
	if sdeVolUUID == "" || sdeSpUUID == "" {
		return fmt.Errorf("provide non empty uuid of volume and storage pool")
	}

	vol, err := client.quarkClient.QuarkV1alpha1().NetAppVolumes(client.nvcNamespace).Get(context.Background(), volumeName, metav1.GetOptions{})
	if err != nil {
		return err
	}
	if len(vol.Labels) == 0 {
		vol.Labels = make(map[string]string)
	}
	vol.Labels[quarkconfig.SDEVolumeUUID] = sdeVolUUID
	vol.Labels[quarkconfig.SDESpUUID] = sdeSpUUID

	_, err = client.netAppVolumeUpdate(vol)

	return err
}

func (client *NVCClient) validateADCredentials(user, password string) error {
	if user == "" {
		return fmt.Errorf("username cannot be empty")
	}
	if len(user) > 104 {
		return fmt.Errorf("username cannot be more than 104 characters long")
	}
	if password == "" {
		return fmt.Errorf("password cannot be empty")
	}
	return nil
}

// Creates AD Secret without ownerRef set
func (client *NVCClient) createOrUpdateADSecret(secretName, user, password string) (*corev1.Secret, error) {
	err := client.validateADCredentials(user, password)
	if err != nil {
		return nil, err
	}

	secretData := map[string][]byte{
		"username": []byte(user),
		"password": []byte(password),
	}
	secret := &corev1.Secret{
		TypeMeta: metav1.TypeMeta{},
		Data:     secretData,
		Type:     corev1.SecretTypeOpaque,
	}
	secret.Name = secretName
	secret.Namespace = client.nvcNamespace
	return client.SecretCreateOrUpdate(secret)
}

// Sets ownerRef field of secret
func (client *NVCClient) setSecretOwnerRef(secretName string, volume *quarkv1alpha1.NetAppVolume) (*corev1.Secret, error) {
	secret, err := client.SecretGet(secretName)
	if err != nil {
		return nil, err
	}
	secret.ObjectMeta = nvc.MakeObjectMeta(volume, nil)
	secret.Name = secretName

	return client.SecretUpdate(secret)

}

// Creates AD Secret with ownerRef set
func (client *NVCClient) createOrUpdateADSecretWithOwnerRef(volume *quarkv1alpha1.NetAppVolume, storagePool *quarkv1alpha1.NetAppStoragePool, secretName, user, password string) (*corev1.Secret, error) {
	var secret *corev1.Secret
	err := client.validateADCredentials(user, password)
	if err != nil {
		return nil, err
	}

	secretData := map[string][]byte{
		"username": []byte(user),
		"password": []byte(password),
	}
	if storagePool != nil {
		secret = &corev1.Secret{
			TypeMeta:   metav1.TypeMeta{},
			ObjectMeta: nvc.MakeObjectMeta(nil, storagePool),
			Data:       secretData,
			Type:       corev1.SecretTypeOpaque,
		}
	} else {
		secret = &corev1.Secret{
			TypeMeta:   metav1.TypeMeta{},
			ObjectMeta: nvc.MakeObjectMeta(volume, nil),
			Data:       secretData,
			Type:       corev1.SecretTypeOpaque,
		}
	}
	secret.Name = secretName
	secret.Namespace = client.nvcNamespace

	return client.SecretCreateOrUpdate(secret)
}

// Initiate NetAppVolume version upgrade
func (client *NVCClient) NetAppVolumeUpgrade(volumeUUID, version string) (*quarkv1alpha1.NetAppVolume, error) {
	volume, err := client.NetAppVolumeGet(volumeUUID)
	if err != nil {
		return nil, err
	}
	errMsg := isConflictingVolUpdateInProgess(volume)
	if errMsg != "" {
		return nil, NewNVCClientConflictError(fmt.Sprintf("failed to upgrade NetAppVolume %s: %s", volumeUUID, errMsg))
	}
	// operator allows creation of one quark CR only, only one element will be present in quarks list
	quarks, err := client.NetAppQuarkGetAll()
	if err != nil {
		return nil, fmt.Errorf("%s: %s", versionUpgradeErr, err)
	}
	if len(quarks) == 0 {
		return nil, fmt.Errorf("%s: no quarks found", versionUpgradeErr)
	}
	quark := quarks[0]
	if quark.Status.Version != version {
		return nil, fmt.Errorf("%s: quark is not updated to version %s", versionUpgradeErr, version)
	}
	// initiate upgrade when Quark is Ready and updated with required version for volume upgrade
	volume.Spec.QuarkVersion = version
	newVol, err := client.netAppVolumeUpdate(volume)

	return newVol, err
}

// Creates SMB Volume
func (client *NVCClient) NetAppVolumeCreateSMB(vol *quarkv1alpha1.NetAppVolume, user, password string) (volume *quarkv1alpha1.NetAppVolume, err error) {
	var secretName string
	if _, ok := vol.Labels[quarkconfig.StoragePoolLabel]; !ok {
		secretName = util.GetVolCreateADSecretName(vol.Name)
		_, err = client.createOrUpdateADSecret(secretName, user, password)
		if err != nil {
			log.WithFields(log.Fields{
				"volume": vol.Name,
				"secret": secretName,
			}).Error("Failed to create secret.")
			return nil, err
		}
	}
	volume, err = client.NetAppVolumeCreate(vol)
	if err != nil {
		log.WithFields(log.Fields{
			"volume": vol.Name,
			"secret": secretName,
		}).Error("Failed to create volume.")
		return nil, err
	}

	if _, ok := vol.Labels[quarkconfig.StoragePoolLabel]; !ok {
		_, err = client.setSecretOwnerRef(secretName, volume)
		// If secret is already deleted ignore the error.
		if err != nil && !errors.IsNotFound(err) {
			// Delete Secret and Volume if can't set ownerRef
			_ = client.SecretDelete(secretName)
			_ = client.NetAppVolumeDelete(volume.Name)
			log.WithFields(log.Fields{
				"volume": vol.Name,
				"secret": secretName,
			}).Error("Failed to set OwnerRef field.")

			return nil, err
		}
	}
	return volume, nil
}

func (client *NVCClient) NetAppVolumeDeleteSMB(volumeUUID string, user, password string) error {

	// Make sure there are no clones associated to this volume
	cloneList, err := client.getAssociatedCloneVolumeList(volumeUUID, "")
	if err != nil {
		return fmt.Errorf("failed to get list of associated clone volumes: %v", err)
	}
	if len(cloneList) != 0 {
		return CloneVolumeSplitScanInProgressError{CloneList: cloneList}
	}
	volume, err := client.NetAppVolumeGet(volumeUUID)
	if err != nil {
		if errors.IsNotFound(err) {
			return nil
		} else {
			return err
		}
	}
	if _, ok := volume.Labels[quarkconfig.StoragePoolLabel]; !ok {
		_, err = client.createOrUpdateADSecretWithOwnerRef(volume, nil, util.GetVolDeleteADSecretName(volume.Name), user, password)
		if err != nil {
			return err
		}
	}

	if client.EnableCRR {
		repl, err := client.isVolPartOfReplication(volumeUUID)
		if err != nil {
			return err
		}

		if repl != "" {
			return fmt.Errorf("failed to delete the NetAppVolume %v: it is the destination volume for NetAppVolumeReplication %v, delete NetAppVolumeReplication before attempting to delete the NetAppVolume",
				volumeUUID, repl)
		}
	}

	return client.deleteWithRetry(volume)
}

func (client *NVCClient) NetAppVolumeSnapshotGet(snapUUID string) (*quarkv1alpha1.NetAppVolumeSnapshot, error) {
	snap := &quarkv1alpha1.NetAppVolumeSnapshot{
		ObjectMeta: metav1.ObjectMeta{
			Name: snapUUID,
		},
	}
	result, err := client.getWithRetry(snap)
	if err != nil {
		return nil, err
	}
	return result.(*quarkv1alpha1.NetAppVolumeSnapshot), nil
}

func (client *NVCClient) NetAppVolumeSnapshotsListGet(volumeUUID string) (*quarkv1alpha1.NetAppVolumeSnapshotsList, error) {
	snapList := &quarkv1alpha1.NetAppVolumeSnapshotsList{
		ObjectMeta: metav1.ObjectMeta{
			Name: volumeUUID,
		},
	}
	result, err := client.getWithRetry(snapList)
	if err != nil {
		return nil, err
	}
	return result.(*quarkv1alpha1.NetAppVolumeSnapshotsList), nil
}

func (client *NVCClient) NetAppVolumeGet(volumeUUID string) (*quarkv1alpha1.NetAppVolume, error) {
	volume := &quarkv1alpha1.NetAppVolume{
		ObjectMeta: metav1.ObjectMeta{
			Name: volumeUUID,
		},
	}
	result, err := client.getWithRetry(volume)
	if err != nil {
		return nil, err
	}
	return result.(*quarkv1alpha1.NetAppVolume), nil
}

func (client *NVCClient) NetAppVolumeGetVersion(volumeUUID string) (string, error) {
	volume, err := client.NetAppVolumeGet(volumeUUID)
	if err != nil {
		return "", fmt.Errorf("failed to get the NetappVolume version: %s", err.Error())
	}
	return volume.Status.QuarkVersion, nil
}

// returns conditions from NetAppVolume CR status
func (client *NVCClient) NetAppVolumeGetConditions(volumeUUID string) ([]quarkv1alpha1.NetAppVolumeCondition, error) {
	volume, err := client.quarkClient.QuarkV1alpha1().NetAppVolumes(client.nvcNamespace).Get(
		context.TODO(), volumeUUID, metav1.GetOptions{})
	if err != nil {
		return nil, fmt.Errorf("failed to get the NetappVolume: %s", err.Error())
	}
	return volume.Status.Conditions, nil
}

func (client *NVCClient) NetAppVolumeStatus(volume *quarkv1alpha1.NetAppVolume) (bool, string, *quarkv1alpha1.NetAppStoragePool, error) {
	var (
		volLifeCycleState        bool
		volLifeCycleStateDetails = "Volume is in error state"
	)

	if _, ok := volume.Labels[quarkconfig.StoragePoolLabel]; !ok {
		return volLifeCycleState, volLifeCycleStateDetails, nil, fmt.Errorf("volume %s is not part of any storagepool", volume.Name)
	}
	storagePoolOnline, storagePoolStateDetails, storagePool, err := client.NetAppStoragePoolStatus(volume.Labels[quarkconfig.StoragePoolLabel])
	if err != nil || !storagePoolOnline {
		return volLifeCycleState, fmt.Sprintf("%s %s", volume.Labels[quarkconfig.StoragePoolLabel], storagePoolStateDetails), storagePool, err
	}

	volCLoudRestoreCondition := client.NetAppVolumeCloudSnapshotRestoreCondition(volume)
	if client.NetAppVolumeOnline(volume) {
		volLifeCycleState = true
		if volCLoudRestoreCondition != nil &&
			volCLoudRestoreCondition.Type == quarkv1alpha1.NetAppVolumeCloudSnapshotRestoring {
			volLifeCycleStateDetails = "Volume is being restored"
		} else {
			volLifeCycleStateDetails = "Volume is available"
		}
	} else if volCLoudRestoreCondition != nil &&
		volCLoudRestoreCondition.Type == quarkv1alpha1.NetAppVolumeCloudSnapshotPreparing {
		volLifeCycleStateDetails = "Volume is being created"
	}

	return volLifeCycleState, volLifeCycleStateDetails, storagePool, err
}

// returns conditions from NetAppVolumeSnapshotsList CR status
func (client *NVCClient) NetAppVolumeSnapshotsListGetConditions(volumeUUID string) ([]quarkv1alpha1.NetAppVolumeSnapshotsListCondition, error) {
	snapList, err := client.NetAppVolumeSnapshotsListGet(volumeUUID)
	if err != nil {
		return nil, fmt.Errorf("failed to get the NetappVolumeSnapshotsList: %s", err.Error())
	}
	return snapList.Status.Conditions, nil
}

func (client *NVCClient) NetAppVolumeGetAll() ([]*quarkv1alpha1.NetAppVolume, error) {
	volumes := make([]*quarkv1alpha1.NetAppVolume, 0)
	result, err := client.listWithRetry(&quarkv1alpha1.NetAppVolumeList{}, metav1.ListOptions{})
	if err != nil {
		return nil, err
	}
	volumesList := result.(*quarkv1alpha1.NetAppVolumeList)
	for i := range volumesList.Items {
		volumes = append(volumes, &volumesList.Items[i])
	}
	return volumes, nil
}

func (client *NVCClient) NetAppVolumeGetAllSMB() ([]*quarkv1alpha1.NetAppVolume, error) {
	volumes := make([]*quarkv1alpha1.NetAppVolume, 0)
	result, err := client.listWithRetry(&quarkv1alpha1.NetAppVolumeList{}, metav1.ListOptions{})
	if err != nil {
		return nil, err
	}
	volumesList := result.(*quarkv1alpha1.NetAppVolumeList)
	for i := range volumesList.Items {
		if volumesList.Items[i].Spec.Protocol == "cifs" {
			volumes = append(volumes, &volumesList.Items[i])
		}
	}
	return volumes, err
}

func (client *NVCClient) isVolPartOfReplication(volumeUUID string) (string, error) {
	volReplications, err := client.NetAppVolumeReplicationGetAll()
	if err != nil {
		return "", err
	}

	for _, volRepl := range volReplications {
		if volRepl.Spec.DestinationVolume.VolumeUUID == volumeUUID || volRepl.Spec.Destination.UUID == volumeUUID {
			return volRepl.Name, nil
		}
	}
	return "", nil
}

func (client *NVCClient) NetAppVolumeDelete(volumeUUID string) error {

	// Make sure there are no clones associated to this volume
	cloneList, err := client.getAssociatedCloneVolumeList(volumeUUID, "")
	if err != nil {
		return fmt.Errorf("failed to get list of associated clone volumes: %v", err)
	}
	if len(cloneList) != 0 {
		return CloneVolumeSplitScanInProgressError{CloneList: cloneList}
	}
	volume, err := client.NetAppVolumeGet(volumeUUID)
	if err != nil {
		return err
	}
	if client.isVolUpgradeInProgess(volume) {
		return NewNVCClientConflictError(fmt.Sprintf("failed to delete the NetAppVolume %s: %s", volumeUUID, quarkconfig.MessageUpgradeInProgress))
	}

	if client.EnableCRR {
		repl, err := client.isVolPartOfReplication(volumeUUID)
		if err != nil {
			return err
		}

		if repl != "" {
			return fmt.Errorf("failed to delete NetAppVolume %v: it is the destination volume for NetAppVolumeReplication %v, delete NetAppVolumeReplication before attempting to delete the NetAppVolume",
				volumeUUID, repl)
		}
	}
	return client.deleteWithRetry(volume)
}

func (client *NVCClient) NetAppVolumeDeleteAll() error {
	if client.EnableCRR {
		volReplications, err := client.NetAppVolumeReplicationGetAll()
		if err != nil {
			return err
		}

		if len(volReplications) > 0 {
			return fmt.Errorf("operation failed: one or more NetAppvolumeReplication exist in the cluster, delete all NetAppvolumeReplication before attempting to delete all NetAppVolume")
		}
	}

	return client.deleteCollectionWithRetry(&quarkv1alpha1.NetAppVolume{})
}

// Error returned by this function should not be overwritten as it has the logic to
// create appropriate concrete error types like StaleVersionError and
// clients (SDE) depend on those concrete error types
func (client *NVCClient) netAppVolumeUpdate(volume *quarkv1alpha1.NetAppVolume) (*quarkv1alpha1.NetAppVolume, error) {
	newVolume, err := client.updateWithRetry(volume)
	if err != nil {
		if strings.Contains(err.Error(), quarkconfig.OldVersionErr) {
			return nil, StaleVersionError{}
		}
		return nil, err
	}
	return newVolume.(*quarkv1alpha1.NetAppVolume), nil
}

// NetAppVolumeOnline returns "true" when the Quark pod is ready to serve data.
func (client *NVCClient) NetAppVolumeOnline(volume *quarkv1alpha1.NetAppVolume) bool {
	for _, cond := range volume.Status.Conditions {
		if cond.Type == quarkv1alpha1.NetAppVolumeOnline && cond.Status == corev1.ConditionTrue {
			return true
		}
	}
	return false
}

// NetAppVolumeResizeAbort aborts the current Vol-Resize operation
func (client *NVCClient) NetAppVolumeResizeAbort(volumeUUID string) (volume *quarkv1alpha1.NetAppVolume, err error) {
	volume, err = client.NetAppVolumeGet(volumeUUID)
	if err != nil {
		return nil, err
	}
	// Reset the new size in spec. The volume error conditions can then be cleared by Podrick.
	volume.Spec.Size = volume.Status.Size
	volume, err = client.netAppVolumeUpdate(volume)
	if err != nil {
		return nil, err
	}
	return volume, nil
}

// NetAppVolumeUsage returns the usage statistics of a given volume.
func (client *NVCClient) NetAppVolumeUsage(volume *quarkv1alpha1.NetAppVolume) (*NetAppVolumeUsage, error) {
	usage := &NetAppVolumeUsage{}
	volUsage, err := client.quarkClient.QuarkV1alpha1().NetAppVolumeUsages(client.nvcNamespace).Get(
		context.TODO(), volume.Name, metav1.GetOptions{})
	if err != nil {
		return nil, err
	}
	usage.LogicalUsed = volUsage.Status.LogicalUsed
	usage.PhysicalUsed = volUsage.Status.PhysicalUsed
	usage.AllocatedSize = volUsage.Status.AllocatedSize
	usage.SnapshotUsed = volUsage.Status.SnapshotUsed
	return usage, nil
}

// NetAppVolumeResized returns "true" when the data PVCs are correctly resized.
func (client *NVCClient) NetAppVolumeResized(volume *quarkv1alpha1.NetAppVolume) bool {
	return volume.Status.Size.Cmp(volume.Spec.Size) == 0
}

// NetAppVolumePermissionUpdated returns "true" when the UnixPermission is correctly updated.
func (client *NVCClient) NetAppVolumePermissionUpdated(volume *quarkv1alpha1.NetAppVolume) bool {
	return (volume.Status.UnixPermission == volume.Spec.UnixPermission)
}

// NetAppVolExportPolicyUpdated returns "true" when the ExportPolicy is applied to the volume.
func (client *NVCClient) NetAppVolExportPolicyUpdated(volume *quarkv1alpha1.NetAppVolume) bool {
	return (volume.Status.ExportPolicy == volume.Spec.ExportPolicy)
}

// NetAppVolumeExportPolicyUpdated returns "true" when the ExportPolicy rule is modified.
func (client *NVCClient) NetAppVolumeExportPolicyUpdated(exportPolicy *quarkv1alpha1.NetAppExportPolicy) bool {
	return reflect.DeepEqual(exportPolicy.Spec.Rules, exportPolicy.Status.Rules)
}

// NetAppVolSnapPolicyUpdated returns "true" when the SnapshotPolicy is applied to the volume.
func (client *NVCClient) NetAppVolSnapPolicyUpdated(volume *quarkv1alpha1.NetAppVolume) bool {
	return (volume.Status.SnapshotPolicy == volume.Spec.SnapshotPolicy)
}

// NetAppVolumeSnapshotPolicyUpdated returns "true" when the SnapshotPolicy rule is modified.
func (client *NVCClient) NetAppVolumeSnapshotPolicyUpdated(snapshotPolicy *quarkv1alpha1.NetAppSnapshotPolicy) bool {
	return (reflect.DeepEqual(snapshotPolicy.Spec.Schedules, snapshotPolicy.Status.Schedules) &&
		snapshotPolicy.Spec.Enabled == snapshotPolicy.Status.Enabled)
}

// NetAppVolNoSnapOptionUpdated returns "true" when the NoSnapDir is applied to the volume.
func (client *NVCClient) NetAppVolNoSnapDirOptionUpdated(volume *quarkv1alpha1.NetAppVolume) bool {
	return (volume.Status.NoSnapDir == volume.Spec.NoSnapDir)
}

// NetAppVolSMBSharePropertiesUpdated returns "true" when SMB shareProperties are applied to the volume.
func (client *NVCClient) NetAppVolSMBSharePropertiesUpdated(volume *quarkv1alpha1.NetAppVolume) bool {
	return util.IsEqualSMBShareProperties(volume)
}

// NetAppVolSMBDNSUpdated returns "true" when SMB DNSIPs are applied to the volume.
func (client *NVCClient) NetAppVolSMBDNSUpdated(volume *quarkv1alpha1.NetAppVolume) bool {
	return util.IsEqualStrings(volume.Spec.SMBConfiguration.DNSIPs, volume.Status.SMBConfiguration.DNSIPs)
}

// NetAppVolSMBPasswordResetDone returns "true" when machine-account password reset is successful.
func (client *NVCClient) NetAppVolSMBPasswordResetDone(volume *quarkv1alpha1.NetAppVolume) bool {
	return (volume.Status.SMBPasswordResetCount == volume.Spec.SMBPasswordResetCount)
}

// NetAppVolSMBEncryptedADCommunicationUpdated returns "true" when EncryptedADCommunication updated for the volume.
func (client *NVCClient) NetAppVolSMBEncryptedADCommunicationUpdated(volume *quarkv1alpha1.NetAppVolume) bool {
	return (volume.Status.SMBConfiguration.EncryptedADCommunication == volume.Spec.SMBConfiguration.EncryptedADCommunication)
}

// NetAppVolumeUpgraded returns "true" when Volume version upgrade is successful.
func (client *NVCClient) NetAppVolumeUpgraded(volume *quarkv1alpha1.NetAppVolume) bool {
	conditions := volume.Status.Conditions
	for _, cond := range conditions {
		if cond.Type == quarkv1alpha1.NetAppVolumeUpgradeDMAPHaltSucceeded || cond.Type == quarkv1alpha1.NetAppVolumeUpgradeDMAPHaltRequested || cond.Type == quarkv1alpha1.NetAppVolumeUpgradeFailed {
			// return false if another upgrade is in progress
			return false
		}
	}
	return (volume.Status.QuarkVersion == volume.Spec.QuarkVersion)

}

// NetAppVolumeCloudSnapshotRestoreCondition returns the state of a cloud snapshot restore operation.
func (client *NVCClient) NetAppVolumeCloudSnapshotRestoreCondition(volume *quarkv1alpha1.NetAppVolume) *quarkv1alpha1.NetAppVolumeCondition {
	for _, condition := range volume.Status.Conditions {
		if condition.Type == quarkv1alpha1.NetAppVolumeCloudSnapshotPreparing ||
			condition.Type == quarkv1alpha1.NetAppVolumeCloudSnapshotRestoring ||
			condition.Type == quarkv1alpha1.NetAppVolumeCloudSnapshotRestored ||
			condition.Type == quarkv1alpha1.NetAppVolumeCloudSnapshotRestoreError {

			return condition.DeepCopy()
		}
	}
	return nil
}

func (client *NVCClient) ValidateExportPolicyCRWithFlag(policy *quarkv1alpha1.NetAppExportPolicy) error {
	for _, rl := range policy.Spec.Rules {
		protocols := strings.Split(rl.Protocol, ",")
		if util.StringSetContains(protocols, quarkconfig.DefaultNFS4Protocol) && !client.EnableNfs41 {
			return fmt.Errorf("Invalid Value of protocol in export policy %s, NFS4.1 not enabled", policy.Name)
		}
	}
	return nil
}

func (client *NVCClient) NetAppExportPolicyCreate(policy *quarkv1alpha1.NetAppExportPolicy) (*quarkv1alpha1.NetAppExportPolicy, error) {
	err := client.ValidateExportPolicyCRWithFlag(policy)
	if err != nil {
		return nil, err
	}

	result, err := client.createWithRetry(policy)
	if err != nil {
		return nil, err
	}
	return result.(*quarkv1alpha1.NetAppExportPolicy), nil
}

func (client *NVCClient) NetAppExportPolicyGet(policyName string) (*quarkv1alpha1.NetAppExportPolicy, error) {
	policy := &quarkv1alpha1.NetAppExportPolicy{
		ObjectMeta: metav1.ObjectMeta{
			Name: policyName,
		},
	}
	result, err := client.getWithRetry(policy)
	if err != nil {
		return nil, err
	}
	return result.(*quarkv1alpha1.NetAppExportPolicy), nil
}

func (client *NVCClient) NetAppExportPolicyDelete(policyName string) error {
	policy := &quarkv1alpha1.NetAppExportPolicy{
		ObjectMeta: metav1.ObjectMeta{
			Name: policyName,
		},
	}
	return client.deleteWithRetry(policy)
}

func (client *NVCClient) NetAppExportPolicyModify(policy *quarkv1alpha1.NetAppExportPolicy) (*quarkv1alpha1.NetAppExportPolicy, error) {
	err := client.ValidateExportPolicyCRWithFlag(policy)
	if err != nil {
		return nil, err
	}
	result, err := client.updateWithRetry(policy)
	if err != nil {
		return nil, err
	}
	return result.(*quarkv1alpha1.NetAppExportPolicy), nil
}

func (client *NVCClient) NetAppSnapshotPolicyCreate(policy *quarkv1alpha1.NetAppSnapshotPolicy) (*quarkv1alpha1.NetAppSnapshotPolicy, error) {
	result, err := client.createWithRetry(policy)
	if err != nil {
		return nil, err
	}
	return result.(*quarkv1alpha1.NetAppSnapshotPolicy), nil
}

func (client *NVCClient) NetAppSnapshotPolicyGet(policyName string) (*quarkv1alpha1.NetAppSnapshotPolicy, error) {
	policy := &quarkv1alpha1.NetAppSnapshotPolicy{
		ObjectMeta: metav1.ObjectMeta{
			Name: policyName,
		},
	}
	result, err := client.getWithRetry(policy)
	if err != nil {
		return nil, err
	}
	return result.(*quarkv1alpha1.NetAppSnapshotPolicy), nil
}

func (client *NVCClient) NetAppSnapshotPolicyDelete(policyName string) error {
	policy := &quarkv1alpha1.NetAppSnapshotPolicy{
		ObjectMeta: metav1.ObjectMeta{
			Name: policyName,
		},
	}
	return client.deleteWithRetry(policy)
}

func (client *NVCClient) NetAppSnapshotPolicyUpdate(policy *quarkv1alpha1.NetAppSnapshotPolicy) (*quarkv1alpha1.NetAppSnapshotPolicy, error) {
	result, err := client.updateWithRetry(policy)
	if err != nil {
		return nil, err
	}
	return result.(*quarkv1alpha1.NetAppSnapshotPolicy), nil
}

func (client *NVCClient) NetAppVolumeSnapshotsListUpdate(snapList *quarkv1alpha1.NetAppVolumeSnapshotsList) (*quarkv1alpha1.NetAppVolumeSnapshotsList, error) {
	result, err := client.updateWithRetry(snapList)
	if err != nil {
		return nil, err
	}
	return result.(*quarkv1alpha1.NetAppVolumeSnapshotsList), nil
}

func (client *NVCClient) NetAppVolumeSnapshotCreate(volumeSnapshot *quarkv1alpha1.NetAppVolumeSnapshot) error {
	metaLabels := volumeSnapshot.ObjectMeta.GetLabels()
	volumeUUID := metaLabels[quarkconfig.VolumeLabel]
	volume, err := client.quarkClient.QuarkV1alpha1().NetAppVolumes(client.nvcNamespace).Get(
		context.TODO(), volumeUUID, metav1.GetOptions{})
	if err != nil {
		return err
	}

	// if snapshot with same display name already exists in snapshotList CR, do not allow to create snapshot
	snapList, err := client.NetAppVolumeSnapshotsListGet(volumeUUID)
	if err != nil {
		return fmt.Errorf("failed to retrieve snapshot list info for volume %s", volumeUUID)
	}

	for _, snap := range snapList.Status.Snapshots {
		if volumeSnapshot.Spec.DisplayName == snap.Name {
			return fmt.Errorf("failed to create snapshot with same name as existing snapshot %s", snap.Name)
		}
	}

	volumeSnapshot.ObjectMeta.OwnerReferences = []metav1.OwnerReference{
		*metav1.NewControllerRef(volume, schema.GroupVersionKind{
			Group:   quarkv1alpha1.SchemeGroupVersion.Group,
			Version: quarkv1alpha1.SchemeGroupVersion.Version,
			Kind:    "NetAppVolume",
		}),
	}

	_, err = client.createWithRetry(volumeSnapshot)
	if err != nil {
		return err
	}
	return nil
}

// NetAppVolumeSnapshotCreateAbort aborts the current snapshot create operation
func (client *NVCClient) NetAppVolumeSnapshotCreateAbort(volumeUUID, snapUUID string) error {

	// Check if the snapshot is ready
	snapReady, err := client.netAppVolumeSnapshotStatusReady(snapUUID, volumeUUID)
	if snapReady {
		return fmt.Errorf("unable to abort snapshot create %s  for volume %s , snapshot already created", snapUUID, volumeUUID)
	} else if err != nil {
		return fmt.Errorf("unable to check snapshot status for snapshot %s : %v", snapUUID, err)
	}

	// snapshot not ready we can proceed with abort , delete snapshot CR
	snap := &quarkv1alpha1.NetAppVolumeSnapshot{
		ObjectMeta: metav1.ObjectMeta{
			Name: snapUUID,
		},
	}
	err = client.deleteWithRetry(snap)
	if err != nil {
		return err
	}

	log.WithFields(log.Fields{
		"volume":   volumeUUID,
		"snapshot": snapUUID,
	}).Debug("NetAppVolumeSnapshots create abort success.")

	return nil
}

func (client *NVCClient) NetAppVolumeSnapshotDelete(volumeUUID, snapDisplayName string) error {

	// Check if snapshot is associated with clones, if so return error
	cloneVolumeList, err := client.getAssociatedCloneVolumeList(volumeUUID, snapDisplayName)
	if err != nil {
		return fmt.Errorf("failed to retrieve clone volume list for given snapshot %s and volume %s, error: %v", snapDisplayName, volumeUUID, err)
	}
	if len(cloneVolumeList) > 0 {
		return CloneVolumeSplitScanInProgressError{CloneList: cloneVolumeList}
	}

	// Check if snapshot is the base snapshot of a clone volume
	isBaseSnapshot, err := client.isBaseCloneSnapshot(volumeUUID, snapDisplayName)
	if err != nil {
		return fmt.Errorf("failed to check if snapshot %s is a base clone snapshot, error: %v", snapDisplayName, err)
	}

	if isBaseSnapshot {
		// Given snapshot is the base snapshot of a clone and split scan is still in progress
		// Snapshot cannot be deleted
		return CloneVolumeBaseSnapshotError{}

	}
	snapList, err := client.NetAppVolumeSnapshotsListGet(volumeUUID)
	if err != nil {
		return fmt.Errorf("failed to retrieve snapshot list info for volume %s when deleting snapshot %s, error: %v", volumeUUID, snapDisplayName, err)
	}
	snapListCopy := snapList.DeepCopy()
	foundSnap := false

	for _, snap := range snapListCopy.Status.Snapshots {
		if snap.Name == snapDisplayName {
			foundSnap = true
			for {
				pendingDeleteSnapshots := util.StringSetDifference(snapListCopy.Spec.DeletedSnapshots, snapListCopy.Status.DeletedSnapshots)
				if util.StringSetContains(pendingDeleteSnapshots, snapDisplayName) {
					log.WithFields(log.Fields{
						"volume":       volumeUUID,
						"snapshot":     snapDisplayName,
						"snapshotList": snapListCopy,
					}).Debug("Snapshot delete operation is already pending. Returning success.")
					return nil
				}
				// Update spec.DeletedSnapshots field of snapshotsList CR to delete a snapshot
				// and clear already deleted snapshots from both spec and status.
				snapListCopy.Spec.DeletedSnapshots = append(pendingDeleteSnapshots, snapDisplayName)
				// Trim from status the snapshots that are not part of spec
				// anymore (just in case spec and status get out of sync).
				snapListCopy.Status.DeletedSnapshots = make([]string, 0)

				_, err := client.NetAppVolumeSnapshotsListUpdate(snapListCopy)
				if err != nil {
					if strings.Contains(err.Error(), quarkconfig.OldVersionErr) {
						snapListCopy, err = client.NetAppVolumeSnapshotsListGet(volumeUUID)
						if err != nil {
							return fmt.Errorf("server error deleting snapshot %s:%s", snapDisplayName, err)
						}
						continue
					}
					return fmt.Errorf("server error deleting snapshot %s: %s", snapDisplayName, err)
				}
				break
			}
			return nil
		}
	}
	if !foundSnap {
		return fmt.Errorf("failed to find snapshot %s", snapDisplayName)
	}
	return nil
}

// NetAppVolumeSnapshotDeleteAbort aborts the current snapshot delete operation
// This call is not idempotent , if this is called for an already aborted deletion , it will error out
func (client *NVCClient) NetAppVolumeSnapshotDeleteAbort(volumeUUID, snapDisplayName string) error {
	errPrefix := "could not abort snapshot delete operation"
	snapList, err := client.NetAppVolumeSnapshotsListGet(volumeUUID)
	if err != nil {
		return fmt.Errorf("failed to retrieve snapshot list info for volume %s when aborting Snapshot Delete", volumeUUID)
	}
	snapListCopy := snapList.DeepCopy()

	for _, snap := range snapListCopy.Status.Snapshots {
		if snap.Name == snapDisplayName {
			for {

				// Check if the snapshot has already been deleted
				if util.StringSetContains(snapListCopy.Status.DeletedSnapshots, snapDisplayName) {
					// Snapshot has already been deleted, cannot abort deletion, return error
					return fmt.Errorf(errPrefix+", snapshot %s has already been deleted", snapDisplayName)
				}

				if !util.StringSetContains(snapListCopy.Spec.DeletedSnapshots, snapDisplayName) {
					// This means a previous abort might have succeeded or a snapshot Deletion never started
					// It is difficult to determine the state , so return error
					return fmt.Errorf(errPrefix+", there is no pending delete for snapshot %s", snapDisplayName)
				}

				// Remove snapshot from DeletedSnapshots list in Spec
				snapListCopy.Spec.DeletedSnapshots = util.RemoveStringFromSlice(snapListCopy.Spec.DeletedSnapshots, snapDisplayName)

				_, err := client.NetAppVolumeSnapshotsListUpdate(snapListCopy)
				if err != nil {
					if strings.Contains(err.Error(), quarkconfig.OldVersionErr) {
						snapListCopy, err = client.NetAppVolumeSnapshotsListGet(volumeUUID)
						if err != nil {
							return fmt.Errorf(errPrefix+" for %s server error:%s", snapDisplayName, err)
						}
						continue
					}
					return fmt.Errorf(errPrefix+" for %s server error:%s", snapDisplayName, err)
				}
				break
			}
			return nil
		}
	}

	// If the snapshotname is not found return an error
	return fmt.Errorf("the snapshot deletion to be aborted %s has already been deleted or doesn't exist", snapDisplayName)
}

func (client *NVCClient) NetAppVolumeSnapshotRestore(volumeUUID, snapDisplayName string) error {
	snapList, err := client.NetAppVolumeSnapshotsListGet(volumeUUID)
	if err != nil {
		return fmt.Errorf("failed to retrieve snapshot list info for volume %s when restoring snapshot %s", volumeUUID, snapDisplayName)
	}

	// Check if the snapshot to be restored has any snapshot taken before it still associated to a clone volume
	isSnapshotCloneLocked, err := client.checkSnapshotForRestoreIsCloneLocked(snapList, volumeUUID, snapDisplayName)
	if err != nil {
		return err
	}
	if isSnapshotCloneLocked {
		return RestoreSnapshotIsCloneLockedError{}
	}
	snapListCopy := snapList.DeepCopy()
	for _, snap := range snapListCopy.Status.Snapshots {
		// Check if the snapshot to be restored exists
		if snap.Name == snapDisplayName {
			for {
				// if spec and status do not match and error condition exists in snapshotsList conditions,
				// send the message to abort restore operation
				// if there is no SnapshotRestoreError condition present, indicate that the restore is pending.
				if snapListCopy.Spec.RestoredSnapshot != snapListCopy.Status.RestoredSnapshot {
					for _, cond := range snapList.Status.Conditions {
						if cond.Type == quarkv1alpha1.SnapshotRestoreError && cond.Status == corev1.ConditionTrue {
							log.WithFields(log.Fields{
								"volume":       volumeUUID,
								"snapshot":     snapDisplayName,
								"snapshotList": snapListCopy,
							}).Error("Fatal error occurred during SnapRestore.")
							return fmt.Errorf("fatal error occurred during SnapRestore of %s. Manually abort this operation to trigger another SnapRestore", snapDisplayName)
						}
					}
					if snapListCopy.Spec.RestoredSnapshot != snapDisplayName {
						log.WithFields(log.Fields{
							"volume":       volumeUUID,
							"snapshot":     snapDisplayName,
							"snapshotList": snapListCopy,
						}).Debug("Another snapshot restore operation is already pending.")
						return fmt.Errorf("another snapshot restore operation is already pending")
					}
					log.WithFields(log.Fields{
						"volume":       volumeUUID,
						"snapshot":     snapDisplayName,
						"snapshotList": snapListCopy,
					}).Debug("Snapshot restore operation is pending, returning success")
					return nil
				}
				// clear the status of already restored snapshot and update the spec with new
				// snapshot to be restored
				snapListCopy.Spec.RestoredSnapshot = snapDisplayName
				snapListCopy.Status.RestoredSnapshot = ""
				_, err := client.NetAppVolumeSnapshotsListUpdate(snapListCopy)
				if err != nil {

					if strings.Contains(err.Error(), quarkconfig.OldVersionErr) {
						snapListCopy, err = client.NetAppVolumeSnapshotsListGet(volumeUUID)
						if err != nil {
							return fmt.Errorf("failed to get snapshots list :%s", err)
						}
						continue
					}
					return fmt.Errorf("server error restoring snapshot %s: %s", snapDisplayName, err)
				}
				break
			}
			return nil
		}
	}
	// Return error, if snapshot not found for restore
	return fmt.Errorf("snapshot not found for restore: %s", snapDisplayName)
}

// NetAppVolumeSnapshotRestoreAbort aborts the current snapRestore operation
func (client *NVCClient) NetAppVolumeSnapshotRestoreAbort(volumeUUID string) error {
	snapList, err := client.NetAppVolumeSnapshotsListGet(volumeUUID)
	if err != nil {
		return fmt.Errorf("failed to retrieve snapshot list info for volume %s when aborting Snap Restore", volumeUUID)
	}
	// Updating conditions by removing the SnapshotRestoreError condition
	var newConditions []quarkv1alpha1.NetAppVolumeSnapshotsListCondition
	for _, cond := range snapList.Status.Conditions {
		if cond.Type != quarkv1alpha1.SnapshotRestoreError {
			newConditions = append(newConditions, cond)
		}
	}
	snapListCopy := snapList.DeepCopy()
	snapListCopy.Spec.RestoredSnapshot = ""
	snapListCopy.Status.RestoredSnapshot = ""
	snapListCopy.Status.Conditions = newConditions
	_, err = client.NetAppVolumeSnapshotsListUpdate(snapListCopy)
	if err != nil {
		return fmt.Errorf("server error aborting NetAppVolumeSnapshotRestore: %s", err)
	}
	log.WithFields(log.Fields{
		"volume": volumeUUID,
	}).Info("NetAppVolumeSnapshotRestore has been aborted.")

	return nil
}

// Checks if any snapshot taken after the mentioned snapshot has a clone volume that still has split scan in progress
func (client *NVCClient) checkSnapshotForRestoreIsCloneLocked(snapList *quarkv1alpha1.NetAppVolumeSnapshotsList, volumeUUID string, snapName string) (bool, error) {

	// Get the list of snapshots
	snapshotsList := snapList.Status.DeepCopy().Snapshots

	// Sort snapList in descending order by timestamp (latest snapshot first and oldest snapshot last)
	sort.SliceStable(snapshotsList, func(i int, j int) bool {
		return !snapshotsList[i].CreationTime.Before(&snapshotsList[j].CreationTime)
	})

	// Map of snapshot names that were taken after the snapshot being restored
	snapshotNameMap := make(map[string]bool)

	// Find the index at which the restore snapshot is present
	snapindex := -1
	for i := 0; i < len(snapshotsList); i++ {
		// Break if we find the snapshot name of restore
		if snapshotsList[i].Name == snapName {
			snapindex = i
			break
		}
		snapshotNameMap[snapshotsList[i].Name] = true
	}

	if snapindex == -1 {
		// If we encounter an error we don't want to move forward with the snapshot restore, hence return true
		return true, fmt.Errorf("snapshot not found for restore: %s", snapName)
	}

	if len(snapshotNameMap) == 0 {
		// There are no snapshots that are after this, return false
		return false, nil
	}

	// Get a list of clone volumes associated with the volumeUUID being restored
	labelSelector := quarkconfig.VolumeCloneParentUUUIDLabel + "=" + volumeUUID
	listOptions := metav1.ListOptions{
		LabelSelector: labelSelector,
	}
	result, err := client.listWithRetry(&quarkv1alpha1.NetAppVolumeList{}, listOptions)
	if err != nil {
		// If we encounter an error we don't want to move forward with the snapshot restore, hence return true
		return true, fmt.Errorf("failed to check if snapshot is locked for clone err: %v", err)
	}
	volumes := result.(*quarkv1alpha1.NetAppVolumeList)

	if len(volumes.Items) == 0 {
		// There are no clone volumes associated with this volume, we can proceed with restore
		return false, nil
	}

	// Check if any of the snapshots associated with the clone volumes is in snapshotNameMap
	// If any of the snapshot is found return true
	for _, volume := range volumes.Items {
		cloneSnapshotName := volume.Spec.CloneParentSnapshot
		if _, ok := snapshotNameMap[cloneSnapshotName]; ok {
			// Snapshot is still associated with a clone, return true
			return true, nil
		}
	}

	return false, nil
}

// NetAppVolumeSnapshotRestored returns "true" when the snapshot is restored.
func (client *NVCClient) NetAppVolumeSnapshotRestored(volumeUUID, snapDisplayName string) bool {
	snapList, err := client.NetAppVolumeSnapshotsListGet(volumeUUID)
	if err != nil {
		return false
	}
	// If spec and status match and snapshot is present in the RestoredSnapshot of snapshoList CR's status, it has been restored, hence return true
	if snapList.Spec.RestoredSnapshot == snapList.Status.RestoredSnapshot && snapList.Status.RestoredSnapshot == snapDisplayName {
		log.WithFields(log.Fields{
			"volume":   volumeUUID,
			"snapshot": snapDisplayName,
		}).Debug("NetAppVolumeSnapshot has been restored.")
		return true
	}
	log.WithFields(log.Fields{
		"volume":   volumeUUID,
		"snapshot": snapDisplayName,
	}).Debug("NetAppVolumeSnapshot is not restored.")
	return false
}

// NetAppVolumeSnapshotDeleted returns "true" when the snapshot is deleted.
func (client *NVCClient) NetAppVolumeSnapshotDeleted(volumeUUID, snapDisplayName string) bool {
	_, err := client.NetAppVolumeGet(volumeUUID)
	if err != nil {
		// if volume is not found, then there is no snapshot
		return true
	}
	snapList, err := client.NetAppVolumeSnapshotsListGet(volumeUUID)
	if err != nil {
		return false
	}
	// If snapshot is not found in the list, it is deleted
	for _, snap := range snapList.Status.Snapshots {
		if snap.Name == snapDisplayName {
			log.WithFields(log.Fields{
				"volume":   volumeUUID,
				"snapshot": snapDisplayName,
			}).Debug("NetAppVolumeSnapshot isn't deleted yet.")
			return false
		}
	}
	log.WithFields(log.Fields{
		"volume":   volumeUUID,
		"snapshot": snapDisplayName,
	}).Debug("NetAppVolumeSnapshot is deleted.")
	return true
}

type NetAppVolumeSnapshotInfo struct {
	Name       string
	Size       string
	VolumeUUID string
	// Creation time of snapshot with location set to UTC.
	CreationTime time.Time
}

func (client *NVCClient) NetAppVolumeSnapshotsListGetByVolume(volumeUUID string) ([]NetAppVolumeSnapshotInfo, error) {
	var snapList []NetAppVolumeSnapshotInfo

	// validate UUID
	_, err := uuid.FromString(volumeUUID)
	if err != nil {
		return nil, fmt.Errorf("invalid value for UUID")
	}
	_, err = client.NetAppVolumeGet(volumeUUID)
	if err != nil {
		if errors.IsNotFound(err) {
			return nil, fmt.Errorf("netappVolume %s was not found", volumeUUID)
		}
	}
	ntapVolSnapList, err := client.NetAppVolumeSnapshotsListGet(volumeUUID)
	if err != nil {
		return nil, fmt.Errorf("server error listing snapshots")
	}

	for _, snap := range ntapVolSnapList.Status.Snapshots {
		snapList = append(snapList, NetAppVolumeSnapshotInfo{
			Name:         snap.Name,
			Size:         snap.Size.String(),
			VolumeUUID:   volumeUUID,
			CreationTime: snap.CreationTime.UTC()})
	}
	return snapList, nil
}

func (client *NVCClient) NetAppVolumeSnapshotGetAll() ([]NetAppVolumeSnapshotInfo, error) {
	var snapList []NetAppVolumeSnapshotInfo
	ntapVols, err := client.NetAppVolumeGetAll()
	if err != nil {
		return nil, err
	}
	for _, volume := range ntapVols {
		ntapVolSnapList, err := client.NetAppVolumeSnapshotsListGet(volume.Name)
		if err != nil {
			return nil, fmt.Errorf("server error listing snapshots")
		}

		for _, snap := range ntapVolSnapList.Status.Snapshots {
			snapList = append(snapList, NetAppVolumeSnapshotInfo{
				Name:       snap.Name,
				Size:       snap.Size.String(),
				VolumeUUID: volume.Name})
		}

	}
	return snapList, nil
}

// NetAppVolumeSnapshotReady returns "true" when the volume snapshot is ready.
func (client *NVCClient) NetAppVolumeSnapshotReady(volumeUUID, snapDisplayName string) bool {
	// Checking snapshot CR in case it is not scheduled snapshot and snap list haven't been updated yet
	// Snap list check could race with the event which is triggered by Snapshot CR update
	snapReady, err := client.pitSnapshotReadyByName(snapDisplayName, volumeUUID)
	if err != nil && !errors.IsNotFound(err) {
		// Error is coming from API server
		// Returning to a caller to retry
		log.WithFields(log.Fields{
			"err": err.Error(),
		}).Debug("Error while calling kube api server")
		return false
	} else if err == nil {
		return snapReady
	}

	// Checking snapshot list in case of a scheduled snapshots
	ntapVolSnapList, err := client.NetAppVolumeSnapshotsListGet(volumeUUID)
	if err != nil {
		return false
	}
	log.WithFields(log.Fields{
		"volume":    volumeUUID,
		"snapshots": ntapVolSnapList,
	}).Debug("NetAppVolumeSnapshotsList get success.")

	// If snapshot found in the list, it is ready
	for _, snap := range ntapVolSnapList.Status.Snapshots {
		if snap.Name == snapDisplayName {
			return true
		}
	}
	return false
}

// NetAppVolumeSnapshotsListUpdated returns "true" when the volume snapshot is found in the list.
func (client *NVCClient) NetAppVolumeSnapshotsListUpdated(volumeUUID, snapDisplayName string) bool {
	ntapVolSnapList, err := client.quarkClient.QuarkV1alpha1().
		NetAppVolumeSnapshotsLists(client.nvcNamespace).
		Get(context.TODO(), volumeUUID, metav1.GetOptions{})
	if err != nil {
		return false
	}
	log.WithFields(log.Fields{
		"volume":    volumeUUID,
		"snapshots": ntapVolSnapList,
	}).Debug("NetAppVolumeSnapshotsList get success.")

	// If snapshot found in the list, it is ready
	for _, snap := range ntapVolSnapList.Status.Snapshots {
		if snap.Name == snapDisplayName {
			return true
		}
	}
	return false
}

func (client *NVCClient) pitSnapshotReadyByName(snapDisplayName string, volumeUUID string) (bool, error) {
	selectorString := quarkconfig.VolumeLabel + "=" + volumeUUID
	listOptions := metav1.ListOptions{
		LabelSelector: selectorString,
	}
	result, err := client.listWithRetry(&quarkv1alpha1.NetAppVolumeSnapshotList{}, listOptions)
	if err != nil {
		return false, err
	}
	ntapVolSnapshots := result.(*quarkv1alpha1.NetAppVolumeSnapshotList)

	for _, snap := range ntapVolSnapshots.Items {
		if snap.Spec.DisplayName == snapDisplayName {
			if len(snap.Status.Conditions) != 0 && IsSnapshotReady(snap.Status.Conditions) {
				return true, nil
			} else {
				return false, nil
			}
		}
	}

	return false, errors.NewNotFound(schema.GroupResource{}, snapDisplayName)
}

// NetAppVolumeSnapshotStatusReady returns "true" when the volume snapshot CR status is ready.
func (client *NVCClient) netAppVolumeSnapshotStatusReady(snapUUID, volumeUUID string) (bool, error) {
	ntapVolSnapshot, err := client.NetAppVolumeSnapshotGet(snapUUID)
	if err != nil {
		return false, err
	}

	log.WithFields(log.Fields{
		"volume":    volumeUUID,
		"snapshots": ntapVolSnapshot,
	}).Debug("NetAppVolumeSnapshots get success.")

	if len(ntapVolSnapshot.Status.Conditions) != 0 && IsSnapshotReady(ntapVolSnapshot.Status.Conditions) {
		//Snapshot is ready
		return true, nil
	}

	return false, nil
}

// NetAppVolumeDelete is used to signal the deletion of a NetAppVolume.
type NetAppVolumeDelete struct {
	Volume *quarkv1alpha1.NetAppVolume
}

// NetAppVolumeUpdate is used to signal the most recent state of a NetAppVolume and the
// most recent changes.
type NetAppVolumeUpdate struct {
	Volume            *quarkv1alpha1.NetAppVolume
	Status            *NetAppVolumeUpdateStatus
	UpdatedConditions []quarkv1alpha1.NetAppVolumeCondition
	Error             error
}

// NetAppQuarkUpdate is used to signal the most recent state of a Quark CR and the
// most recent changes.
type NetAppQuarkUpdate struct {
	Quark  *quarkv1alpha1.Quark
	Status NetAppQuarkUpdateStatus
	Error  error
}

// NetAppQuarkDelete is used to signal the most recent state of a Quark CR and the
// most recent changes.
type NetAppQuarkDelete struct {
	Quark *quarkv1alpha1.Quark
}

// QuarkOperatorUpdateStatus captures the status of any pending updates.
type NetAppQuarkUpdateStatus struct {
	HighAvailabilityReady        bool
	ExternalVolumeAccessReady    bool
	CloudsnapshotCapabilityReady bool
	VolumeControllerReady        bool
	ServiceMeshControllerReady   bool
	OperatorReady                bool
	IPDReady                     bool
	EtcdClusterReady             bool
	NodevolReady                 bool
	NodeMonitorReady             bool
	OtelCollectorReady           bool
	Version                      string
}

type NetAppQuarkThirdPartyComponentStatus struct {
	KubeStateMetricsReady bool
	PrometheusServerReady bool
	ADCReady              bool
}

// NetAppVolumeUpdateStatus captures the status of any pending updates.
type NetAppVolumeUpdateStatus struct {
	SizeUpdated                        bool
	ExportPolicyUpdated                bool
	SnapshotPolicyUpdated              bool
	SMBSharePropertiesUpdated          bool
	SMBDNSUpdated                      bool
	SMBEncryptedADCommunicationUpdated bool
	SMBPasswordReset                   bool
	QuarkVersionUpdated                bool
	NoSnapDirOptionUpdated             bool
	UnixPermissionUpdated              bool
}

// NetAppVolumeSnapshotDelete is used to signal the snapshot delete completion.
type NetAppVolumeSnapshotDelete struct {
	VolumeUUID       string
	DeletedSnapshots []string
}

// NetAppVolumeSnapshotUpdate is used to signal the most recent changes to NetAppVolumeSnapshots.
type NetAppVolumeSnapshotUpdate struct {
	Snapshot          *quarkv1alpha1.NetAppVolumeSnapshot
	UpdatedConditions []quarkv1alpha1.NetAppVolumeSnapshotCondition
}

// NetAppVolumeSnapshotRestore is used to signal the snapshot restore operation.
type NetAppVolumeSnapshotRestore struct {
	VolumeUUID       string
	ErrorCondition   string
	RestoredSnapshot bool
}

// NetAppVolumeSnapshotsListUpdate is used to signal the most recent changes to NetAppVolumeSnapshotsList.
type NetAppVolumeSnapshotsListUpdate struct {
	VolumeUUID           string
	SnapshotsListUpdated bool
	SnapshotNames        []string
}

// NetAppVolumeCloudSnapshotDelete is used to signal the deletion of a NetAppVolumeCloudSnapshot.
type NetAppVolumeCloudSnapshotDelete struct {
	CloudSnapshot *quarkv1alpha1.NetAppVolumeCloudSnapshot
}

// NetAppVolumeCloudSnapshotUpdate is used to signal the most recent state of a NetAppVolumeCloudSnapshot and the
// most recent changes.
type NetAppVolumeCloudSnapshotUpdate struct {
	CloudSnapshot *quarkv1alpha1.NetAppVolumeCloudSnapshot
	Status        NetAppVolumeCloudSnapshotInfo
}

// NetAppVolumeExportPolicyUpdate is used to signal the most recent state of a NetAppVolumeExportPolicy and the
// most recent changes.
type NetAppVolumeExportPolicyUpdate struct {
	ExportPolicy *quarkv1alpha1.NetAppExportPolicy
	Rules        []quarkv1alpha1.ExportRule
}

// NetAppVolumeSnapshotPolicyUpdate is used to signal the most recent state of a NetAppVolumeExportPolicy and the
// most recent changes.
type NetAppVolumeSnapshotPolicyUpdate struct {
	SnapshotPolicy *quarkv1alpha1.NetAppSnapshotPolicy
	Enabled        bool
	Schedules      []quarkv1alpha1.CronSchedule
}

// NetAppVolumeUpdated returns "true" when the actual state has converged to the
// desired state. If not, it will return an object describing what actions are pending.
func (client *NVCClient) NetAppVolumeUpdated(volume *quarkv1alpha1.NetAppVolume) (bool, *NetAppVolumeUpdateStatus, error) {
	statusConverged := true

	status := &NetAppVolumeUpdateStatus{NoSnapDirOptionUpdated: true, SizeUpdated: true, ExportPolicyUpdated: true, SnapshotPolicyUpdated: true, SMBSharePropertiesUpdated: true, SMBDNSUpdated: true, SMBEncryptedADCommunicationUpdated: true, SMBPasswordReset: true, QuarkVersionUpdated: true, UnixPermissionUpdated: true}
	// Check Unix Permission for NFS volumes
	if !client.NetAppVolumePermissionUpdated(volume) {
		status.UnixPermissionUpdated = false
		statusConverged = false
	}
	// Check size
	if !client.NetAppVolumeResized(volume) {
		status.SizeUpdated = false
		statusConverged = false
	}
	if !client.NetAppVolExportPolicyUpdated(volume) {
		status.ExportPolicyUpdated = false
		statusConverged = false
	}
	if !client.NetAppVolSnapPolicyUpdated(volume) {
		status.SnapshotPolicyUpdated = false
		statusConverged = false
	}
	if !client.NetAppVolSMBSharePropertiesUpdated(volume) {
		status.SMBSharePropertiesUpdated = false
		statusConverged = false
	}
	if !client.NetAppVolSMBDNSUpdated(volume) {
		status.SMBDNSUpdated = false
		statusConverged = false
	}
	if !client.NetAppVolSMBEncryptedADCommunicationUpdated(volume) {
		status.SMBEncryptedADCommunicationUpdated = false
		statusConverged = false
	}
	if !client.NetAppVolSMBPasswordResetDone(volume) {
		status.SMBPasswordReset = false
		statusConverged = false
	}
	if !client.NetAppVolumeUpgraded(volume) {
		status.QuarkVersionUpdated = false
		statusConverged = false
	}
	if !client.NetAppVolNoSnapDirOptionUpdated(volume) {
		status.NoSnapDirOptionUpdated = false
		statusConverged = false
	}
	// Return the status of any potentially pending updates.
	if statusConverged {
		return true, status, nil
	} else {
		return false, status, nil
	}
}

// Start the NVC Client controller to receive updates from NVC.
// stopCh: Channel to terminate the controller.
// volUpdateChan: Channel to receive updates to NetAppVolumes.
// volDeleteChan: Channel to receive deletions of NetAppVolumes.
// snapUpdateChan: Channel to receive updates to NetAppVolumeSnapshots.
// snapDeleteChan: Channel to receive deletions of NetAppVolumeSnapshots.
// quarkUpdateChan: Channel to receive updates to Quark
// quarkDeleteChan: Channel to receive deletions of Quark
// clusterPeeringUpdateChan: Channel to receive updates of NetAppClusterPeering.
// clusterPeeringDeleteChan: Channel to receive deletes of NetAppClusterPeering.
// volumeReplicationUpdateChan: Channel to receive updates of NetAppVolumeReplication.
// volumeReplicationDeleteChan: Channel to receive deletes of NetAppVolumeReplication.
// objectStoreUpdateChan: Channel to receive updates of NetAppObjectStore
// objectStoreDeleteChan: Channel to receive deletes of NetAppObjectStore
// volumeCloudBackupListUpdateChan: Channel to receive updates of NetAppVolumeCloudBackupList
// threadiness: Number of controller threads
func (client *NVCClient) StartController(stopCh <-chan struct{}, threadiness int) (*NVCClientEventHandler, error) {

	volUpdateChan := make(chan *NetAppVolumeUpdate, EventHandlerQueueSize)
	volDeleteChan := make(chan *NetAppVolumeDelete, EventHandlerQueueSize)
	spUpdateChan := make(chan *NetAppStoragePoolUpdate, EventHandlerQueueSize)
	spDeleteChan := make(chan *NetAppStoragePoolDelete, EventHandlerQueueSize)
	snapDeleteChan := make(chan *NetAppVolumeSnapshotDelete, EventHandlerQueueSize)
	snapRestoreChan := make(chan *NetAppVolumeSnapshotRestore, EventHandlerQueueSize)
	snapUpdateChan := make(chan *NetAppVolumeSnapshotUpdate, EventHandlerQueueSize)
	snapListChan := make(chan *NetAppVolumeSnapshotsListUpdate, EventHandlerQueueSize)
	cloudSnapUpdateChan := make(chan *NetAppVolumeCloudSnapshotUpdate, EventHandlerQueueSize)
	cloudSnapDeleteChan := make(chan *NetAppVolumeCloudSnapshotDelete, EventHandlerQueueSize)
	exportPolUpdateChan := make(chan *NetAppVolumeExportPolicyUpdate, EventHandlerQueueSize)
	snapshotPolUpdateChan := make(chan *NetAppVolumeSnapshotPolicyUpdate, EventHandlerQueueSize)
	quarkUpdateChan := make(chan *NetAppQuarkUpdate, EventHandlerQueueSize)
	quarkDeleteChan := make(chan *NetAppQuarkDelete, EventHandlerQueueSize)
	clusterPeeringUpdateChan := make(chan *NetAppClusterPeeringUpdate, EventHandlerQueueSize)
	clusterPeeringDeleteChan := make(chan *NetAppClusterPeeringDelete, EventHandlerQueueSize)
	volumeReplicationUpdateChan := make(chan *NetAppVolumeReplicationUpdate, EventHandlerQueueSize)
	volumeReplicationDeleteChan := make(chan *NetAppVolumeReplicationDelete, EventHandlerQueueSize)
	objectStoreUpdateChan := make(chan *NetAppObjectStoreUpdate, EventHandlerQueueSize)
	objectStoreDeleteChan := make(chan *NetAppObjectStoreDelete, EventHandlerQueueSize)
	volumeCloudBackupListUpdateChan := make(chan *NetAppVolumeCloudBackupListUpdate, EventHandlerQueueSize)

	eventHandler := newNVCClientEventHandler(client, volUpdateChan, volDeleteChan, spUpdateChan, spDeleteChan,
		snapUpdateChan, snapDeleteChan, snapRestoreChan, snapListChan,
		cloudSnapUpdateChan, cloudSnapDeleteChan, exportPolUpdateChan, snapshotPolUpdateChan, quarkUpdateChan, quarkDeleteChan,
		volumeReplicationUpdateChan, volumeReplicationDeleteChan, clusterPeeringUpdateChan, clusterPeeringDeleteChan,
		objectStoreUpdateChan, objectStoreDeleteChan, volumeCloudBackupListUpdateChan)

	client.volumeUpdateChannel = volUpdateChan
	client.volumeDeleteChannel = volDeleteChan
	client.storagePoolUpdateChannel = spUpdateChan
	client.storagePoolDeleteChannel = spDeleteChan
	client.snapshotUpdateChannel = snapUpdateChan
	client.snapshotDeleteChannel = snapDeleteChan
	client.snapshotRestoreChannel = snapRestoreChan
	client.snapshotsListChannel = snapListChan
	client.cloudSnapshotUpdateChannel = cloudSnapUpdateChan
	client.cloudSnapshotDeleteChannel = cloudSnapDeleteChan
	client.exportPolicyUpdateChannel = exportPolUpdateChan
	client.snapshotPolicyUpdateChannel = snapshotPolUpdateChan
	client.quarkUpdateChannel = quarkUpdateChan
	client.quarkDeleteChannel = quarkDeleteChan
	client.clusterPeeringUpdateChannel = clusterPeeringUpdateChan
	client.clusterPeeringDeleteChannel = clusterPeeringDeleteChan
	client.volumeReplicationUpdateChannel = volumeReplicationUpdateChan
	client.volumeReplicationDeleteChannel = volumeReplicationDeleteChan
	client.objectStoreUpdateChannel = objectStoreUpdateChan
	client.objectStoreDeleteChannel = objectStoreDeleteChan
	client.volumeCloudBackupListUpdateChannel = volumeCloudBackupListUpdateChan

	if err := client.SetupController(stopCh); err != nil {
		return nil, err
	}

	//go client.runNetAppVolumeController(threadiness, stopCh)
	return eventHandler, nil
}

// Stops the event controller.
func (client *NVCClient) StopController() {
	close(client.volumeUpdateChannel)
	client.volumeUpdateChannel = nil
	close(client.snapshotUpdateChannel)
	client.snapshotUpdateChannel = nil
	close(client.volumeDeleteChannel)
	client.volumeDeleteChannel = nil
	close(client.snapshotDeleteChannel)
	client.snapshotDeleteChannel = nil
	close(client.snapshotRestoreChannel)
	client.snapshotRestoreChannel = nil
	close(client.snapshotsListChannel)
	client.snapshotsListChannel = nil
	close(client.cloudSnapshotUpdateChannel)
	client.cloudSnapshotUpdateChannel = nil
	close(client.cloudSnapshotDeleteChannel)
	client.cloudSnapshotDeleteChannel = nil
	close(client.exportPolicyUpdateChannel)
	client.exportPolicyUpdateChannel = nil
	close(client.snapshotPolicyUpdateChannel)
	client.snapshotPolicyUpdateChannel = nil
	close(client.quarkUpdateChannel)
	client.quarkUpdateChannel = nil
	close(client.quarkDeleteChannel)
	client.quarkDeleteChannel = nil
	close(client.storagePoolUpdateChannel)
	client.storagePoolUpdateChannel = nil
	close(client.storagePoolDeleteChannel)
	client.storagePoolDeleteChannel = nil
	close(client.clusterPeeringUpdateChannel)
	client.clusterPeeringUpdateChannel = nil
	close(client.clusterPeeringDeleteChannel)
	client.clusterPeeringDeleteChannel = nil
	close(client.volumeReplicationUpdateChannel)
	client.volumeReplicationUpdateChannel = nil
	close(client.volumeReplicationDeleteChannel)
	client.volumeReplicationDeleteChannel = nil
	close(client.objectStoreUpdateChannel)
	client.objectStoreUpdateChannel = nil
	close(client.objectStoreDeleteChannel)
	client.objectStoreDeleteChannel = nil
	close(client.volumeCloudBackupListUpdateChannel)
	client.volumeCloudBackupListUpdateChannel = nil
}

func (client *NVCClient) SetupController(stopCh <-chan struct{}) error {
	informerFactory := quarkinformers.NewSharedInformerFactoryWithOptions(client.quarkClient, time.Second*30, quarkinformers.WithNamespace(client.nvcNamespace))
	volumeInformer := informerFactory.Quark().V1alpha1().NetAppVolumes()
	client.volumesLister = volumeInformer.Lister()
	client.volumesSynced = volumeInformer.Informer().HasSynced
	//client.workqueue = workqueue.NewNamedRateLimitingQueue(workqueue.DefaultControllerRateLimiter(), "NetAppVolumes")

	volumeInformer.Informer().AddEventHandler(cache.ResourceEventHandlerFuncs{
		DeleteFunc: func(obj interface{}) {
			volume, ok := obj.(*quarkv1alpha1.NetAppVolume)
			if ok {
				client.handleVolumeDeletes(volume)
			}
		},
		UpdateFunc: func(old, new interface{}) {
			newVolume := new.(*quarkv1alpha1.NetAppVolume)
			oldVolume := old.(*quarkv1alpha1.NetAppVolume)

			// Filter previously seen NetAppVolumes.
			//key, err := cache.MetaNamespaceKeyFunc(new)

			_, err := cache.MetaNamespaceKeyFunc(new)
			if err != nil {
				utilruntime.HandleError(err)
				return
			}
			if newVolume.ResourceVersion == oldVolume.ResourceVersion {
				return
			}
			log.WithFields(log.Fields{
				"volume":       newVolume.Name,
				"newRV":        newVolume.ResourceVersion,
				"oldRV":        oldVolume.ResourceVersion,
				"newVolStatus": newVolume.Status,
				"oldVolStatus": oldVolume.Status,
			}).Debug("NetAppVolume is updated.")

			//client.enqueueNetAppVolume(newVolume) //TODO: DEL
			//TODO: Ideally, a scheme like the way Trident diffs the old and
			// new objects make more sense here:
			//TODO: add support for multithreading
			client.handleVolumeUpdates(oldVolume, newVolume)
		},
	})

	storagePoolInformer := informerFactory.Quark().V1alpha1().NetAppStoragePools()
	client.storagePoolLister = storagePoolInformer.Lister()
	client.storagePoolSynced = storagePoolInformer.Informer().HasSynced

	storagePoolInformer.Informer().AddEventHandler(cache.ResourceEventHandlerFuncs{
		DeleteFunc: func(obj interface{}) {
			storagePool, ok := obj.(*quarkv1alpha1.NetAppStoragePool)
			if ok {
				client.handleStoragePoolDeletes(storagePool)
			}
		},
		UpdateFunc: func(old, new interface{}) {
			newStoragePool := new.(*quarkv1alpha1.NetAppStoragePool)
			oldStoragePool := old.(*quarkv1alpha1.NetAppStoragePool)

			_, err := cache.MetaNamespaceKeyFunc(new)
			if err != nil {
				utilruntime.HandleError(err)
				return
			}
			if newStoragePool.ResourceVersion == oldStoragePool.ResourceVersion {
				return
			}
			log.WithFields(log.Fields{
				"storagePool": newStoragePool.Name,
				"newRV":       newStoragePool.ResourceVersion,
				"oldRV":       oldStoragePool.ResourceVersion,
				"newSpStatus": newStoragePool.Status,
				"oldSpStatus": oldStoragePool.Status,
			}).Debug("NetAppStoragePool is updated.")

			// new objects make more sense here:
			client.handleStoragePoolUpdates(oldStoragePool, newStoragePool)
		},
	})

	snapshotInformer := informerFactory.Quark().V1alpha1().NetAppVolumeSnapshots()
	client.snapshotsLister = snapshotInformer.Lister()
	client.snapshotsSynced = snapshotInformer.Informer().HasSynced
	//client.workqueueSnapshot = workqueue.NewNamedRateLimitingQueue(workqueue.DefaultControllerRateLimiter(), "NetAppVolumeSnapshots")
	snapshotInformer.Informer().AddEventHandler(cache.ResourceEventHandlerFuncs{
		UpdateFunc: func(old, new interface{}) {
			newVolSnap := new.(*quarkv1alpha1.NetAppVolumeSnapshot)
			oldVolSnap := old.(*quarkv1alpha1.NetAppVolumeSnapshot)

			// Filter previously seen NetAppVolumeSnapshots.
			//key, err := cache.MetaNamespaceKeyFunc(new)
			_, err := cache.MetaNamespaceKeyFunc(new)
			if err != nil {
				utilruntime.HandleError(err)
				return
			}
			if newVolSnap.ResourceVersion == oldVolSnap.ResourceVersion {
				return
			}
			log.WithFields(log.Fields{
				"snapshot":         newVolSnap.Name,
				"newRV":            newVolSnap.ResourceVersion,
				"oldRV":            oldVolSnap.ResourceVersion,
				"newVolSnapStatus": newVolSnap.Status,
				"oldVolSnapStatus": oldVolSnap.Status,
			}).Debug("NetAppVolumeSnapshot is updated.")

			//TODO: add support for multithreading
			client.handleSnapshotUpdates(oldVolSnap, newVolSnap)
		},
	})

	snapshotsListInformer := informerFactory.Quark().V1alpha1().NetAppVolumeSnapshotsLists()
	client.snapshotsListLister = snapshotsListInformer.Lister()
	client.snapshotsListSynced = snapshotsListInformer.Informer().HasSynced
	//client.workqueueSnapshotsList = workqueue.NewNamedRateLimitingQueue(workqueue.DefaultControllerRateLimiter(), "NetAppVolumeSnapshotsList")
	snapshotsListInformer.Informer().AddEventHandler(cache.ResourceEventHandlerFuncs{
		UpdateFunc: func(old, new interface{}) {
			newVolSnapLst := new.(*quarkv1alpha1.NetAppVolumeSnapshotsList)
			oldVolSnapLst := old.(*quarkv1alpha1.NetAppVolumeSnapshotsList)

			// Filter previously seen NetAppVolumeSnapshots.
			//key, err := cache.MetaNamespaceKeyFunc(new)
			_, err := cache.MetaNamespaceKeyFunc(new)
			if err != nil {
				utilruntime.HandleError(err)
				return
			}
			if newVolSnapLst.ResourceVersion == oldVolSnapLst.ResourceVersion {
				return
			}
			log.WithFields(log.Fields{
				"snapshotsList":         newVolSnapLst.Name,
				"newRV":                 newVolSnapLst.ResourceVersion,
				"oldRV":                 oldVolSnapLst.ResourceVersion,
				"newVolSnapsListStatus": newVolSnapLst.Status,
				"oldVolSnapsListStatus": oldVolSnapLst.Status,
			}).Debug("NetAppVolumeSnapshotsList is updated.")

			//TODO: add support for multithreading
			client.handleSnapshotsListUpdates(oldVolSnapLst, newVolSnapLst)
			client.handleSnapshotRestoreUpdates(oldVolSnapLst, newVolSnapLst)
		},
	})

	cloudSnapshotInformer := informerFactory.Quark().V1alpha1().NetAppVolumeCloudSnapshots()
	client.cloudSnapshotsLister = cloudSnapshotInformer.Lister()
	client.cloudSnapshotsSynced = cloudSnapshotInformer.Informer().HasSynced
	//client.workqueuecloudSnapshot = workqueue.NewNamedRateLimitingQueue(workqueue.DefaultControllerRateLimiter(), "NetAppVolumeCloudSnapshot")
	cloudSnapshotInformer.Informer().AddEventHandler(cache.ResourceEventHandlerFuncs{
		DeleteFunc: func(obj interface{}) {
			cloudSnapshot, ok := obj.(*quarkv1alpha1.NetAppVolumeCloudSnapshot)
			if ok {
				client.handleCloudSnapshotDeletes(cloudSnapshot)
			}
		},
		UpdateFunc: func(old, new interface{}) {
			newCloudSnap := new.(*quarkv1alpha1.NetAppVolumeCloudSnapshot)
			oldCloudSnap := old.(*quarkv1alpha1.NetAppVolumeCloudSnapshot)

			// Filter previously seen NetAppVolumeCloudSnapshots.
			_, err := cache.MetaNamespaceKeyFunc(new)
			if err != nil {
				utilruntime.HandleError(err)
				return
			}
			if newCloudSnap.ResourceVersion == oldCloudSnap.ResourceVersion {
				return
			}
			log.WithFields(log.Fields{
				"cloudSnapshot":      newCloudSnap.Name,
				"newRV":              newCloudSnap.ResourceVersion,
				"oldRV":              oldCloudSnap.ResourceVersion,
				"newCloudSnapStatus": newCloudSnap.Status,
				"oldCloudSnapStatus": oldCloudSnap.Status,
			}).Debug("NetAppVolumeCloudSnapshot is updated.")

			//TODO: add support for multithreading
			client.handleCloudSnapshotUpdates(oldCloudSnap, newCloudSnap)
		},
	})

	exportPolicyInformer := informerFactory.Quark().V1alpha1().NetAppExportPolicies()
	client.exportPolicyLister = exportPolicyInformer.Lister()
	client.exportPolicySynced = exportPolicyInformer.Informer().HasSynced
	exportPolicyInformer.Informer().AddEventHandler(cache.ResourceEventHandlerFuncs{
		UpdateFunc: func(old, new interface{}) {
			newExportPolicy := new.(*quarkv1alpha1.NetAppExportPolicy)
			oldExportPolicy := old.(*quarkv1alpha1.NetAppExportPolicy)

			_, err := cache.MetaNamespaceKeyFunc(new)
			if err != nil {
				utilruntime.HandleError(err)
				return
			}
			if newExportPolicy.ResourceVersion == oldExportPolicy.ResourceVersion {
				return
			}
			log.WithFields(log.Fields{
				"ExportPolicy":          newExportPolicy.Name,
				"newRV":                 newExportPolicy.ResourceVersion,
				"oldRV":                 oldExportPolicy.ResourceVersion,
				"newExportPolicyStatus": newExportPolicy.Status,
				"oldExportPolicyStatus": oldExportPolicy.Status,
			}).Debug("NetAppVolumeExportPolicy is updated.")

			//TODO: add support for multithreading
			client.handleExportPolicyUpdates(oldExportPolicy, newExportPolicy)
		},
	})

	snapshotPolicyInformer := informerFactory.Quark().V1alpha1().NetAppSnapshotPolicies()
	client.snapshotPolicyLister = snapshotPolicyInformer.Lister()
	client.snapshotPolicySynced = snapshotPolicyInformer.Informer().HasSynced
	//client.workqueuecloudSnapshot = workqueue.NewNamedRateLimitingQueue(workqueue.DefaultControllerRateLimiter(), "NetAppVolumeCloudSnapshot")
	snapshotPolicyInformer.Informer().AddEventHandler(cache.ResourceEventHandlerFuncs{
		UpdateFunc: func(old, new interface{}) {
			newSnapshotPolicy := new.(*quarkv1alpha1.NetAppSnapshotPolicy)
			oldSnapshotPolicy := old.(*quarkv1alpha1.NetAppSnapshotPolicy)

			_, err := cache.MetaNamespaceKeyFunc(new)
			if err != nil {
				utilruntime.HandleError(err)
				return
			}
			if newSnapshotPolicy.ResourceVersion == oldSnapshotPolicy.ResourceVersion {
				return
			}
			log.WithFields(log.Fields{
				"snapshotPolicy":          newSnapshotPolicy.Name,
				"newRV":                   newSnapshotPolicy.ResourceVersion,
				"oldRV":                   oldSnapshotPolicy.ResourceVersion,
				"newSnapshotPolicyStatus": newSnapshotPolicy.Status,
				"oldSnapshotPolicyStatus": oldSnapshotPolicy.Status,
			}).Debug("NetAppVolumeSnapshotPolicy is updated.")

			//TODO: add support for multithreading
			client.handleSnapshotPolicyUpdates(oldSnapshotPolicy, newSnapshotPolicy)
		},
	})

	quarkInformer := informerFactory.Quark().V1alpha1().Quarks()
	client.quarksLister = quarkInformer.Lister()
	client.quarksSynced = quarkInformer.Informer().HasSynced

	quarkInformer.Informer().AddEventHandler(cache.ResourceEventHandlerFuncs{
		DeleteFunc: func(obj interface{}) {
			quark, ok := obj.(*quarkv1alpha1.Quark)
			if ok {
				client.handleQuarkDeletes(quark)
			}
		},
		UpdateFunc: func(old, new interface{}) {
			newQuark := new.(*quarkv1alpha1.Quark)
			oldQuark := old.(*quarkv1alpha1.Quark)

			// Filter previously seen Quark.
			_, err := cache.MetaNamespaceKeyFunc(new)
			if err != nil {
				utilruntime.HandleError(err)
				return
			}
			if newQuark.ResourceVersion == oldQuark.ResourceVersion {
				return
			}
			log.WithFields(log.Fields{
				"quark":          newQuark.Name,
				"newRV":          newQuark.ResourceVersion,
				"oldRV":          oldQuark.ResourceVersion,
				"newQuarkStatus": newQuark.Status,
				"oldQuarkStatus": oldQuark.Status,
			}).Debug("Quark is updated.")

			//TODO: add support for multithreading
			client.handleQuarkUpdates(oldQuark, newQuark)
		},
	})

	if client.EnableCRR {
		volumeReplicationInformer := informerFactory.Quark().V1alpha1().NetAppVolumeReplications()
		client.volumeReplicationsLister = volumeReplicationInformer.Lister()
		client.volumeReplicationSynced = volumeReplicationInformer.Informer().HasSynced

		volumeReplicationInformer.Informer().AddEventHandler(cache.ResourceEventHandlerFuncs{
			DeleteFunc: func(obj interface{}) {
				volRepl, ok := obj.(*quarkv1alpha1.NetAppVolumeReplication)
				if ok {
					client.handleVolumeReplicationDelete(volRepl)
				}
			},
			UpdateFunc: func(old, new interface{}) {
				newVolumeReplication := new.(*quarkv1alpha1.NetAppVolumeReplication)
				oldVolumeReplication := old.(*quarkv1alpha1.NetAppVolumeReplication)

				_, err := cache.MetaNamespaceKeyFunc(new)
				if err != nil {
					utilruntime.HandleError(err)
					return
				}
				if newVolumeReplication.ResourceVersion == oldVolumeReplication.ResourceVersion {
					return
				}
				log.WithFields(log.Fields{
					"volumeReplication": newVolumeReplication.Name,
					"newRV":             newVolumeReplication.ResourceVersion,
					"oldRV":             oldVolumeReplication.ResourceVersion,
					"newVRStatus":       newVolumeReplication.Status,
					"oldVRStatus":       oldVolumeReplication.Status,
				}).Debug("NetAppVolumeReplication is updated.")
				client.handleNetappVolumeReplicationUpdate(oldVolumeReplication, newVolumeReplication)
			},
		})
	}

	if client.EnableCRR || client.EnableServiceMesh {
		clusterPeeringInformer := informerFactory.Quark().V1alpha1().NetAppClusterPeerings()
		client.clusterPeeringLister = clusterPeeringInformer.Lister()
		client.clusterPeeringSynced = clusterPeeringInformer.Informer().HasSynced
		clusterPeeringInformer.Informer().AddEventHandler(cache.ResourceEventHandlerFuncs{
			DeleteFunc: func(obj interface{}) {
				clusterPeering, ok := obj.(*quarkv1alpha1.NetAppClusterPeering)
				if ok {
					client.handleClusterPeeringDelete(clusterPeering)
				}
			},
			UpdateFunc: func(old, new interface{}) {
				newClusterPeering := new.(*quarkv1alpha1.NetAppClusterPeering)
				oldClusterPeering := old.(*quarkv1alpha1.NetAppClusterPeering)

				_, err := cache.MetaNamespaceKeyFunc(new)
				if err != nil {
					utilruntime.HandleError(err)
					return
				}
				if newClusterPeering.ResourceVersion == oldClusterPeering.ResourceVersion {
					return
				}
				log.WithFields(log.Fields{
					"clusterPeeringName":      newClusterPeering.Name,
					"newRV":                   newClusterPeering.ResourceVersion,
					"oldRV":                   oldClusterPeering.ResourceVersion,
					"newClusterPeeringStatus": newClusterPeering.Status,
					"oldClusterPeeringStatus": oldClusterPeering.Status,
				}).Debug("NetAppClusterPeering is updated.")

				//TODO: add support for multithreading
				client.handleClusterPeeringUpdates(oldClusterPeering, newClusterPeering)
			},
		})
	}

	if client.EnableIntegratedBackup {
		objectStoreInformer := informerFactory.Quark().V1alpha1().NetAppObjectStores()
		client.objectStoreLister = objectStoreInformer.Lister()
		client.objectStoreSynced = objectStoreInformer.Informer().HasSynced

		objectStoreInformer.Informer().AddEventHandler(cache.ResourceEventHandlerFuncs{
			DeleteFunc: func(obj interface{}) {
				objStore, ok := obj.(*quarkv1alpha1.NetAppObjectStore)
				if ok {
					client.handleObjectStoreDelete(objStore)
				}
			},
			UpdateFunc: func(oldObjStore, newObjStore interface{}) {
				newObjectStore := newObjStore.(*quarkv1alpha1.NetAppObjectStore)
				oldObjectStore := oldObjStore.(*quarkv1alpha1.NetAppObjectStore)

				_, err := cache.MetaNamespaceKeyFunc(newObjectStore)
				if err != nil {
					utilruntime.HandleError(err)
					return
				}
				if newObjectStore.ResourceVersion == oldObjectStore.ResourceVersion {
					return
				}
				log.WithFields(log.Fields{
					"objectStoreName":      newObjectStore.Name,
					"newRV":                newObjectStore.ResourceVersion,
					"oldRV":                oldObjectStore.ResourceVersion,
					"newObjectStoreStatus": newObjectStore.Status,
					"oldObjectStoreStatus": oldObjectStore.Status,
				}).Debug("NetAppObjectStore is updated.")

				client.handleObjectStoreUpdates(oldObjectStore, newObjectStore)
			},
		})

		volumeCloudBackupListInformer := informerFactory.Quark().V1alpha1().NetAppVolumeCloudBackupLists()
		client.volumeCloudBackupListLister = volumeCloudBackupListInformer.Lister()
		client.volumeCloudBackupListSynced = volumeCloudBackupListInformer.Informer().HasSynced

		volumeCloudBackupListInformer.Informer().AddEventHandler(cache.ResourceEventHandlerFuncs{
			UpdateFunc: func(oldVolCloudBackupList, newVolCloudBackupList interface{}) {
				newVolumeCloudBackupList := newVolCloudBackupList.(*quarkv1alpha1.NetAppVolumeCloudBackupList)
				oldVolumeCloudBackupList := oldVolCloudBackupList.(*quarkv1alpha1.NetAppVolumeCloudBackupList)

				_, err := cache.MetaNamespaceKeyFunc(newVolumeCloudBackupList)
				if err != nil {
					utilruntime.HandleError(err)
					return
				}
				if newVolumeCloudBackupList.ResourceVersion == oldVolumeCloudBackupList.ResourceVersion {
					return
				}
				log.WithFields(log.Fields{
					"volumeCloudBackupListName":      newVolumeCloudBackupList.Name,
					"newRV":                          newVolumeCloudBackupList.ResourceVersion,
					"oldRV":                          oldVolumeCloudBackupList.ResourceVersion,
					"newVolumeCloudBackupListStatus": newVolumeCloudBackupList.Status,
					"oldVolumeCloudBackupListStatus": oldVolumeCloudBackupList.Status,
				}).Debug("NetAppVolumeCloudBackupList is updated.")

				client.handleVolumeCloudBackupListUpdates(oldVolumeCloudBackupList, newVolumeCloudBackupList)
			},
		})
	}

	informerFactory.Start(stopCh)
	log.Debug("Waiting for cache sync.")
	cacheSyncs := []cache.InformerSynced{client.volumesSynced, client.snapshotsSynced, client.snapshotsListSynced,
		client.cloudSnapshotsSynced, client.quarksSynced, client.storagePoolSynced, client.exportPolicySynced,
		client.snapshotPolicySynced}
	if client.EnableCRR || client.EnableServiceMesh {
		cacheSyncs = append(cacheSyncs, client.clusterPeeringSynced)
	}
	if client.EnableCRR {
		cacheSyncs = append(cacheSyncs, client.volumeReplicationSynced)
	}
	if client.EnableIntegratedBackup {
		cacheSyncs = append(cacheSyncs, client.objectStoreSynced)
		cacheSyncs = append(cacheSyncs, client.volumeCloudBackupListSynced)

	}
	if ok := cache.WaitForCacheSync(stopCh, cacheSyncs...); !ok {
		return fmt.Errorf("failed to wait for caches to sync")
	}

	log.Debug("Controller finished cache sync.")
	return nil
}

func (client *NVCClient) handleSnapshotRestoreUpdates(oldVolSnapList *quarkv1alpha1.NetAppVolumeSnapshotsList, newVolSnapList *quarkv1alpha1.NetAppVolumeSnapshotsList) {
	// send an event when SnapshotRestoreError condition added to SnapshotsList conditions
	if len(newVolSnapList.Status.Conditions) != len(oldVolSnapList.Status.Conditions) {
		for _, cond := range newVolSnapList.Status.Conditions {
			if cond.Type == quarkv1alpha1.SnapshotRestoreError && cond.Status == corev1.ConditionTrue {
				client.snapshotRestoreChannel <- &NetAppVolumeSnapshotRestore{
					VolumeUUID:       newVolSnapList.Name,
					ErrorCondition:   cond.Message,
					RestoredSnapshot: false,
				}
				log.WithFields(log.Fields{
					"snapshot":      newVolSnapList.Name,
					"oldConditions": oldVolSnapList.Status.Conditions,
					"newConditions": newVolSnapList.Status.Conditions,
				}).Debug("NetAppVolumeSnapshotsList conditions have been updated with SnapshotRestoreError condition.")
			}
		}
	}

	if newVolSnapList.Status.RestoredSnapshot != oldVolSnapList.Status.RestoredSnapshot {
		client.snapshotRestoreChannel <- &NetAppVolumeSnapshotRestore{
			VolumeUUID:       newVolSnapList.Name,
			ErrorCondition:   "",
			RestoredSnapshot: true,
		}
		log.WithFields(log.Fields{
			"snapshot":            newVolSnapList.Name,
			"oldRestoredSnapshot": oldVolSnapList.Status.RestoredSnapshot,
			"newRestoredSnapshot": newVolSnapList.Status.RestoredSnapshot,
		}).Debug("Detected SnapRestore.")
	}
}

func (client *NVCClient) handleVolumeDeletes(vol *quarkv1alpha1.NetAppVolume) {
	client.volumeDeleteChannel <- &NetAppVolumeDelete{vol}
}

func (client *NVCClient) handleVolumeUpdates(oldVol *quarkv1alpha1.NetAppVolume, newVol *quarkv1alpha1.NetAppVolume) {
	sendEvent := false

	// We send events when any of the NetAppVolume has converged to the desired state.
	_, newStatus, _ := client.NetAppVolumeUpdated(newVol)
	_, oldStatus, _ := client.NetAppVolumeUpdated(oldVol)
	if newStatus != oldStatus {
		log.WithFields(log.Fields{
			"volume":    newVol.Name,
			"newStatus": newStatus,
		}).Debug("NetApp Volume state has been updated.")
		sendEvent = true
	}
	// We send events when the NetAppVolume's conditions have changed.
	if len(newVol.Status.Conditions) != len(oldVol.Status.Conditions) {
		sendEvent = true
		log.WithFields(log.Fields{
			"volume":     newVol.Name,
			"conditions": newVol.Status.Conditions,
		}).Debug("NetApp Volume has updated conditions.")
	} else {
		for i, cond := range newVol.Status.Conditions {
			if oldVol.Status.Conditions[i].Type != cond.Type || oldVol.Status.Conditions[i].Status != cond.Status {
				sendEvent = true
				log.WithFields(log.Fields{
					"volume":     newVol.Name,
					"conditions": newVol.Status.Conditions,
				}).Debug("NetApp Volume has updated conditions.")
				break
			}
		}
	}

	if sendEvent {
		client.volumeUpdateChannel <- &NetAppVolumeUpdate{
			Volume:            newVol,
			Status:            newStatus,
			UpdatedConditions: newVol.Status.Conditions,
		}
	}
}

func (client *NVCClient) handleSnapshotUpdates(oldVolSnap *quarkv1alpha1.NetAppVolumeSnapshot, newVolSnap *quarkv1alpha1.NetAppVolumeSnapshot) {
	sendEvent := false

	// We send events when the NetAppVolumeSnapshot's conditions have changed.
	if len(newVolSnap.Status.Conditions) != len(oldVolSnap.Status.Conditions) {
		sendEvent = true
		log.WithFields(log.Fields{
			"snapshot":   newVolSnap.Name,
			"conditions": newVolSnap.Status.Conditions,
		}).Debug("NetApp Volume Snapshot has updated conditions.")
	} else {
		for i, cond := range newVolSnap.Status.Conditions {
			if oldVolSnap.Status.Conditions[i].Type != cond.Type || oldVolSnap.Status.Conditions[i].Status != cond.Status {
				sendEvent = true
				log.WithFields(log.Fields{
					"snapshot":   newVolSnap.Name,
					"conditions": newVolSnap.Status.Conditions,
				}).Debug("NetApp Volume Snapshot has updated conditions.")
				break
			}
		}
	}

	if sendEvent {
		client.snapshotUpdateChannel <- &NetAppVolumeSnapshotUpdate{
			Snapshot:          newVolSnap,
			UpdatedConditions: newVolSnap.Status.Conditions,
		}
	}
}

// Difference between two snapshotsList. All the snapshots in the oldList that are not in the newList
func differenceSnapshots(oldList, newList []quarkv1alpha1.NetAppVolumeSnapshotInfo) (diff []string) {
	listMap := make(map[string]int)

	key := ""
	for _, snap := range newList {
		key = snap.Name + snap.CreationTime.String()
		listMap[key] = listMap[key] + 1
	}

	for _, snap := range oldList {
		key = snap.Name + snap.CreationTime.String()
		if count, ok := listMap[key]; !ok || count == 0 {
			diff = append(diff, snap.Name)
		} else {
			listMap[key] = listMap[key] - 1
		}
	}
	return
}

func (client *NVCClient) snapshotsListConditionsUpdated(oldVolSnap *quarkv1alpha1.NetAppVolumeSnapshotsList, newVolSnap *quarkv1alpha1.NetAppVolumeSnapshotsList) bool {
	if len(newVolSnap.Status.Conditions) != len(oldVolSnap.Status.Conditions) {
		return true
	} else {
		var (
			newConditionLastTransitionTime metav1.Time
			oldConditionLastTransitionTime metav1.Time
		)
		for _, cond := range newVolSnap.Status.Conditions {
			if cond.Type == quarkv1alpha1.SnapshotsListReady && cond.Status == corev1.ConditionTrue {
				newConditionLastTransitionTime = cond.LastTransitionTime
			}
		}
		for _, cond := range oldVolSnap.Status.Conditions {
			if cond.Type == quarkv1alpha1.SnapshotsListReady && cond.Status == corev1.ConditionTrue {
				oldConditionLastTransitionTime = cond.LastTransitionTime
			}
		}
		newConditionUpdateTime := newConditionLastTransitionTime.Time
		oldConditionUpdateTime := oldConditionLastTransitionTime.Time
		log.WithFields(log.Fields{
			"snapshot":      newVolSnap.Name,
			"oldConditions": oldVolSnap.Status.Conditions,
			"newConditions": newVolSnap.Status.Conditions,
		}).Debug("NetApp Volume Snapshots condition has been updated.")

		return !newConditionUpdateTime.Equal(oldConditionUpdateTime)
	}

}

func (client *NVCClient) handleSnapshotsListUpdates(oldVolSnap *quarkv1alpha1.NetAppVolumeSnapshotsList, newVolSnap *quarkv1alpha1.NetAppVolumeSnapshotsList) {
	// We send events when the NetAppVolumeSnapshotsList's list have changed.
	if client.snapshotsListConditionsUpdated(oldVolSnap, newVolSnap) {
		// newSnapshotsList contains all snapshots within the NetAppVolumeSnapshotsList
		var newSnapshotsList []string
		for _, snap := range newVolSnap.Status.Snapshots {
			newSnapshotsList = append(newSnapshotsList, snap.Name)
		}
		client.snapshotsListChannel <- &NetAppVolumeSnapshotsListUpdate{
			VolumeUUID:           newVolSnap.Name,
			SnapshotsListUpdated: true,
			SnapshotNames:        newSnapshotsList,
		}
		log.WithFields(log.Fields{
			"volume":    newVolSnap.Name,
			"snapshots": newVolSnap.Status.Snapshots,
		}).Debug("NetAppVolumeSnapshotsList has been updated.")
	}
	// All the snapshots that were in snapshots old list before but not in new list are deleted.
	// We are comparing Snapshots instead of deletedSnapshots since multiple updates can be coalesced together at client-go.
	deletedList := differenceSnapshots(oldVolSnap.Status.Snapshots, newVolSnap.Status.Snapshots)
	if len(deletedList) > 0 {
		client.snapshotDeleteChannel <- &NetAppVolumeSnapshotDelete{
			VolumeUUID:       newVolSnap.Name,
			DeletedSnapshots: deletedList,
		}
		log.WithFields(log.Fields{
			"volume":    newVolSnap.Name,
			"snapshots": newVolSnap.Status.DeletedSnapshots,
		}).Debug("Detected deletion of snapshots.")
	}
}

func (client *NVCClient) handleCloudSnapshotDeletes(cloudSnap *quarkv1alpha1.NetAppVolumeCloudSnapshot) {
	client.cloudSnapshotDeleteChannel <- &NetAppVolumeCloudSnapshotDelete{cloudSnap}
}

func (client *NVCClient) handleCloudSnapshotUpdates(oldCloudSnap *quarkv1alpha1.NetAppVolumeCloudSnapshot, newCloudSnap *quarkv1alpha1.NetAppVolumeCloudSnapshot) {
	sendEvent := false

	oldCloudSnapInfo := getCloudSnapshotInfo(oldCloudSnap)
	newCloudSnapInfo := getCloudSnapshotInfo(newCloudSnap)

	if oldCloudSnapInfo.Ready != newCloudSnapInfo.Ready {
		sendEvent = true
	}
	if oldCloudSnapInfo.Uploading != newCloudSnapInfo.Uploading {
		sendEvent = true
	}
	if oldCloudSnapInfo.Deleting != newCloudSnapInfo.Deleting {
		sendEvent = true
	}
	if oldCloudSnapInfo.EndpointUUID != newCloudSnapInfo.EndpointUUID {
		sendEvent = true
	}
	if oldCloudSnapInfo.LogicalSize != newCloudSnapInfo.LogicalSize {
		sendEvent = true
	}
	if oldCloudSnapInfo.CompletionPercent != newCloudSnapInfo.CompletionPercent {
		sendEvent = true
	}
	if oldCloudSnapInfo.Error != newCloudSnapInfo.Error {
		if (oldCloudSnapInfo.Error != nil) &&
			(newCloudSnapInfo.Error != nil) {
			if oldCloudSnapInfo.Error.Error() != newCloudSnapInfo.Error.Error() {
				sendEvent = true
			}
		} else {
			sendEvent = true
		}
	}
	if oldCloudSnapInfo.CreationTime != newCloudSnapInfo.CreationTime {
		if (oldCloudSnapInfo.CreationTime != nil) &&
			(newCloudSnapInfo.CreationTime != nil) {
			if *oldCloudSnapInfo.CreationTime != *newCloudSnapInfo.CreationTime {
				sendEvent = true
			}
		} else {
			sendEvent = true
		}
	}

	if sendEvent {
		client.cloudSnapshotUpdateChannel <- &NetAppVolumeCloudSnapshotUpdate{
			CloudSnapshot: newCloudSnap,
			Status:        newCloudSnapInfo,
		}
	}
}

func (client *NVCClient) handleExportPolicyUpdates(oldExportpolicy *quarkv1alpha1.NetAppExportPolicy, newExportPolicy *quarkv1alpha1.NetAppExportPolicy) {
	sendEvent := false

	if !reflect.DeepEqual(oldExportpolicy.Status.Rules, newExportPolicy.Status.Rules) {
		sendEvent = true
	}
	if sendEvent {
		client.exportPolicyUpdateChannel <- &NetAppVolumeExportPolicyUpdate{
			ExportPolicy: newExportPolicy,
			Rules:        newExportPolicy.Status.Rules,
		}
	}
}

func (client *NVCClient) handleSnapshotPolicyUpdates(oldSnapshotpolicy *quarkv1alpha1.NetAppSnapshotPolicy, newSnapshotPolicy *quarkv1alpha1.NetAppSnapshotPolicy) {
	sendEvent := false

	if (oldSnapshotpolicy.Status.Enabled != newSnapshotPolicy.Status.Enabled) ||
		(!reflect.DeepEqual(oldSnapshotpolicy.Status.Schedules, newSnapshotPolicy.Status.Schedules)) {
		sendEvent = true
	}
	if sendEvent {
		client.snapshotPolicyUpdateChannel <- &NetAppVolumeSnapshotPolicyUpdate{
			SnapshotPolicy: newSnapshotPolicy,
			Enabled:        newSnapshotPolicy.Status.Enabled,
			Schedules:      newSnapshotPolicy.Status.Schedules,
		}
	}
}

func (client *NVCClient) handleClusterPeeringUpdates(oldClusterPeering *quarkv1alpha1.NetAppClusterPeering, newClusterPeering *quarkv1alpha1.NetAppClusterPeering) {
	sendEvent := false

	if (oldClusterPeering.Status.State != newClusterPeering.Status.State) ||
		(!reflect.DeepEqual(oldClusterPeering.Status.Conditions, newClusterPeering.Status.Conditions)) {
		sendEvent = true
	}

	if sendEvent {
		client.clusterPeeringUpdateChannel <- &NetAppClusterPeeringUpdate{
			ClusterPeering:    newClusterPeering,
			UpdatedConditions: newClusterPeering.Status.Conditions,
		}
	}
}

// handleClusterPeeringDelete is used to terminate the watch on ClusterPeering events
func (client *NVCClient) handleClusterPeeringDelete(clusterPeering *quarkv1alpha1.NetAppClusterPeering) {
	client.clusterPeeringDeleteChannel <- &NetAppClusterPeeringDelete{ClusterPeering: clusterPeering}
}

func (client *NVCClient) handleObjectStoreUpdates(oldObjectStore *quarkv1alpha1.NetAppObjectStore, newObjectStore *quarkv1alpha1.NetAppObjectStore) {
	sendEvent := false

	if (oldObjectStore.Status.SecretRefCounter != newObjectStore.Status.SecretRefCounter) ||
		(!reflect.DeepEqual(oldObjectStore.Status.Conditions, newObjectStore.Status.Conditions)) {
		sendEvent = true
	}

	if sendEvent {
		client.objectStoreUpdateChannel <- &NetAppObjectStoreUpdate{
			ObjectStore:       newObjectStore,
			UpdatedConditions: newObjectStore.Status.Conditions,
		}
	}
}

func (client *NVCClient) handleVolumeCloudBackupListUpdates(oldVolumeCloudBackupList *quarkv1alpha1.NetAppVolumeCloudBackupList, newVolumeCloudBackupList *quarkv1alpha1.NetAppVolumeCloudBackupList) {
	if !reflect.DeepEqual(oldVolumeCloudBackupList.Status.DeleteBackups, newVolumeCloudBackupList.Status.DeleteBackups) {
		client.volumeCloudBackupListUpdateChannel <- &NetAppVolumeCloudBackupListUpdate{
			VolumeCloudBackupList: newVolumeCloudBackupList,
			UpdatedConditions:     newVolumeCloudBackupList.Status.Conditions,
		}
	}
}

// handleObjectStoreDelete is used to terminate the watch on ObjectStore events
func (client *NVCClient) handleObjectStoreDelete(objStore *quarkv1alpha1.NetAppObjectStore) {
	client.objectStoreDeleteChannel <- &NetAppObjectStoreDelete{ObjectStore: objStore}
}

func (client *NVCClient) handleQuarkUpdates(oldQuark *quarkv1alpha1.Quark, newQuark *quarkv1alpha1.Quark) {
	sendEvent := false

	oldNtapQuarkStatus := getNetAppQuarkStatus(oldQuark)
	newNtapQuarkStatus := getNetAppQuarkStatus(newQuark)

	if oldNtapQuarkStatus.OperatorReady != newNtapQuarkStatus.OperatorReady {
		sendEvent = true
	}

	if oldNtapQuarkStatus.VolumeControllerReady != newNtapQuarkStatus.VolumeControllerReady {
		sendEvent = true
	}

	if oldNtapQuarkStatus.HighAvailabilityReady != newNtapQuarkStatus.HighAvailabilityReady {
		sendEvent = true
	}

	if oldNtapQuarkStatus.ExternalVolumeAccessReady != newNtapQuarkStatus.ExternalVolumeAccessReady {
		sendEvent = true
	}

	if oldNtapQuarkStatus.CloudsnapshotCapabilityReady != newNtapQuarkStatus.CloudsnapshotCapabilityReady {
		sendEvent = true
	}

	if oldNtapQuarkStatus.IPDReady != newNtapQuarkStatus.IPDReady {
		sendEvent = true
	}

	if oldNtapQuarkStatus.EtcdClusterReady != newNtapQuarkStatus.EtcdClusterReady {
		sendEvent = true
	}

	if oldNtapQuarkStatus.NodeMonitorReady != newNtapQuarkStatus.NodeMonitorReady {
		sendEvent = true
	}

	if oldNtapQuarkStatus.NodevolReady != newNtapQuarkStatus.NodevolReady {
		sendEvent = true
	}

	if oldNtapQuarkStatus.ServiceMeshControllerReady != newNtapQuarkStatus.ServiceMeshControllerReady {
		sendEvent = true
	}

	if oldNtapQuarkStatus.Version != newNtapQuarkStatus.Version {
		sendEvent = true
	}

	if oldNtapQuarkStatus.OtelCollectorReady != newNtapQuarkStatus.OtelCollectorReady {
		sendEvent = true
	}

	if sendEvent {
		client.quarkUpdateChannel <- &NetAppQuarkUpdate{
			Quark:  newQuark,
			Status: newNtapQuarkStatus,
		}
	}
}

func (client *NVCClient) handleQuarkDeletes(quark *quarkv1alpha1.Quark) {
	client.quarkDeleteChannel <- &NetAppQuarkDelete{Quark: quark}
}

// ConstructNetAppVolumeByRef - Constructs a new newapp volume by passing in references to export & snapshot policy objects
func ConstructNetAppVolumeByRef(ownerID, displayName, sizeString, volumeUUID, storagePoolUUID, volumePath string, serviceLevel ServiceLevel, exportPolicy *quarkv1alpha1.NetAppExportPolicy, snapshotPolicy *quarkv1alpha1.NetAppSnapshotPolicy, noSnapDir bool, unixPermission string) (volume *quarkv1alpha1.NetAppVolume, err error) {
	return ConstructNetAppVolume(ownerID, displayName, sizeString, volumeUUID, storagePoolUUID, volumePath, exportPolicy.Name, snapshotPolicy.Name, serviceLevel, noSnapDir, unixPermission)
}

// Constructs NetAppVolume definition.
// Constructs NetAppVolume definition helper routine.
// Sets the common field for NetAppVolume
// ownerID: OwnerID of the volume.
// dispalyName: Name of the volume provided by customer.
// sizeString: Volume Size.
// volumeUUID: Unique ID of volume in k8s cluster.
// volumePath: Volume mount path.
// serviceLevel: Standard for alpha.
// snapshotPolicy: Snapshot Policy object
func constructBaseNetAppVolume(ownerID, displayName, sizeString, volumeUUID, storagePoolUUID, volumePath, snapshotPolicyName string,
	serviceLevel ServiceLevel, noSnapDir bool) (volume *quarkv1alpha1.NetAppVolume, err error) {
	defer func() {
		if e := recover(); e != nil {
			err = fmt.Errorf("invalid value for size")
		}
	}()
	if storagePoolUUID == "" {
		// Validate service level
		switch serviceLevel {
		case ServiceLevelStandard, ServiceLevelRegionalStandard:
		default:
			return nil, fmt.Errorf("invalid value for service level")
		}
	}

	// validate UUID
	_, err = uuid.FromString(volumeUUID)
	if err != nil {
		return nil, fmt.Errorf("invalid value for volumeUUID")
	}
	/*	Commenting out to support backward compatibility
		_, err = uuid.FromString(storagePoolUUID)
		if err != nil {
			return nil, fmt.Errorf("invalid value for storagePoolUUID")
		}
	*/
	// Validate size
	size, err := resource.ParseQuantity(sizeString)
	if err != nil {
		return nil, fmt.Errorf("invalid value for size")
	}
	var minSize resource.Quantity
	if storagePoolUUID == "" {
		minSize = minVolSize
	} else {
		minSize = minSpVolSize
	}

	var minVolSizeDifference = size.Value() - minSize.Value()
	var maxVolSizeDifference = size.Value() - maxVolSize.Value()
	if minVolSizeDifference < 0 || maxVolSizeDifference > 0 || minVolSizeDifference%incrementalSize.Value() != 0 {
		return nil, fmt.Errorf("size=%s, volume size should be a minimum of %s, maximum of %s and a multiple of %s", size.String(), minSize.String(), maxVolSize.String(), incrementalSize.String())
	}

	// Validate ownerID
	if ownerID == "" {
		return nil, fmt.Errorf("volume OwnerID required")
	}
	volume = &quarkv1alpha1.NetAppVolume{
		TypeMeta: metav1.TypeMeta{},
		ObjectMeta: metav1.ObjectMeta{
			Name: volumeUUID,
			Labels: map[string]string{
				quarkconfig.VolumeLabel:  volumeUUID,
				quarkconfig.OwnerIDLabel: ownerID,
			},
		},
		Spec: quarkv1alpha1.NetAppVolumeSpec{
			Size:           size,
			DisplayName:    displayName,
			VolumePath:     volumePath,
			ServiceLevel:   string(serviceLevel),
			SnapshotPolicy: snapshotPolicyName,
			NoSnapDir:      noSnapDir,
			DataProtection: false,
		},
		Status: quarkv1alpha1.NetAppVolumeStatus{},
	}

	if storagePoolUUID != "" {
		volume.Labels[quarkconfig.StoragePoolLabel] = storagePoolUUID
	}
	return
}

// ConstructNetAppVolumeDP creates a NetAppVolume definition for DataProtection (DP) volume (Secondary volume)
func ConstructNetAppVolumeDP(ownerID, displayName, sizeString, volumeUUID, storagePoolUUID, volumePath, exportPolicyName, snapshotPolicyName string,
	serviceLevel ServiceLevel, noSnapDir bool) (volume *quarkv1alpha1.NetAppVolume, err error) {
	volume, err = ConstructNetAppVolume(ownerID,
		displayName,
		sizeString,
		volumeUUID,
		storagePoolUUID,
		volumePath,
		exportPolicyName,
		snapshotPolicyName,
		serviceLevel,
		noSnapDir,
		"")
	if err != nil {
		return nil, err
	}
	volume.Spec.DataProtection = true
	volume.Spec.UnixPermission = ""
	return volume, nil
}

func isValidUnixPermission(permission string) bool {
	// checking if length of permission is == quarkconfig.StandardUnixPermissionLength and all the characters when converted to int are octal
	if len(permission) != quarkconfig.StandardUnixPermissionLength {
		return false
	} else {
		for i := 0; i < len(permission); i++ {
			_, err := strconv.ParseInt(string(permission[i]), 8, 32) // checking for octal(base 8) digits
			if err != nil {
				return false
			}
		}
	}
	return true
}

// Constructs NetAppVolume definition.
// ownerID: OwnerID of the volume.
// dispalyName: Name of the volume provided by customer.
// sizeString: Volume Size.
// volumeUUID: Unique ID of volume in k8s cluster.
// volumePath: Volume mount path.
// serviceLevel: Standard for alpha.
// exportPolicy: Export Policy object
// snapshotPolicy: Snapshot Policy object
// unixPermission: unix volume permission
func ConstructNetAppVolume(ownerID, displayName, sizeString, volumeUUID, storagePoolUUID, volumePath, exportPolicyName string,
	snapshotPolicyName string, serviceLevel ServiceLevel, noSnapDir bool, unixPermission string) (volume *quarkv1alpha1.NetAppVolume, err error) {
	err = nvc.ValidateNFSConfiguration(volumePath)
	if err != nil {
		return nil, err
	}
	volume, err = constructBaseNetAppVolume(ownerID, displayName, sizeString, volumeUUID, storagePoolUUID, volumePath, snapshotPolicyName, serviceLevel, noSnapDir)
	if err != nil {
		return nil, err
	}
	volume.Spec.ExportPolicy = exportPolicyName
	volume.Spec.Protocol = quarkconfig.DefaultNFS3Protocol

	if unixPermission == "" {
		// 0777 will be the default permission in create workflow.
		// For clone create & snapshot cloud restore, unixPermission will be internally set by fetching it from DMAP.
		unixPermission = quarkconfig.DefaultUnixPermission
	} else {
		isValid := isValidUnixPermission(unixPermission)
		if !isValid {
			return nil, NewNVCClientConflictError(fmt.Sprintf("failed to update the permission for volume %s: %s", volume.Name, "Invalid Unix Permission"))
		}
	}
	volume.Spec.UnixPermission = unixPermission
	return volume, err
}

func (client *NVCClient) setVolumeProtocol(protocols string) ([]string, error) {
	protocolArr := strings.Split(protocols, ",")
	if len(protocolArr) > 2 {
		return nil, fmt.Errorf("More than 2 protocols not supported")
	}
	// Ensure leading/trailing spaces dont get in the way
	var protocolArrTrimmed = make([]string, len(protocolArr))
	for i := range protocolArr {
		protocolArrTrimmed[i] = strings.TrimSpace(protocolArr[i])
	}

	var convertedProtocolArr []string
	nfsVariations := []string{quarkconfig.NFS41Protocol, quarkconfig.NFSV41Protocol, quarkconfig.NFS4Protocol, quarkconfig.NFSV4Protocol}
	for _, protocol := range protocolArrTrimmed {
		protocol = strings.ToLower(protocol)
		if util.StringSetContains(nfsVariations, protocol) {
			convertedProtocolArr = append(convertedProtocolArr, quarkconfig.DefaultNFS4Protocol)
		} else if (protocol == quarkconfig.NFS40Protocol) || (protocol == quarkconfig.NFSV40Protocol) ||
			(protocol == quarkconfig.CIFSProtocol) || (protocol == quarkconfig.SMBProtocol) {
			return nil, fmt.Errorf("Invalid value of protocol %s", protocol)
		} else {
			convertedProtocolArr = append(convertedProtocolArr, quarkconfig.DefaultNFS3Protocol)
		}
	}
	convertedProtocolArr = util.RemoveDuplicates(convertedProtocolArr)
	return convertedProtocolArr, nil
}

// ConstructNetAppVolumeDP creates a NetAppVolume definition for DataProtection (DP) volume (Secondary volume)
func (client *NVCClient) ConstructNetAppVolumeDPNFS4(ownerID, displayName, sizeString, volumeUUID, storagePoolUUID, volumePath, exportPolicyName, snapshotPolicyName string,
	serviceLevel ServiceLevel, noSnapDir bool, protocols string, kerbEnabled bool) (volume *quarkv1alpha1.NetAppVolume, err error) {
	volume, err = client.ConstructNetAppVolumeNFS4(ownerID,
		displayName,
		sizeString,
		volumeUUID,
		storagePoolUUID,
		volumePath,
		exportPolicyName,
		snapshotPolicyName,
		serviceLevel,
		noSnapDir,
		"",
		protocols,
		kerbEnabled)
	if err != nil {
		return nil, err
	}
	volume.Spec.DataProtection = true
	volume.Spec.UnixPermission = ""
	return volume, nil
}

func (client *NVCClient) ConstructNetAppVolumeNFS4(ownerID, displayName, sizeString, volumeUUID, storagePoolUUID, volumePath, exportPolicyName, snapshotPolicyName string, serviceLevel ServiceLevel, noSnapDir bool, unixPermission, protocols string, kerbEnabled bool) (volume *quarkv1alpha1.NetAppVolume, err error) {
	if !client.EnableNfs41 {
		return nil, fmt.Errorf("NFS4.1 not enabled")
	}

	if kerbEnabled && !client.EnableCCPd {
		return nil, fmt.Errorf("CCPd not enabled for quark.")
	}

	err = nvc.ValidateNFSConfiguration(volumePath)
	if err != nil {
		return nil, err
	}
	volume, err = constructBaseNetAppVolume(ownerID, displayName, sizeString, volumeUUID, storagePoolUUID, volumePath, snapshotPolicyName, serviceLevel, noSnapDir)
	if err != nil {
		return nil, err
	}
	volume.Spec.ExportPolicy = exportPolicyName
	protocolArr, err := client.setVolumeProtocol(protocols)
	if err != nil {
		return nil, err
	}
	if len(protocolArr) == 1 {
		volume.Spec.Protocol = protocolArr[0]
	} else if len(protocolArr) == 2 {
		volume.Spec.Protocol = protocolArr[0] + "," + protocolArr[1]
	} else {
		volume.Spec.Protocol = quarkconfig.DefaultNFS4Protocol
	}
	if unixPermission == "" {
		// 0777 will be the default permission in create workflow.
		// For clone create & snapshot cloud restore, unixPermission will be internally set by fetching it from DMAP.
		unixPermission = quarkconfig.DefaultUnixPermission
	} else {
		isValid := isValidUnixPermission(unixPermission)
		if !isValid {
			return nil, NewNVCClientConflictError(fmt.Sprintf("failed to update the permission for volume %s: %s", volume.Name, "Invalid Unix Permission"))
		}
	}
	volume.Spec.UnixPermission = unixPermission

	if kerbEnabled {
		// Kerberos can only be enabled with NFS4.1 volumes
		if !util.ContainsNFSv4(strings.Split(protocols, ",")) || util.ContainsBothNFS(strings.Split(protocols, ",")) {
			return nil, NewNVCClientConflictError(fmt.Sprintf("failed to create volume %s: %s", volume.Name,
				"kerberos not allowed with NFSv3,SMB,both(NFSv3,NFSv4.1) type volumes"))
		}
		volume.Spec.KerberosEnabled = true
	}
	return volume, err
}

// validate the protocols are same in export policy and volume
func (client *NVCClient) ValidateProtocolMatch(exportPolicyName string, volType []string) bool {
	exportPolicy, err := client.NetAppExportPolicyGet(exportPolicyName)
	if err != nil {
		log.Fatal("Failed to get export policy", err)
	}
	rules := exportPolicy.Spec.Rules

	// since we allow creation of empty export policies
	if rules == nil || len(rules) == 0 {
		return true
	}

	for _, rl := range rules {
		// if protocol type is empty, we allow creation of volume
		if rl.Protocol == "" {
			return true
		}
		if reflect.DeepEqual([]string{rl.Protocol}, volType) {
			return true
		}
	}
	return false
}

func (client *NVCClient) NetAppVolumeNoSnapDirUpdate(vol *quarkv1alpha1.NetAppVolume, val bool) (volume *quarkv1alpha1.NetAppVolume, err error) {
	vol.SetNoSnapDir(val)
	newVol, err := client.netAppVolumeUpdate(vol)

	return newVol, err

}

func (client *NVCClient) NetAppVolumeSnapshotPolicyUpdate(vol *quarkv1alpha1.NetAppVolume, snapshotPolicy *quarkv1alpha1.NetAppSnapshotPolicy) (volume *quarkv1alpha1.NetAppVolume, err error) {
	if snapshotPolicy == nil {
		return nil, fmt.Errorf("snapshot policy can not be nil")
	}
	vol.Spec.SnapshotPolicy = snapshotPolicy.Name
	newVol, err := client.netAppVolumeUpdate(vol)

	return newVol, err
}

func (client *NVCClient) NetAppVolumeExportPolicyUpdate(vol *quarkv1alpha1.NetAppVolume, exportPolicy *quarkv1alpha1.NetAppExportPolicy) (volume *quarkv1alpha1.NetAppVolume, err error) {
	if exportPolicy == nil {
		return nil, fmt.Errorf("export policy can not be nil")
	}

	if client.EnableNfs41 && vol.Spec.Protocol == quarkconfig.DefaultNFS4Protocol {
		isValidProtocol := client.ValidateProtocolMatch(exportPolicy.Name, []string{vol.Spec.Protocol})
		// protocol of the export policy and volume should match before update
		if !isValidProtocol {
			return nil, NewNVCClientConflictError(fmt.Sprintf("failed to update the export policy for volume %s: %s", vol.Name, "Protocol mismatch"))
		}
	}
	vol.Spec.ExportPolicy = exportPolicy.Name
	newVol, err := client.netAppVolumeUpdate(vol)
	return newVol, err

}

func (client *NVCClient) NetAppVolumePermissionUpdate(vol *quarkv1alpha1.NetAppVolume, newUnixPermission string) (volume *quarkv1alpha1.NetAppVolume, err error) {

	if vol.Spec.Protocol == "cifs" {
		return nil, NewNVCClientConflictError(fmt.Sprintf("permission update not allowed for CIFS/SMB volume : %s", vol.Name))
	}
	// permission in DP volumes cannot be modified
	if vol.Spec.DataProtection {
		return nil, NewNVCClientConflictError(fmt.Sprintf("failed to update the permission for volume %s: %s", vol.Name, "Modification of unix permission not allowed for DP volume"))
	}

	isValid := isValidUnixPermission(newUnixPermission)
	if !isValid {
		return nil, NewNVCClientConflictError(fmt.Sprintf("failed to update the permission for volume %s: %s", vol.Name, "Invalid Unix Permission"))
	}

	if vol.Spec.UnixPermission == newUnixPermission {
		return nil, NewNVCClientConflictError(fmt.Sprintf("New permission same as existing permission for volume %s", vol.Name))
	}

	// Check if the volume is online. If not return
	if !client.NetAppVolumeOnline(vol) {
		return nil, NewNVCClientConflictError(fmt.Sprintf("failed to update the permission for volume %s: %s", vol.Name, quarkconfig.MessageVolumeOffline))
	}

	// block vol permission update during upgrade
	if client.isVolUpgradeInProgess(vol) {
		return nil, NewNVCClientConflictError(fmt.Sprintf("failed to update the permission for volume %s: %s", vol.Name, quarkconfig.MessageUpgradeInProgress))
	}

	vol.Spec.UnixPermission = newUnixPermission
	newVol, err := client.netAppVolumeUpdate(vol)

	return newVol, err
}

func (client *NVCClient) NetAppVolumeSizeUpdate(vol *quarkv1alpha1.NetAppVolume, newSizeString string) (volume *quarkv1alpha1.NetAppVolume, err error) {
	// block vol resize during upgrade
	if client.isVolUpgradeInProgess(vol) {
		return nil, NewNVCClientConflictError(fmt.Sprintf("failed to update the size for volume %s: %s", vol.Name, quarkconfig.MessageUpgradeInProgress))
	}
	// Validate size
	size, err := resource.ParseQuantity(newSizeString)
	if err != nil {
		return nil, fmt.Errorf("invalid value for size")
	}
	var minSize resource.Quantity
	if _, ok := vol.Labels[quarkconfig.StoragePoolLabel]; !ok {
		minSize = minVolSize
	} else {
		minSize = minSpVolSize
	}

	var minVolSizeDifference = size.Value() - minSize.Value()
	var maxVolSizeDifference = size.Value() - maxVolSize.Value()
	if minVolSizeDifference < 0 || maxVolSizeDifference > 0 || minVolSizeDifference%incrementalSize.Value() != 0 {
		return nil, fmt.Errorf("size=%s, volume size should be a minimum of %s, maximum of %s and a multiple of %s", size.String(), minSize.String(), maxVolSize.String(), incrementalSize.String())
	}
	if _, ok := vol.Labels[quarkconfig.StoragePoolLabel]; ok {
		availableSpace, err := client.NetAppStoragePoolGetAvailableSpace(vol.Labels[quarkconfig.StoragePoolLabel])
		if err != nil {
			return nil, fmt.Errorf("getting free space from storagePool failed: %s", err.Error())
		}
		if size.Value() > vol.Status.Size.Value() && availableSpace < size.Value()-vol.Status.Size.Value() {
			return nil, fmt.Errorf("insufficient space left in storagePool, free space=%d, requested=%d", availableSpace, size.Value()-vol.Status.Size.Value())
		}
	}
	vol.Spec.Size = size

	newVol, err := client.netAppVolumeUpdate(vol)

	return newVol, err
}

// Update Share properties for SMB volume
func (client *NVCClient) NetAppVolumeSMBSharePropertiesUpdate(vol *quarkv1alpha1.NetAppVolume, shareProperties NetAppVolumeSMBShareProperties) (volume *quarkv1alpha1.NetAppVolume, err error) {
	volume, err = client.NetAppVolumeGet(vol.Name)
	if err != nil {
		return nil, err
	}
	volume.Spec.SMBConfiguration.EncryptionEnabled = shareProperties.EncryptionEnabled
	volume.Spec.SMBConfiguration.HideShare = shareProperties.HideShare
	volume.Spec.SMBConfiguration.ShareAccessBasedEnumeration = shareProperties.AccessBasedEnumeration
	volume, err = client.netAppVolumeUpdate(volume)

	return
}

// Abort NetAppVolumeSMBSharePropertiesUpdate() operation
func (client *NVCClient) NetAppVolumeSMBSharePropertiesUpdateAbort(vol *quarkv1alpha1.NetAppVolume) (*quarkv1alpha1.NetAppVolume, error) {
	oldShareProperties := NetAppVolumeSMBShareProperties{
		EncryptionEnabled:      vol.Status.SMBConfiguration.EncryptionEnabled,
		HideShare:              vol.Status.SMBConfiguration.HideShare,
		AccessBasedEnumeration: vol.Status.SMBConfiguration.ShareAccessBasedEnumeration,
	}
	return client.NetAppVolumeSMBSharePropertiesUpdate(vol, oldShareProperties)
}

// Reset Machine Account password for SMB volume
func (client *NVCClient) NetAppVolumeSMBPasswordReset(volumeUUID string, user, password string) (volume *quarkv1alpha1.NetAppVolume, err error) {
	volume, err = client.NetAppVolumeGet(volumeUUID)
	if err != nil {
		return nil, err
	}
	secretName := util.GetVolPwdResetADSecretName(volume.Name)
	_, err = client.createOrUpdateADSecretWithOwnerRef(volume, nil, secretName, user, password)
	if err != nil {
		log.WithFields(log.Fields{
			"volume": volume.Name,
			"secret": secretName,
		}).Error("Failed to create secret.")
		return nil, err
	}

	volume.Spec.SMBPasswordResetCount = volume.Spec.SMBPasswordResetCount + 1
	newVolume, err := client.netAppVolumeUpdate(volume)
	if err != nil {
		delErr := client.SecretDelete(secretName)
		if delErr != nil {
			log.WithFields(log.Fields{
				"volume": volume.Name,
				"secret": secretName,
				"error":  delErr,
			}).Error("Failed to delete secret.")
		}

		return nil, err
	}
	return newVolume, err
}

// Abort machine-account password reset operation for SMB volume.
// SDE should abort the password reset operation if it times out.
// Abort will delete the secret and avoid retrying operation.
func (client *NVCClient) NetAppVolumeSMBPasswordResetAbort(volumeUUID string) (volume *quarkv1alpha1.NetAppVolume, err error) {
	secretName := util.GetVolPwdResetADSecretName(volumeUUID)
	err = client.SecretDelete(secretName)
	if err != nil && !errors.IsNotFound(err) {
		return nil, err
	}
	volume, err = client.NetAppVolumeGet(volumeUUID)
	if err != nil {
		return nil, err
	}
	volume.Spec.SMBPasswordResetCount = volume.Status.SMBPasswordResetCount
	volume, err = client.netAppVolumeUpdate(volume)

	return volume, err
}

// Update DNS IP for SMB volume
func (client *NVCClient) NetAppVolumeSMBDNSUpdate(vol *quarkv1alpha1.NetAppVolume, dnsIPs []string) (volume *quarkv1alpha1.NetAppVolume, err error) {
	// block SMB vol DNS update during upgrade
	if client.isVolUpgradeInProgess(vol) {
		return nil, NewNVCClientConflictError(fmt.Sprintf("failed to update DNS IPs for volume %s: %s", vol.Name, quarkconfig.MessageUpgradeInProgress))
	}
	volume, err = client.NetAppVolumeGet(vol.Name)
	if err != nil {
		return nil, err
	}
	volume.Spec.SMBConfiguration.DNSIPs = dnsIPs
	volume, err = client.netAppVolumeUpdate(volume)

	return
}

func (client *NVCClient) NetAppVolumeSMBDNSUpdateAbort(vol *quarkv1alpha1.NetAppVolume) (volume *quarkv1alpha1.NetAppVolume, err error) {
	volume, err = client.NetAppVolumeGet(vol.Name)
	if err != nil {
		return nil, err
	}
	volume.Spec.SMBConfiguration.DNSIPs = volume.Status.SMBConfiguration.DNSIPs
	volume, err = client.netAppVolumeUpdate(volume)

	return
}

// Update EncryptedADCommunication for SMB volume
func (client *NVCClient) NetAppVolumeSMBEncryptedADCommunicationUpdate(vol *quarkv1alpha1.NetAppVolume, encryptedADCommunication bool) (volume *quarkv1alpha1.NetAppVolume, err error) {
	// block AD field update during upgrade
	if client.isVolUpgradeInProgess(vol) {
		return nil, NewNVCClientConflictError(fmt.Sprintf("failed to update EncryptedADCommunication: %s", quarkconfig.MessageUpgradeInProgress))
	}
	volume, err = client.NetAppVolumeGet(vol.Name)
	if err != nil {
		return nil, err
	}
	volume.Spec.SMBConfiguration.EncryptedADCommunication = encryptedADCommunication
	volume, err = client.netAppVolumeUpdate(volume)

	return
}

func (client *NVCClient) NetAppVolumeSMBEncryptedADCommunicationUpdateAbort(vol *quarkv1alpha1.NetAppVolume) (volume *quarkv1alpha1.NetAppVolume, err error) {
	volume, err = client.NetAppVolumeGet(vol.Name)
	if err != nil {
		return nil, err
	}
	volume.Spec.SMBConfiguration.EncryptedADCommunication = volume.Status.SMBConfiguration.EncryptedADCommunication
	volume, err = client.netAppVolumeUpdate(volume)

	return
}

// isKerberosRulePresent checks if a Kerberos rule is present in the given ExportRule slice.
func isKerberosRulePresent(rules []ExportRule) bool {
	for _, rl := range rules {
		if rl.SecurityType == quarkconfig.SecurityTypeKerberos {
			return true
		}
	}
	return false
}

func validateExportRules(name string, rl ExportRule, i int, kerberosRulePresent bool) error {
	// TODO: Validate NetAppExpolicy Name if we are following any naming convention

	validProtocols := []string{quarkconfig.DefaultNFS3Protocol, "cifs", quarkconfig.DefaultNFS4Protocol}
	validSecurityType := []string{"sys", "any", "never", quarkconfig.SecurityTypeKerberos}
	validRootAccess := []string{"on", "off", ""}
	validKerberosSecurityTypes := []string{"krb5", "krb5i", "krb5p"}

	if !util.StringSetContains(validSecurityType, rl.SecurityType) {
		return fmt.Errorf("exportPolicy: %s: invalid security type %s for rule %d", name, rl.SecurityType, i)
	}

	protocolArr := strings.Split(rl.Protocol, ",")
	for protocolValues := range protocolArr {
		if !util.StringSetContains(validProtocols, protocolArr[protocolValues]) {
			return fmt.Errorf("exportPolicy: %s: invalid protocol %s for rule %d", name, rl.Protocol, i)
		}
	}

	if rl.AccessMode != "rw" && rl.AccessMode != "ro" && rl.AccessMode != "none" {
		return fmt.Errorf("exportPolicy: %s: invalid access mode %s for rule %d", name, rl.AccessMode, i)
	}

	// Validating 'Clients' field
	clientsList := strings.Split(rl.Clients, ",")
	for _, client := range clientsList {
		client = strings.Trim(client, " ")
		if strings.HasSuffix(client, "/0") {
			if client != "0.0.0.0/0" {
				return fmt.Errorf("exportPolicy: %s: invalid clients %s: address bytes masked out by netmask are non zero for rule %d", name, rl.Clients, i)
			}
		}
	}

	// Validating 'rootAccess' field
	if !util.StringSetContains(validRootAccess, rl.RootAccess) {
		return fmt.Errorf("exportPolicy: %s: invalid rootAccess %s for rule %d", name, rl.RootAccess, i)
	}

	// kerberos security type should either be present in all the export policy rules or none of them.
	if rl.SecurityType != quarkconfig.SecurityTypeKerberos && kerberosRulePresent {
		return fmt.Errorf("exportPolicy: %s: cannot have both kerberos and non-kerberos security types in rules", name)
	}

	// kerberos security type is only applicable for 'nfsv4' protocols and 'none' access mode.
	if rl.SecurityType == quarkconfig.SecurityTypeKerberos {
		if util.ContainsNFSv3(protocolArr) {
			return fmt.Errorf("exportPolicy: %s: %s security type is applicable only for nfsv4 protocol for rule %d ", name, quarkconfig.SecurityTypeKerberos, i)
		}
		if rl.AccessMode != "none" {
			return fmt.Errorf("exportPolicy: %s: %s security type is applicable only for none access mode for rule %d ", name, quarkconfig.SecurityTypeKerberos, i)
		}

		for _, item := range rl.KerberosROSecurityType {
			if !util.StringSetContains(validKerberosSecurityTypes, item) {
				return fmt.Errorf("exportPolicy: %s: invalid kerberos RO security type %s for rule %d", name, item, i)
			}
		}
		for _, item := range rl.KerberosRWSecurityType {
			if !util.StringSetContains(validKerberosSecurityTypes, item) {
				return fmt.Errorf("exportPolicy: %s: invalid kerberos RW security type %s for rule %d", name, item, i)
			}
		}
	} else {
		if rl.AccessMode == "none" {
			return fmt.Errorf("exportPolicy: %s: none access mode is applicable only for %s security type for rule %d ", name, quarkconfig.SecurityTypeKerberos, i)
		}
	}

	return nil
}

func (client *NVCClient) ModifyNetAppExportPolicy(exportPolicy *quarkv1alpha1.NetAppExportPolicy, rules []ExportRule) (policy *quarkv1alpha1.NetAppExportPolicy, err error) {
	exportPolicyCopy := exportPolicy.DeepCopy()
	var policyRule quarkv1alpha1.ExportRule
	policyRuleList := []quarkv1alpha1.ExportRule{}

	if exportPolicy.Status.Rules == nil {
		return nil, fmt.Errorf("mutable export policy is not supported")
	}

	kerberosRulePresent := isKerberosRulePresent(rules)
	for i, rl := range rules {
		err1 := validateExportRules(exportPolicyCopy.Name, rl, i, kerberosRulePresent)
		if err1 != nil {
			return nil, err1
		}
		policyRule = quarkv1alpha1.ExportRule{
			RuleIndex:              rl.RuleIndex,
			Clients:                rl.Clients,
			AccessMode:             rl.AccessMode,
			SecurityType:           rl.SecurityType,
			KerberosROSecurityType: rl.KerberosROSecurityType,
			KerberosRWSecurityType: rl.KerberosRWSecurityType,
			Protocol:               rl.Protocol,
			RootAccess:             rl.RootAccess,
		}
		policyRuleList = append(policyRuleList, policyRule)
	}

	if len(exportPolicy.Spec.Rules) != 0 {
		oldRule := exportPolicy.Spec.Rules[0]

		// Cannot modify export policy with kerberos rules to have non-kerberos rules and vice versa.
		if oldRule.SecurityType == quarkconfig.SecurityTypeKerberos && !kerberosRulePresent {
			return nil, fmt.Errorf("cannot modify export policy with kerberos rules to have non-kerberos rules")
		} else if oldRule.SecurityType != quarkconfig.SecurityTypeKerberos && kerberosRulePresent {
			return nil, fmt.Errorf("cannot modify export policy with non-kerberos rules to have kerberos rules")
		}
	}

	exportPolicyCopy.Spec.Rules = policyRuleList
	newExportPolicy, err := client.NetAppExportPolicyModify(exportPolicyCopy)
	return newExportPolicy, err

}

// Construct NetAppExportPolicy definition
// name: Name should be unique. Suggested Naming convention is volName-revisionNumber.
// rules: ExportPolicy rule
func ConstructNetAppExportPolicy(name string, rules []ExportRule, volumeUUID string, storagePoolUUID string, ownerID string) (policy *quarkv1alpha1.NetAppExportPolicy, err error) {
	// validate volumeUUID
	_, err = uuid.FromString(volumeUUID)
	if err != nil {
		return nil, fmt.Errorf("invalid value for volumeUUID")
	}
	// validate storagePoolUUID
	/*	commenting out to support backward compatibility
		_, err = uuid.FromString(storagePoolUUID)
		if err != nil {
			return nil, fmt.Errorf("invalid value for storagePoolUUID")
		}
	*/
	policy = &quarkv1alpha1.NetAppExportPolicy{
		TypeMeta: metav1.TypeMeta{},
		ObjectMeta: metav1.ObjectMeta{
			Name: name,
			Labels: map[string]string{
				quarkconfig.VolumeLabel:  volumeUUID,
				quarkconfig.OwnerIDLabel: ownerID,
			},
		},
		Spec: quarkv1alpha1.NetAppExportPolicySpec{
			Rules: []quarkv1alpha1.ExportRule{},
		},
	}
	if storagePoolUUID != "" {
		policy.Labels[quarkconfig.StoragePoolLabel] = storagePoolUUID
	}
	var policyRule quarkv1alpha1.ExportRule
	kerberosRulePresent := isKerberosRulePresent(rules)

	for i, rl := range rules {
		err1 := validateExportRules(name, rl, i, kerberosRulePresent)
		if err1 != nil {
			return nil, err1
		}
		policyRule = quarkv1alpha1.ExportRule{
			RuleIndex:              rl.RuleIndex,
			Clients:                rl.Clients,
			AccessMode:             rl.AccessMode,
			SecurityType:           rl.SecurityType,
			Protocol:               rl.Protocol,
			KerberosROSecurityType: rl.KerberosROSecurityType,
			KerberosRWSecurityType: rl.KerberosRWSecurityType,
			RootAccess:             rl.RootAccess,
		}
		policy.Spec.Rules = append(policy.Spec.Rules, policyRule)
	}
	return
}

type CronSchedule quarkv1alpha1.CronSchedule

func validateTimeInput(policyName string, arr []int64, maxItems int, minVal int64, maxVal int64, item string) (err error) {
	if len(arr) > maxItems {
		return fmt.Errorf("snapshotPolicy: %s: more than %d items in the %s list not allowed", policyName, maxItems, item)
	}
	keys := make(map[int64]bool)
	for _, val := range arr {
		if _, value := keys[val]; !value {
			keys[val] = true
		} else {
			return fmt.Errorf("snapshotPolicy: %s: duplicate value in an array %s", policyName, item)
		}
		if val < minVal || val > maxVal {
			return fmt.Errorf("snapshotPolicy: %s: %d is an invalid value for %s, Valid values are %d-%d",
				policyName, val, item, minVal, maxVal)
		}
	}
	return nil
}

// Construct Netapp Snapshot Policy definition
// DaysOfWeek: Format Sunday -0,... ,Saturday - 6
func ConstructNetAppSnapshotPolicy(name string, isEnabled bool, schedules []CronSchedule, volumeUUID string, storagePoolUUID string, ownerID string) (policy *quarkv1alpha1.NetAppSnapshotPolicy, err error) {
	// validate volumeUUID
	_, err = uuid.FromString(volumeUUID)
	if err != nil {
		return nil, fmt.Errorf("invalid value for volumeUUID")
	}
	// validate storagePoolUUID
	/*	commenting out to support backward compatibility
		_, err = uuid.FromString(storagePoolUUID)
		if err != nil {
			return nil, fmt.Errorf("invalid value for storagePoolUUID")
		}
	*/
	policy = &quarkv1alpha1.NetAppSnapshotPolicy{
		TypeMeta: metav1.TypeMeta{},
		ObjectMeta: metav1.ObjectMeta{
			Name: name,
			Labels: map[string]string{
				quarkconfig.VolumeLabel:  volumeUUID,
				quarkconfig.OwnerIDLabel: ownerID,
			},
		},
		Spec: quarkv1alpha1.NetAppSnapshotPolicySpec{
			Schedules: []quarkv1alpha1.CronSchedule{},
		},
	}
	if storagePoolUUID != "" {
		policy.Labels[quarkconfig.StoragePoolLabel] = storagePoolUUID
	}

	policy.Spec.Enabled = isEnabled
	var policySchedule quarkv1alpha1.CronSchedule
	for _, sc := range schedules {
		if err = validateTimeInput(name, sc.Months, 12, 1, 12, "Months"); err != nil {
			return nil, err
		}
		if err = validateTimeInput(name, sc.DaysOfMonth, 31, 1, 31, "DaysOfMonth"); err != nil {
			return nil, err
		}
		if err = validateTimeInput(name, sc.DaysOfWeek, 7, 0, 6, "DaysOfWeek"); err != nil {
			return nil, err
		}
		if err = validateTimeInput(name, sc.Hours, 24, 0, 23, "Hours"); err != nil {
			return nil, err
		}
		if err = validateTimeInput(name, sc.Minutes, 60, 0, 59, "Minutes"); err != nil {
			return nil, err
		}

		policySchedule = quarkv1alpha1.CronSchedule{
			Prefix:          sc.Prefix,
			Months:          sc.Months,
			DaysOfMonth:     sc.DaysOfMonth,
			DaysOfWeek:      sc.DaysOfWeek,
			Hours:           sc.Hours,
			Minutes:         sc.Minutes,
			SnapshotsToKeep: sc.SnapshotsToKeep,
		}
		policy.Spec.Schedules = append(policy.Spec.Schedules, policySchedule)
	}
	return
}

func validateSnapshotDisplayName(displayName string, snapUUID string) error {

	// !!!! READ ME: These rules are what SDE uses to validate displayNames as supplied by the user. DO NOT change without coordindation with SDE !!!!
	if regexp.MustCompile(`^$`).MatchString(strings.TrimSpace(displayName)) {
		return fmt.Errorf("snapshot display name cannot be an empty string")
	}
	// name of length 0 is handled by previous check
	if len(displayName) > 255 {
		return fmt.Errorf("snapshot name should be 1-255 characters long, you supplied a name %d characters long", len(displayName))
	}
	nameRuleSDE := `^[\w()+.-]+$`
	if !regexp.MustCompile(nameRuleSDE).MatchString(displayName) {
		return fmt.Errorf("invalid snapshot name, must satisfy rule %v", nameRuleSDE)
	}
	if strings.Contains(displayName, "..") {
		return fmt.Errorf("invalid snapshot display name, cannot contain \"..\"")
	}
	if displayName == "." {
		return fmt.Errorf("invalid snapshot display name, using \".\" as name not allowed")
	}
	if regexp.MustCompile("^ref_ss_volmove.*$").MatchString(displayName) ||
		regexp.MustCompile("^snapmirror.*$").MatchString(displayName) {
		return fmt.Errorf("invalid snapshot name, cannot begin with \"snapmirror\" or \"ref_ss_volmove\"")
	}
	// END SDE supplied rules

	// Pre-existing rule
	if regexp.MustCompile("^cloud-snapshot.*$").MatchString(displayName) {
		return fmt.Errorf("invalid prefix for snapshot display name, cannot start with \"cloud-snapshot\"")
	}
	if strings.EqualFold(displayName, snapUUID) {
		return fmt.Errorf("snapshot display name cannot be the same as snapshot UUID")
	}
	return nil
}

// Construct Netapp Volume Snapshot definition
func ConstructNetAppVolumeSnapshot(ownerID, snapUUID, volumeUUID, storagePoolUUID, displayName string) (volumeSnapshot *quarkv1alpha1.NetAppVolumeSnapshot, err error) {

	// Trim whitespace from all incoming strings
	ownerID = strings.TrimSpace(ownerID)
	snapUUID = strings.TrimSpace(snapUUID)
	volumeUUID = strings.TrimSpace(volumeUUID)
	storagePoolUUID = strings.TrimSpace(storagePoolUUID)
	displayName = strings.TrimSpace(displayName)

	// validate volumeUUID
	_, err = uuid.FromString(volumeUUID)
	if err != nil {
		return nil, fmt.Errorf("invalid value for volumeUUID")
	}
	// validate storagePoolUUID
	/*	commenting out to support backward compatibility
		_, err = uuid.FromString(storagePoolUUID)
		if err != nil {
			return nil, fmt.Errorf("invalid value for storagePoolUUID")
		}
	*/

	// validate snapUUID
	_, err = uuid.FromString(snapUUID)
	if err != nil {
		return nil, fmt.Errorf("invalid value for snapUUID")
	}

	if err = validateSnapshotDisplayName(displayName, snapUUID); err != nil {
		return nil, err
	}

	volumeSnapshot = &quarkv1alpha1.NetAppVolumeSnapshot{
		TypeMeta: metav1.TypeMeta{},
		ObjectMeta: metav1.ObjectMeta{
			Name: snapUUID,
			Labels: map[string]string{
				quarkconfig.VolumeLabel:  volumeUUID,
				quarkconfig.OwnerIDLabel: ownerID,
			},
		},
		Spec: quarkv1alpha1.NetAppVolumeSnapshotSpec{
			DisplayName: displayName,
			Name:        snapUUID,
		},
	}
	if storagePoolUUID != "" {
		volumeSnapshot.Labels[quarkconfig.StoragePoolLabel] = storagePoolUUID
	}
	return
}

// Modify Netapp Snapshot Policy definition
// DaysOfWeek: Format Sunday -0,... ,Saturday - 6
// func (client *NVCClient) ModifyNetAppSnapshotPolicy(name string, isEnabled bool, schedules []CronSchedule) (*quarkv1alpha1.NetAppSnapshotPolicy, error) {
func (client *NVCClient) ModifyNetAppSnapshotPolicy(snapshotPolicy *quarkv1alpha1.NetAppSnapshotPolicy, isEnabled bool, schedules []CronSchedule) (*quarkv1alpha1.NetAppSnapshotPolicy, error) {
	var newSnapshotPolicy *quarkv1alpha1.NetAppSnapshotPolicy
	var policySchedule quarkv1alpha1.CronSchedule
	policyScheduleList := []quarkv1alpha1.CronSchedule{}

	if snapshotPolicy.Status.Schedules == nil {
		return nil, fmt.Errorf("mutable snapshot policy is not supported")
	}

	policy := snapshotPolicy.DeepCopy()

	_, err := client.NetAppVolumeGet(snapshotPolicy.Labels[quarkconfig.VolumeLabel])
	if err != nil {
		return nil, err
	}

	_, err = client.NetAppStoragePoolGet(snapshotPolicy.Labels[quarkconfig.StoragePoolLabel])
	if err != nil {
		return nil, err
	}

	for _, sc := range schedules {
		if err = validateTimeInput(policy.Name, sc.Months, 12, 1, 12, "Months"); err != nil {
			return nil, err
		}
		if err = validateTimeInput(policy.Name, sc.DaysOfMonth, 31, 1, 31, "DaysOfMonth"); err != nil {
			return nil, err
		}
		if err = validateTimeInput(policy.Name, sc.DaysOfWeek, 7, 0, 6, "DaysOfWeek"); err != nil {
			return nil, err
		}
		if err = validateTimeInput(policy.Name, sc.Hours, 24, 0, 23, "Hours"); err != nil {
			return nil, err
		}
		if err = validateTimeInput(policy.Name, sc.Minutes, 60, 0, 59, "Minutes"); err != nil {
			return nil, err
		}

		policySchedule = quarkv1alpha1.CronSchedule{
			Prefix:          sc.Prefix,
			Months:          sc.Months,
			DaysOfMonth:     sc.DaysOfMonth,
			DaysOfWeek:      sc.DaysOfWeek,
			Hours:           sc.Hours,
			Minutes:         sc.Minutes,
			SnapshotsToKeep: sc.SnapshotsToKeep,
		}
		policyScheduleList = append(policyScheduleList, policySchedule)
	}

	policy.Spec.Enabled = isEnabled
	policy.Spec.Schedules = policyScheduleList
	newSnapshotPolicy, err = client.NetAppSnapshotPolicyUpdate(policy)

	return newSnapshotPolicy, err
}
func ConstructNetAppVolumeCloudSnapshot(providerType quarkv1alpha1.NetAppVolumeCloudSnapshotProviderType, ownerID, cloudSnapshotUUID, volumeUUID, storagePoolUUID, displayName, container, server string, port int32, secretName, secretNamespace string) (volumeCloudSnapshot *quarkv1alpha1.NetAppVolumeCloudSnapshot, err error) {
	// validate volumeUUID
	_, err = uuid.FromString(volumeUUID)
	if err != nil {
		return nil, fmt.Errorf("invalid value for volumeUUID")
	}
	// validate cloudSnapshotUUID
	_, err = uuid.FromString(cloudSnapshotUUID)
	if err != nil {
		return nil, fmt.Errorf("invalid value for cloudSnapshotUUID")
	}

	volumeCloudSnapshot = &quarkv1alpha1.NetAppVolumeCloudSnapshot{
		TypeMeta: metav1.TypeMeta{},
		ObjectMeta: metav1.ObjectMeta{
			Name: cloudSnapshotUUID,
			Labels: map[string]string{
				quarkconfig.VolumeLabel:  volumeUUID,
				quarkconfig.OwnerIDLabel: ownerID,
			},
			Finalizers: []string{
				quarkconfig.CloudSnapshotFinalizer,
			},
		},
		Spec: quarkv1alpha1.NetAppVolumeCloudSnapshotSpec{
			ProviderType: providerType,
			DisplayName:  displayName,
			Imported:     false,
			Container:    container,
			Server:       server,
			Port:         port,
			SecretRef: corev1.SecretReference{
				Name:      secretName,
				Namespace: secretNamespace,
			},
		},
	}
	if storagePoolUUID != "" {
		volumeCloudSnapshot.Labels[quarkconfig.StoragePoolLabel] = storagePoolUUID
	}
	return
}

func (client *NVCClient) NetAppVolumeCloudSnapshotCreate(volumeCloudSnapshot *quarkv1alpha1.NetAppVolumeCloudSnapshot) (*quarkv1alpha1.NetAppVolumeCloudSnapshot, error) {
	metaLabels := volumeCloudSnapshot.ObjectMeta.GetLabels()
	volumeUUID := metaLabels[quarkconfig.VolumeLabel]
	ntapVol, err := client.quarkClient.QuarkV1alpha1().NetAppVolumes(client.nvcNamespace).Get(
		context.TODO(), volumeUUID, metav1.GetOptions{})
	if err != nil {
		return nil, err
	}
	// Block cloud snapshot creation if the volume is being deleted
	if ntapVol.ObjectMeta.DeletionTimestamp != nil {
		return nil, NewNVCClientConflictError(fmt.Sprintf("cannot create cloud snapshot for volume '%s', volume deletion in progress", volumeUUID))
	}

	// Block cloud snapshot creation if the volume is a clone volume and split scan is still in progress
	// This is a best case effort, SDE will block such operations on their end as well
	_, cloneLabelExists := ntapVol.Labels[quarkconfig.VolumeCloneParentUUUIDLabel]
	if ntapVol.Spec.CloneParentSnapshot != "" || ntapVol.Spec.CloneParentVolumeUUID != "" || cloneLabelExists {
		return nil, CloneVolumeSplitScanInProgressError{CloneList: []quarkv1alpha1.NetAppVolume{*ntapVol}}

	}
	if volumeCloudSnapshot.Spec.Imported {
		return nil, NewNVCClientFatalError("cannot create imported cloud snapshot")
	}
	return client.quarkClient.QuarkV1alpha1().NetAppVolumeCloudSnapshots(client.nvcNamespace).Create(
		context.TODO(), volumeCloudSnapshot, metav1.CreateOptions{})
}

func (client *NVCClient) getCloudSnapshot(cloudSnapshotUUID string) (*quarkv1alpha1.NetAppVolumeCloudSnapshot, error) {
	return client.quarkClient.QuarkV1alpha1().NetAppVolumeCloudSnapshots(client.nvcNamespace).Get(
		context.TODO(), cloudSnapshotUUID, metav1.GetOptions{})
}

func cloudSnapshotReady(cloudSnapshot *quarkv1alpha1.NetAppVolumeCloudSnapshot) bool {
	if cloudSnapshot.Spec.Imported {
		return true
	}
	conditions := cloudSnapshot.Status.Conditions
	for i := range conditions {
		if conditions[i].Type == quarkv1alpha1.CloudSnapshotReady {
			return conditions[i].Status == corev1.ConditionTrue
		}
	}
	return false
}

func cloudSnapshotUploading(cloudSnapshot *quarkv1alpha1.NetAppVolumeCloudSnapshot) bool {
	if cloudSnapshot.Spec.Imported {
		return false
	}
	conditions := cloudSnapshot.Status.Conditions
	for i := range conditions {
		if conditions[i].Type == quarkv1alpha1.CloudSnapshotCreating {
			return conditions[i].Status == corev1.ConditionTrue
		}
	}
	return false
}

func cloudSnapshotDeleting(cloudSnapshot *quarkv1alpha1.NetAppVolumeCloudSnapshot) bool {
	return (cloudSnapshot.ObjectMeta.DeletionTimestamp != nil)
}

func cloudSnapshotError(cloudSnapshot *quarkv1alpha1.NetAppVolumeCloudSnapshot) error {
	conditions := cloudSnapshot.Status.Conditions
	for _, cond := range conditions {
		if cond.Type == quarkv1alpha1.CloudSnapshotError &&
			cond.Status == corev1.ConditionTrue {
			return fmt.Errorf(cond.Message)
		}
	}
	return nil
}

func (client *NVCClient) NetAppVolumeCloudSnapshotReady(cloudSnapshotUUID string) bool {
	cloudSnapshotInfo, err := client.NetAppVolumeCloudSnapshotGet(cloudSnapshotUUID)
	if err != nil {
		return false
	}

	return cloudSnapshotInfo.Ready
}

func ConstructNetAppVolumeCloudSnapshotImport(providerType quarkv1alpha1.NetAppVolumeCloudSnapshotProviderType, ownerID, cloudSnapshotUUID, volumeUUID, storagePoolUUID, displayName, container, server string, port int32, secretName, secretNamespace, endpointUUID string, logicalSize int64, creationTime time.Time) (volumeCloudSnapshot *quarkv1alpha1.NetAppVolumeCloudSnapshot, err error) {
	// validate endpointUUID
	if endpointUUID == "" {
		return nil, fmt.Errorf("endpointUUID not provided")
	}

	volumeCloudSnapshot, err = ConstructNetAppVolumeCloudSnapshot(providerType, ownerID, cloudSnapshotUUID, volumeUUID, storagePoolUUID, displayName, container, server, port, secretName, secretNamespace)
	if err != nil {
		return
	}

	volumeCloudSnapshot.Spec.Imported = true

	// XXX Client shouldn't be setting status. We'll handle populating the status in the C2C controller. This info can be placed in the spec or perhaps stored in the bucket for the controller to pull down.
	volumeCloudSnapshot.Status = quarkv1alpha1.NetAppVolumeCloudSnapshotStatus{
		EndpointUUID: endpointUUID,
		LogicalSize:  logicalSize,
		CreationTime: metav1.NewTime(creationTime),
	}
	return
}

func (client *NVCClient) NetAppVolumeCloudSnapshotImport(volumeCloudSnapshot *quarkv1alpha1.NetAppVolumeCloudSnapshot) (*quarkv1alpha1.NetAppVolumeCloudSnapshot, error) {
	if !volumeCloudSnapshot.Spec.Imported {
		return nil, fmt.Errorf("cannot import created cloud snapshot")
	}

	newCloudSnap, err := client.quarkClient.QuarkV1alpha1().NetAppVolumeCloudSnapshots(client.nvcNamespace).Create(
		context.TODO(), volumeCloudSnapshot, metav1.CreateOptions{})

	if err != nil {
		return nil, err
	}

	newCloudSnap.Status = volumeCloudSnapshot.Status

	return client.quarkClient.QuarkV1alpha1().NetAppVolumeCloudSnapshots(client.nvcNamespace).UpdateStatus(
		context.TODO(), newCloudSnap, metav1.UpdateOptions{})
}

func (client *NVCClient) NetAppVolumeCloudSnapshotRestore(volume *quarkv1alpha1.NetAppVolume, cloudSnapshotUUID string) (*quarkv1alpha1.NetAppVolume, error) {

	cloudSnapshot, err := client.getCloudSnapshot(cloudSnapshotUUID)
	if err != nil {
		return nil, err
	}

	if !cloudSnapshotReady(cloudSnapshot) {
		return nil, fmt.Errorf("cannot restore, CloudSnapshot not ready: %s", cloudSnapshotUUID)
	}

	snapSize := cloudSnapshot.Status.LogicalSize
	volSize := volume.Spec.Size.Value()

	// Volume must be at least as large as the snapshot
	if volSize < snapSize {
		return nil, fmt.Errorf("cannot restore, volume size must be at least %d bytes", snapSize)
	}

	// restored volume's permission should match with the snapshot's permission, which will be done after volume creation/restoration in podrick
	volume.Spec.UnixPermission = ""
	volume.Spec.CloudSnapshot = cloudSnapshotUUID

	return client.NetAppVolumeCreate(volume)
}

func (client *NVCClient) NetAppVolumeCloudSnapshotRestoreSMB(volume *quarkv1alpha1.NetAppVolume, cloudSnapshotUUID, user, password string) (*quarkv1alpha1.NetAppVolume, error) {

	cloudSnapshot, err := client.getCloudSnapshot(cloudSnapshotUUID)
	if err != nil {
		return nil, err
	}

	if !cloudSnapshotReady(cloudSnapshot) {
		return nil, fmt.Errorf("cannot restore, CloudSnapshot not ready: %s", cloudSnapshotUUID)
	}

	snapSize := cloudSnapshot.Status.LogicalSize
	volSize := volume.Spec.Size.Value()

	// Volume must be at least as large as the snapshot
	if volSize < snapSize {
		return nil, fmt.Errorf("cannot restore, volume size must be at least %d bytes", snapSize)
	}

	volume.Spec.CloudSnapshot = cloudSnapshotUUID

	return client.NetAppVolumeCreateSMB(volume, user, password)
}

// NetAppVolumeCloudSnapshotRestored returns "true" when the cloud snapshot is restored for a given volume.
func (client *NVCClient) NetAppVolumeCloudSnapshotRestored(volume *quarkv1alpha1.NetAppVolume) bool {

	cloudSnapUUID := volume.Spec.CloudSnapshot
	if cloudSnapUUID != "" {
		// check for cloud snap restored condition to be true
		for _, condition := range volume.Status.Conditions {
			if condition.Type == quarkv1alpha1.NetAppVolumeCloudSnapshotRestored {
				return condition.Status == corev1.ConditionTrue
			}
		}
	}
	return false
}

func (client *NVCClient) NetAppVolumeCloudSnapshotDelete(cloudSnapshotUUID string) error {
	cloudSnapshot, err := client.getCloudSnapshot(cloudSnapshotUUID)
	if err != nil {
		return err
	}
	metaLabels := cloudSnapshot.ObjectMeta.GetLabels()
	volumeUUID := metaLabels[quarkconfig.VolumeLabel]
	for {
		ntapVol, err := client.quarkClient.QuarkV1alpha1().NetAppVolumes(client.nvcNamespace).Get(
			context.TODO(), volumeUUID, metav1.GetOptions{})
		if err != nil {
			if errors.IsNotFound(err) {
				break
			}
			log.WithFields(log.Fields{
				"volume":        volumeUUID,
				"cloudsnapshot": cloudSnapshotUUID,
				"error":         err,
			}).Error("Error getting NetappVolume associated with cloud snapshot.")
			return err
		}
		if ntapVol != nil && ntapVol.DeletionTimestamp == nil {
			break
		}
		waitTime := 2 * time.Second
		log.WithFields(log.Fields{
			"volume":        volumeUUID,
			"cloudsnapshot": cloudSnapshotUUID,
			"waittime":      waitTime,
		}).Info("Waiting for NetAppVolume deletion before deleting cloud snapshot.")
		time.Sleep(waitTime)
	}
	return client.quarkClient.QuarkV1alpha1().NetAppVolumeCloudSnapshots(client.nvcNamespace).Delete(
		context.TODO(), cloudSnapshotUUID, metav1.DeleteOptions{})
}

func (client *NVCClient) NetAppVolumeCloudSnapshotDeleted(cloudSnapshotUUID string) bool {
	_, err := client.getCloudSnapshot(cloudSnapshotUUID)
	if err != nil {
		if errors.IsNotFound(err) {
			return true
		} else {
			log.WithFields(log.Fields{
				"error": err,
			}).Debug("getCloudSnapshot failed.")
		}
	}
	return false
}

type NetAppVolumeCloudSnapshotInfo struct {
	CloudSnapshotUUID string
	DisplayName       string
	OwnerID           string
	VolumeUUID        string
	Ready             bool
	Uploading         bool
	Deleting          bool
	EndpointUUID      string
	BytesTransferred  int64
	LogicalSize       int64
	CompletionPercent int64
	Error             error
	// Timestamps with location set to UTC.
	LastTransitionTime *time.Time
	LastHeartbeatTime  *time.Time
	CreationTime       *time.Time
}

func getCloudSnapshotInfo(cloudSnapshot *quarkv1alpha1.NetAppVolumeCloudSnapshot) NetAppVolumeCloudSnapshotInfo {
	cloudSnapInfo := NetAppVolumeCloudSnapshotInfo{
		CloudSnapshotUUID: cloudSnapshot.Name,
		DisplayName:       cloudSnapshot.Spec.DisplayName,
		OwnerID:           cloudSnapshot.Labels[quarkconfig.OwnerIDLabel],
		VolumeUUID:        cloudSnapshot.Labels[quarkconfig.VolumeLabel],
		Ready:             cloudSnapshotReady(cloudSnapshot),
		Uploading:         cloudSnapshotUploading(cloudSnapshot),
		Deleting:          cloudSnapshotDeleting(cloudSnapshot),
		EndpointUUID:      cloudSnapshot.Status.EndpointUUID,
		BytesTransferred:  cloudSnapshot.Status.BytesTransferred,
		LogicalSize:       cloudSnapshot.Status.LogicalSize,
		CompletionPercent: cloudSnapshot.Status.CompletionPercent,
		Error:             cloudSnapshotError(cloudSnapshot),
	}

	if len(cloudSnapshot.Status.Conditions) > 0 {
		// There should only be 0 or 1 conditions
		transitionUTC := cloudSnapshot.Status.Conditions[0].LastTransitionTime.UTC()
		cloudSnapInfo.LastTransitionTime = &transitionUTC
		heartbeatUTC := cloudSnapshot.Status.Conditions[0].LastHeartbeatTime.UTC()
		cloudSnapInfo.LastHeartbeatTime = &heartbeatUTC
	}

	if !cloudSnapshot.Status.CreationTime.IsZero() {
		creationUTC := cloudSnapshot.Status.CreationTime.UTC()
		cloudSnapInfo.CreationTime = &creationUTC
	}

	return cloudSnapInfo
}

func getNetAppQuarkStatus(quark *quarkv1alpha1.Quark) NetAppQuarkUpdateStatus {
	ntapQuarkStatus := NetAppQuarkUpdateStatus{
		OperatorReady:                false,
		VolumeControllerReady:        false,
		ServiceMeshControllerReady:   false,
		HighAvailabilityReady:        false,
		ExternalVolumeAccessReady:    false,
		CloudsnapshotCapabilityReady: false,
		IPDReady:                     false,
		EtcdClusterReady:             false,
		NodevolReady:                 false,
		NodeMonitorReady:             false,
		OtelCollectorReady:           false,
	}

	// Check if quark is reconciled, if in progress return false as the statuses might not be accurate
	quarkReconciled := (quark.Spec.Version == quark.Status.Version) && reflect.DeepEqual(quark.Spec.Images, quark.Status.Images)

	// Return false status if quark has not reconciled
	if !quarkReconciled {
		return ntapQuarkStatus
	}

	if quarkReconciled &&
		quark.Status.Health.Operator == quarkv1alpha1.QuarkReadyStatusReady {
		ntapQuarkStatus.OperatorReady = true
	}

	if quarkReconciled &&
		quark.Status.Health.NVC == quarkv1alpha1.QuarkReadyStatusReady {
		ntapQuarkStatus.VolumeControllerReady = true
	}

	if quarkReconciled {
		// If service mesh is not enabled, default controllerready to True
		// so as to not break any workflows.
		// TODO: This is temporary. Eventually service mesh controller will
		// be a stable part of Quark stack, and we won't need to worry about
		// EnableServiceMesh feature flag.
		if !quark.Spec.EnableServiceMesh || quark.Status.Health.SvcMeshCtrl == quarkv1alpha1.QuarkReadyStatusReady {
			ntapQuarkStatus.ServiceMeshControllerReady = true
		}
	}

	if quarkReconciled &&
		quark.Status.Health.NHC == quarkv1alpha1.QuarkReadyStatusReady {
		ntapQuarkStatus.HighAvailabilityReady = true
	}

	if quarkReconciled &&
		quark.Status.Health.NSC == quarkv1alpha1.QuarkReadyStatusReady {
		ntapQuarkStatus.ExternalVolumeAccessReady = true
	}

	if quarkReconciled &&
		quark.Status.Health.CSC == quarkv1alpha1.QuarkReadyStatusReady && quark.Status.Health.ADC == quarkv1alpha1.QuarkReadyStatusReady {
		ntapQuarkStatus.CloudsnapshotCapabilityReady = true
	}
	// If the cloud snapshot deployment is not created/enabled, override it's status to Ready
	if quark.Spec.Images.CSC == "" && quark.Spec.Images.ADC == "" {
		ntapQuarkStatus.CloudsnapshotCapabilityReady = true
	}

	if quarkReconciled &&
		quark.Status.Health.IPD == quarkv1alpha1.QuarkReadyStatusReady {
		ntapQuarkStatus.IPDReady = true
	}

	if quarkReconciled &&
		quark.Status.Health.EtcdCluster == quarkv1alpha1.QuarkReadyStatusReady {
		ntapQuarkStatus.EtcdClusterReady = true
	}

	if quarkReconciled &&
		quark.Status.Health.Nodevol == quarkv1alpha1.QuarkReadyStatusReady {
		ntapQuarkStatus.NodevolReady = true
	}

	if quarkReconciled &&
		quark.Status.Health.NodeMonitor == quarkv1alpha1.QuarkReadyStatusReady {
		ntapQuarkStatus.NodeMonitorReady = true
	}

	if quark.Spec.EnableGcp1p || quark.Spec.Cloud == quarkconfig.CloudAzure {
		if quarkReconciled &&
			quark.Status.Health.OtelCollector == quarkv1alpha1.QuarkReadyStatusReady {
			ntapQuarkStatus.OtelCollectorReady = true
		}
	} else {
		ntapQuarkStatus.OtelCollectorReady = true
	}

	return ntapQuarkStatus
}

func getNetAppQuarkThirdPartyComponentStatus(quark *quarkv1alpha1.Quark) NetAppQuarkThirdPartyComponentStatus {
	ntapQuarkThirdPartyComponentsStatus := NetAppQuarkThirdPartyComponentStatus{
		KubeStateMetricsReady: false,
		PrometheusServerReady: false,
		ADCReady:              false,
	}

	// Check if quark is reconciled, if in progress return false as the statuses might not be accurate
	quarkReconciled := (quark.Spec.Version == quark.Status.Version) && reflect.DeepEqual(quark.Spec.Images, quark.Status.Images)

	if !quarkReconciled {
		return ntapQuarkThirdPartyComponentsStatus
	}
	if quark.Status.Health.KubeStateMetrics == quarkv1alpha1.QuarkReadyStatusReady {
		ntapQuarkThirdPartyComponentsStatus.KubeStateMetricsReady = true
	}

	if quark.Status.Health.PrometheusServer == quarkv1alpha1.QuarkReadyStatusReady {
		ntapQuarkThirdPartyComponentsStatus.PrometheusServerReady = true
	}

	if quark.Status.Health.ADC == quarkv1alpha1.QuarkReadyStatusReady {
		ntapQuarkThirdPartyComponentsStatus.ADCReady = true
	}

	return ntapQuarkThirdPartyComponentsStatus
}

func (client *NVCClient) NetAppVolumeCloudSnapshotGet(cloudSnapshotUUID string) (NetAppVolumeCloudSnapshotInfo, error) {
	cloudSnapshot, err := client.getCloudSnapshot(cloudSnapshotUUID)
	if err != nil {
		return NetAppVolumeCloudSnapshotInfo{}, err
	}
	return getCloudSnapshotInfo(cloudSnapshot), nil
}

func (client *NVCClient) NetAppVolumeCloudSnapshotGetByVolume(volumeUUID string) (cloudSnapshotList []NetAppVolumeCloudSnapshotInfo, err error) {
	allCloudSnapshotList, err := client.NetAppVolumeCloudSnapshotGetAll()
	if err != nil {
		return
	}

	for i := range allCloudSnapshotList {
		if allCloudSnapshotList[i].VolumeUUID == volumeUUID {
			cloudSnapshotList = append(cloudSnapshotList, allCloudSnapshotList[i])
		}
	}
	return
}

func (client *NVCClient) NetAppVolumeCloudSnapshotGetAll() ([]NetAppVolumeCloudSnapshotInfo, error) {
	listResult, err := client.quarkClient.QuarkV1alpha1().NetAppVolumeCloudSnapshots(client.nvcNamespace).List(
		context.TODO(), metav1.ListOptions{})
	if err != nil {
		return nil, err
	}

	var cloudSnapshotList []NetAppVolumeCloudSnapshotInfo

	for i := range listResult.Items {
		cloudSnapshotList = append(cloudSnapshotList, getCloudSnapshotInfo(&listResult.Items[i]))
	}

	return cloudSnapshotList, nil
}

func ConstructSecret(secretName string, data map[string][]byte) *corev1.Secret {
	return &corev1.Secret{
		TypeMeta: metav1.TypeMeta{},
		ObjectMeta: metav1.ObjectMeta{
			Name: secretName,
		},
		Data: data,
		Type: corev1.SecretTypeOpaque,
	}
}

func ConstructBucketCredentialsSecret(secretName, accessKey, secret string) *corev1.Secret {
	data := make(map[string][]byte)
	data["accessKey"] = []byte(accessKey)
	data["secret"] = []byte(secret)
	return ConstructSecret(secretName, data)
}

type ActiveDirectoryConfiguration struct {
	Netbios                               string
	Domain                                string
	OrganizationUnit                      string
	DnsIPs                                []string
	EncryptedActiveDirectoryCommunication bool
	PreferredADSite                       string
}

type NetAppVolumeSMBShareProperties struct {
	EncryptionEnabled      bool
	HideShare              bool
	AccessBasedEnumeration bool
}

// NetAppVolumeCloneCreate constructs clone definition and
// creates clone of a NetApp volume with the given parent volume UUID and
// parent snapshot name. It returns the created clone volume and any error encountered.
// Parameters:
//
//	volume - the volume to be cloned
//	parentVolumeUUID - the UUID of the parent volume
//	parentSnapshotName - the name of the parent snapshot. This is snapshot.Spec.DisplayName field value.
//
// Returns:
//
//	*quarkv1alpha1.NetAppVolume - the created clone volume
//	error - error if there is any
func (client *NVCClient) NetAppVolumeCloneCreate(volume *quarkv1alpha1.NetAppVolume, parentVolumeUUID string, parentSnapshotName string) (*quarkv1alpha1.NetAppVolume, error) {

	cloneVolume, err := client.constructNetAppCloneVolume(volume, parentVolumeUUID, parentSnapshotName)
	if err != nil {
		return nil, err
	}
	return client.NetAppVolumeCreate(cloneVolume)
}

func (client *NVCClient) NetAppVolumeSMBCloneCreate(volume *quarkv1alpha1.NetAppVolume, parentVolumeUUID string, parentSnapshotName string, user string, password string) (*quarkv1alpha1.NetAppVolume, error) {

	cloneVolume, err := client.constructNetAppCloneVolume(volume, parentVolumeUUID, parentSnapshotName)
	if err != nil {
		return nil, err
	}

	return client.NetAppVolumeCreateSMB(cloneVolume, user, password)
}

// compareProtocolIgnoreVersions compares two protocol strings, ignoring the version numbers.
// It returns true if the protocol strings are equal after removing the version numbers and
// ignoring any differences in the order of protocols. Otherwise, it returns false.
//
// Parameters:
// - prot1: The first protocol string to compare.
// - prot2: The second protocol string to compare.
//
// Example:
// compareProtocolIgnoreVersions("nfs3,nfs4.1", "nfs4.1,nfs4.2") returns true.
// compareProtocolIgnoreVersions("nfs3,nfs4.1", "smb") returns false.
func compareProtocolIgnoreVersions(prot1, prot2 string) bool {
	res1 := strings.Split(prot1, ",")
	for i := range res1 {
		res1[i] = strings.TrimSpace(res1[i])
		if len(res1[i]) >= 3 {
			res1[i] = res1[i][:3]
		}
	}
	res2 := strings.Split(prot2, ",")
	for i := range res2 {
		res2[i] = strings.TrimSpace(res2[i])
		if len(res2[i]) >= 3 {
			res2[i] = res2[i][:3]
		}
	}
	if util.StringSetDifference(res1, res2) != nil ||
		util.StringSetDifference(res2, res1) != nil {
		return false
	}
	return true
}

func (client *NVCClient) constructNetAppCloneVolume(volume *quarkv1alpha1.NetAppVolume, parentVolumeUUID string, parentSnapshotName string) (*quarkv1alpha1.NetAppVolume, error) {
	parentVolume, err := client.NetAppVolumeGet(parentVolumeUUID)
	if err != nil {
		return nil, err
	}

	// Check if snapshot exists in the parent snapshotlist
	parentSnapshotList, err := client.NetAppVolumeSnapshotsListGet(parentVolumeUUID)
	if err != nil {
		return nil, fmt.Errorf("failed to retrieve snapshot list info for volume %s", parentVolumeUUID)
	}

	foundSnapshot := false
	for _, snap := range parentSnapshotList.Status.Snapshots {
		if snap.Name == parentSnapshotName {
			foundSnapshot = true
			break
		}
	}

	if !foundSnapshot {
		// We did not find the snapshot return error
		return nil, fmt.Errorf("failed to find parent snapshot named \"%s\" in parent snapshotlist", parentSnapshotName)
	}

	cloneVolumeCopy := volume.DeepCopy()

	// Make sure parent volume and clone have the same protocol
	if equal := compareProtocolIgnoreVersions(parentVolume.Spec.Protocol, cloneVolumeCopy.Spec.Protocol); !equal {
		return nil, fmt.Errorf("clone protocol should match parent protocol, parentProtocol:%s cloneProtocol:%s", parentVolume.Spec.Protocol, cloneVolumeCopy.Spec.Protocol)
	}

	// Make sure parent and clone belong to the same storage pool
	if parentVolume.Labels[quarkconfig.StoragePoolLabel] != cloneVolumeCopy.Labels[quarkconfig.StoragePoolLabel] {
		return nil, fmt.Errorf("clone volume and parent volume should belong to the same storagepool")
	}

	// unixPermission for the cloned volume will be internally set by fetching it from DMAP.
	cloneVolumeCopy.Spec.UnixPermission = ""

	// Propagate Volume Encryption label from parent volume
	if volEncLabel, ok := parentVolume.Labels[quarkconfig.VolumeEncryptionLabel]; ok {
		if volEncLabel == quarkv1alpha1.EncryptionStateScheduled.String() {
			cloneVolumeCopy.ObjectMeta.Labels[quarkconfig.VolumeEncryptionLabel] = volEncLabel
		}
	}

	// Update clone information in volume
	cloneVolumeCopy.Spec.CloneParentVolumeUUID = parentVolumeUUID
	cloneVolumeCopy.Spec.CloneParentSnapshot = parentSnapshotName
	cloneVolumeCopy.ObjectMeta.Labels[quarkconfig.VolumeCloneParentUUUIDLabel] = parentVolumeUUID
	return cloneVolumeCopy, nil
}

// Checks if passed in parent volume is still associated with any clone in the cluster
// If snapshotname is not passed in we check if any snapshot with the provided volumeUUID is still associated to a clone
// If snapshot name is provided we verify if that specific snapshot is still associated with a clone
// Returns a list of all clones associated with the volume
func (client *NVCClient) getAssociatedCloneVolumeList(parentVolumeUUID string, snapshotName string) ([]quarkv1alpha1.NetAppVolume, error) {

	var associatedCloneVolumeList []quarkv1alpha1.NetAppVolume
	// Construct selector to only list clone volumes with split scan running
	selectorString := quarkconfig.VolumeCloneParentUUUIDLabel + "=" + parentVolumeUUID
	listOptions := metav1.ListOptions{
		LabelSelector: selectorString,
	}

	result, err := client.listWithRetry(&quarkv1alpha1.NetAppVolumeList{}, listOptions)
	if err != nil {
		return []quarkv1alpha1.NetAppVolume{}, err
	}
	volumes := result.(*quarkv1alpha1.NetAppVolumeList)

	if snapshotName != "" {
		// Filter out the volumes that have the required snapshotName and return only those volumes
		for _, volume := range volumes.Items {
			if volume.Spec.CloneParentSnapshot == snapshotName {
				associatedCloneVolumeList = append(associatedCloneVolumeList, volume)
			}
		}
		return associatedCloneVolumeList, nil
	}

	return volumes.Items, nil
}

// Check if given snapshot is a base clone snapshot of the given volume and if it is locked due to ongoing split scan
// Returns true if locked due to ongoing split scan
func (client *NVCClient) isBaseCloneSnapshot(volumeUUID string, snapshotName string) (bool, error) {
	volume, err := client.NetAppVolumeGet(volumeUUID)
	if err != nil {
		return false, err
	}
	// Check if Clone Parent Snapshot is the snapshot passed in
	if volume.Spec.CloneParentSnapshot == snapshotName {
		// Clone fields are still populated, splitscan is still ongoing, return true
		return true, nil
	}
	return false, nil
}

// ConstructNetAppVolumeDPSMB creates a  NetAppVolume definitionfor DataProtection (DP) volume (secondary volume)
func ConstructNetAppVolumeDPSMB(ownerID, displayName, sizeString, volumeUUID, storagePoolUUID, volumePath, snapshotPolicyName string,
	serviceLevel ServiceLevel,
	noSnapDir bool,
	activeDirectoryConfig ActiveDirectoryConfiguration,
	shareProperties NetAppVolumeSMBShareProperties) (volume *quarkv1alpha1.NetAppVolume, err error) {
	volume, err = ConstructNetAppVolumeSMB(ownerID,
		displayName,
		sizeString,
		volumeUUID,
		storagePoolUUID,
		volumePath,
		snapshotPolicyName,
		serviceLevel,
		noSnapDir,
		activeDirectoryConfig,
		shareProperties)
	if err != nil {
		return nil, err
	}
	volume.Spec.DataProtection = true
	return volume, nil
}

// Constructs SMB NetAppVolume definition.
// ownerID: OwnerID of the volume.
// dispalyName: Name of the volume provided by customer.
// sizeString: Volume Size.
// volumeUUID: Unique ID of volume in k8s cluster.
// volumePath: Volume mount path.
// serviceLevel: Standard for alpha.
// snapshotPolicy: Snapshot Policy object
// noSnapDir: Hide snapshot directory
// activeDirectoryConfig: Active directory configuration
// shareProperties: SMB share properties
func ConstructNetAppVolumeSMB(ownerID, displayName, sizeString, volumeUUID, storagePoolUUID, volumePath, snapshotPolicyName string, serviceLevel ServiceLevel, noSnapDir bool, activeDirectoryConfig ActiveDirectoryConfiguration, shareProperties NetAppVolumeSMBShareProperties) (volume *quarkv1alpha1.NetAppVolume, err error) {
	if storagePoolUUID != "" {
		err = nvc.ValidateVolumePath(volumePath)
		if err != nil {
			return nil, err
		}
	} else {
		err = nvc.ValidateSMBConfiguration(volumePath, activeDirectoryConfig.Netbios, activeDirectoryConfig.Domain, activeDirectoryConfig.OrganizationUnit, activeDirectoryConfig.PreferredADSite)
		if err != nil {
			return nil, err
		}
	}
	volume, err = constructBaseNetAppVolume(ownerID, displayName, sizeString, volumeUUID, storagePoolUUID, volumePath, snapshotPolicyName, serviceLevel, noSnapDir)
	if err != nil {
		return nil, err
	}

	volume.Spec.Protocol = "cifs"
	if storagePoolUUID == "" {
		volume.Spec.SMBConfiguration.Netbios = activeDirectoryConfig.Netbios
		volume.Spec.SMBConfiguration.Domain = activeDirectoryConfig.Domain
		volume.Spec.SMBConfiguration.OrganizationUnit = activeDirectoryConfig.OrganizationUnit
		volume.Spec.SMBConfiguration.DNSIPs = activeDirectoryConfig.DnsIPs
		volume.Spec.SMBPasswordResetCount = 0
		volume.Spec.SMBConfiguration.EncryptedADCommunication = activeDirectoryConfig.EncryptedActiveDirectoryCommunication
		volume.Spec.SMBConfiguration.PreferredADSite = activeDirectoryConfig.PreferredADSite
	}
	volume.Spec.SMBConfiguration.EncryptionEnabled = shareProperties.EncryptionEnabled
	volume.Spec.SMBConfiguration.HideShare = shareProperties.HideShare
	volume.Spec.SMBConfiguration.ShareAccessBasedEnumeration = shareProperties.AccessBasedEnumeration

	return volume, nil
}

func (client *NVCClient) SecretCreate(secret *corev1.Secret) (*corev1.Secret, error) {
	return client.kubeClient.CoreV1().Secrets(client.nvcNamespace).Create(
		context.TODO(), secret, metav1.CreateOptions{})
}

func (client *NVCClient) SecretGet(secretName string) (*corev1.Secret, error) {
	return client.kubeClient.CoreV1().Secrets(client.nvcNamespace).Get(
		context.TODO(), secretName, metav1.GetOptions{})
}

func (client *NVCClient) SecretUpdate(secret *corev1.Secret) (*corev1.Secret, error) {
	return client.kubeClient.CoreV1().Secrets(client.nvcNamespace).Update(
		context.TODO(), secret, metav1.UpdateOptions{})
}

func (client *NVCClient) SecretCreateOrUpdate(secret *corev1.Secret) (*corev1.Secret, error) {
	newSecret, err := client.SecretCreate(secret)
	if err != nil && errors.IsAlreadyExists(err) {
		newSecret, err = client.SecretGet(secret.Name)
		if err == nil {
			newSecret.Data = secret.Data
			newSecret, err = client.SecretUpdate(newSecret)
		}
	}
	return newSecret, err
}

func (client *NVCClient) SecretDelete(secretName string) error {
	return client.kubeClient.CoreV1().Secrets(client.nvcNamespace).Delete(
		context.TODO(), secretName, metav1.DeleteOptions{})
}

// GetKubeClient returns the current kubeclient for NVCClient
func (client *NVCClient) GetKubeClient() kubernetes.Interface {
	return client.kubeClient
}

// GetKubeClient returns the current kubeclient for NVCClient
func (client *NVCClient) GetQuarkClient() quarkclientset.Interface {
	return client.quarkClient
}

func (client *NVCClient) NetAppGetZoneStatus() (map[string]string, error) {
	var (
		err error
	)

	regionalHACM, err := client.kubeClient.CoreV1().ConfigMaps(
		client.nvcNamespace,
	).Get(
		context.TODO(), quarkconfig.RegionalHAConfigMapName, metav1.GetOptions{})
	if err != nil {
		return nil, fmt.Errorf("unable to get regional HA configmap")
	}

	// Remove the nhcRecoveryAcks if they are present
	delete(regionalHACM.Data, "nhcRecoveryAcks")

	return regionalHACM.Data, err
}

func (client *NVCClient) NetAppSetZoneStatusActive(zone string) error {
	var (
		err error
	)

	regionalHACM, err := client.kubeClient.CoreV1().ConfigMaps(
		client.nvcNamespace,
	).Get(context.TODO(), quarkconfig.RegionalHAConfigMapName, metav1.GetOptions{})
	if err != nil {
		return fmt.Errorf("unable to get regional HA configmap")
	}

	_, found := regionalHACM.Data[zone]
	if !found {
		return fmt.Errorf("zone %s not found", zone)
	}

	// Mark the zone active
	cmCopy := regionalHACM.DeepCopy()
	cmCopy.Data[zone] = fmt.Sprintf("%s",
		quarkconfig.RegionalHAActiveZone)
	if _, err = client.kubeClient.CoreV1().ConfigMaps(client.nvcNamespace).Update(
		context.TODO(), cmCopy, metav1.UpdateOptions{},
	); err != nil {
		log.WithFields(log.Fields{
			"error": err.Error(),
		}).Error("Failed to update regional HA configmap.")
		return fmt.Errorf("failed to update zone status active")
	}

	log.WithFields(log.Fields{
		"zone": zone,
	}).Info("Updated zone status active")
	return nil
}

/*
// TODO: To be evaluated for use in 2021.12 -> .12+ upgrades in place of isVolUpgradeinProgress to take care of corner cases
func (client *NVCClient) isStoragePoolUpgradeInProgess(vol *quarkv1alpha1.NetAppVolume) bool {
	// Get the storage pool that the volume is associated with
	if storagePoolUuid, ok := vol.Labels[quarkconfig.StoragePoolLabel]; ok {
		storagePool, _ := client.NetAppStoragePoolGet(storagePoolUuid)
		// check whether storage pool is upgrading
		if storagePool != nil {
			return storagePool.Status.QuarkVersion != "" && storagePool.Spec.QuarkVersion != storagePool.Status.QuarkVersion
		}
	}
	return false
}
*/

// returns true if volume is currently undergoing upgrade
func (client *NVCClient) isVolUpgradeInProgess(vol *quarkv1alpha1.NetAppVolume) (volUpgrading bool) {
	var requiredQuarkVersion string
	if spUUID, ok := vol.Labels[quarkconfig.StoragePoolLabel]; ok {
		// When storage pool is present, we want to compare versions on volume status and sp status,
		// since podrick sets volume.Status.QuarkVersion field when storagePool != nil
		spVersion, err := client.NetAppStoragePoolGetVersion(spUUID)
		if err == nil {
			requiredQuarkVersion = spVersion
		}
	} else {
		requiredQuarkVersion = vol.Spec.QuarkVersion
	}
	if vol.Status.QuarkVersion != "" && vol.Status.QuarkVersion != requiredQuarkVersion {
		volUpgrading = true
		for _, cond := range vol.Status.Conditions {
			if cond.Type == quarkv1alpha1.NetAppVolumeUpgradeFailed {
				// if upgrade has failed, there is no upgrade operation underway
				volUpgrading = false
			}
		}
	}
	return
}

// returns error message if an AD field is being updated
func ADFieldUpdateConflict(volume *quarkv1alpha1.NetAppVolume) (errMsg string) {
	if !reflect.DeepEqual(volume.Spec.SMBConfiguration.DNSIPs, volume.Status.SMBConfiguration.DNSIPs) {
		errMsg = "DNSIPs update in progress"
	} else if volume.Spec.SMBConfiguration.EncryptedADCommunication != volume.Status.SMBConfiguration.EncryptedADCommunication {
		errMsg = "EncryptionADCommunication update in progress"
	}
	return
}

// returns the correct error message if any of the volume AD fields or volume size are undergoing an update,
// or volume is undergoing deletion
func isConflictingVolUpdateInProgess(vol *quarkv1alpha1.NetAppVolume) (errMsg string) {
	errMsg = ADFieldUpdateConflict(vol)
	if errMsg != "" {
		return
	}
	if vol.Spec.Size.Value() != vol.Status.Size.Value() {
		errMsg = "volume size update in progress"
	} else if vol.DeletionTimestamp != nil {
		errMsg = "volume deletion in progress"
	}
	return
}

// Check if any quark component is being patched
func isQuarkComponentBeingPatchedValid(quarkCR *quarkv1alpha1.Quark, images map[string]interface{}) (bool, error) {
	if val, exists := images["DMAP"]; exists {
		// make sure all other known quark components also have the same tag
		quarkimageTag, err := util.GetImageTagOrDigest(val.(string))
		if err != nil {
			return false, err
		}
		for componentName, _ := range quarkconfig.KnownQuarkComponentImages {
			if imageTag, componentExists := images[componentName]; componentExists {
				componentImageTag, err := util.GetImageTagOrDigest(imageTag.(string))
				if err != nil {
					return false, err
				}
				if componentImageTag != quarkimageTag {
					return false, fmt.Errorf("all quark component tags don't match, expected %s, have %s", quarkimageTag, componentImageTag)
				}
			} else {
				// Do not error if CCPD is not enabled
				if componentName == "CCPD" && !quarkCR.Spec.EnableCcpd {
					continue
				}
				return false, fmt.Errorf("quark component %s not found in images map", componentName)
			}
		}
	} else {
		// Make sure no other quark component exists in the images map
		for componetName, _ := range quarkconfig.KnownQuarkComponentImages {
			if _, exists := images[componetName]; exists {
				// Found some other quark component that exists, which indicates not all quark components are being updated
				var knownQuarkComponents []string
				for component, _ := range quarkconfig.KnownQuarkComponentImages {
					knownQuarkComponents = append(knownQuarkComponents, component)
				}
				return false, fmt.Errorf("not all quark components are being updated, expected quark components %v", knownQuarkComponents)
			}
		}
	}
	return true, nil
}

func patchImages(quarkCR *quarkv1alpha1.Quark, images map[string]interface{}, registry string) (*quarkv1alpha1.Quark, error) {
	// DeepCopy
	quarkCRCopy := quarkCR.DeepCopy()

	if !quarkCRCopy.Spec.EnableCcpd {
		// Remove CCPD from images map if it is not enabled
		delete(images, "CCPD")
	}

	// Check if any of the Ontap or quark components are being updated
	isValid, err := isQuarkComponentBeingPatchedValid(quarkCRCopy, images)
	if !isValid || err != nil {
		return quarkCRCopy, err
	}

	imagesStruct := &quarkCRCopy.Spec.Images

	// If quark components are being updated make sure they have the same tag
	err = util.FillStructFieldsFromMap(imagesStruct, images)
	return quarkCRCopy, err
}

// Error returned by this function should not be overwritten as it has the logic to
// create appropriate concrete error types like StaleVersionError and
// clients (SDE) depend on those concrete error types
func (client *NVCClient) netAppVolumeDataProtectionUpdate(
	volumeUUID string,
	dp quarkv1alpha1.DataProtectionType) error {
	volume, err := client.NetAppVolumeGet(volumeUUID)
	if err != nil {
		return err
	}
	if volume.Spec.DataProtection == dp {
		return nil
	}
	volumeCopy := volume.DeepCopy()
	volumeCopy.Spec.DataProtection = dp
	_, err = client.netAppVolumeUpdate(volumeCopy)
	return err
}
