package v2

import (
	"context"
	"fmt"
	"strings"

	uuid "github.com/satori/go.uuid"
	corev1 "k8s.io/api/core/v1"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	quarkv1alpha1 "quark.netapp.io/volume-controller/pkg/apis/volumecontroller/v1alpha1"
	"quark.netapp.io/volume-controller/pkg/quarkconfig"
)

// NetAppObjectStoreDelete is used to signal the deletion of a NetAppObjectStore resource
type NetAppObjectStoreDelete struct {
	ObjectStore *quarkv1alpha1.NetAppObjectStore
}

// NetAppObjectStoreUpdate is used to signal the most recent state of a NetAppObjectStoreUpdate and the
// most recent changes.
type NetAppObjectStoreUpdate struct {
	ObjectStore       *quarkv1alpha1.NetAppObjectStore
	UpdatedConditions []quarkv1alpha1.NetAppObjectStoreCondition
}

// ConstructNetAppObjectStore creates a NetAppObjectStore struct resource
func ConstructNetAppObjectStore(
	ownerID string,
	displayName string,
	objectStoreUUID uuid.UUID,
	storagePoolUUID uuid.UUID,
	providerType quarkv1alpha1.NetAppVolumeCloudSnapshotProviderType,
	container string,
	serverName string,
	secretName string,
	secretNamespace string,
) (objectStore *quarkv1alpha1.NetAppObjectStore, err error) {

	if len(strings.TrimSpace(displayName)) == 0 ||
		len(strings.TrimSpace(container)) == 0 ||
		len(strings.TrimSpace(serverName)) == 0 ||
		len(strings.TrimSpace(secretName)) == 0 ||
		len(strings.TrimSpace(secretNamespace)) == 0 {
		return nil, fmt.Errorf("invalid input, one or more of the following values are not populated: %v: \"%v\", %v: \"%v\", %v: \"%v\", %v: \"%v\", %v: \"%v\"",
			"displayName", displayName, "container", container, "serverName", serverName, "secretName", secretName, "secretNamespace", secretNamespace)
	}

	if providerType != quarkv1alpha1.GoogleCloud {
		return nil, fmt.Errorf("invalid provider type, supported provider type is %v", quarkv1alpha1.GoogleCloud)
	}

	objectStore = &quarkv1alpha1.NetAppObjectStore{
		TypeMeta: metav1.TypeMeta{},
		ObjectMeta: metav1.ObjectMeta{
			Name: objectStoreUUID.String(),
			Labels: map[string]string{
				quarkconfig.StoragePoolLabel: storagePoolUUID.String(),
				quarkconfig.OwnerIDLabel:     ownerID,
			},
		},
		Spec: quarkv1alpha1.NetAppObjectStoreSpec{
			ProviderType: providerType,
			DisplayName:  displayName,
			Container:    container,
			ServerName:   serverName,
			SecretRef: corev1.SecretReference{
				Name:      secretName,
				Namespace: secretNamespace,
			},
			SecretRefCounter: 0,
		},
	}

	return objectStore, nil
}

// NetAppObjectStoreCreate creates a NetAppObjectStore CR by making a REST API call to k8s api server
func (client *NVCClient) NetAppObjectStoreCreate(
	objectStore *quarkv1alpha1.NetAppObjectStore,
) (*quarkv1alpha1.NetAppObjectStore, error) {
	_, err := client.storagePoolLister.NetAppStoragePools(client.nvcNamespace).Get(objectStore.Labels[quarkconfig.StoragePoolLabel])
	if err != nil {
		return nil, fmt.Errorf("cannot create NetAppObjectStore. Failed to verify if NetAppStoragePool %v exists. %v",
			objectStore.Labels[quarkconfig.StoragePoolLabel], err)
	}

	_, err = client.kubeClient.CoreV1().Secrets(objectStore.Spec.SecretRef.Namespace).Get(
		context.TODO(), objectStore.Spec.SecretRef.Name, metav1.GetOptions{})
	if err != nil {
		return nil, fmt.Errorf("cannot create NetAppObjectStore. Failed to verify if Secret %v/%v exists. %v",
			objectStore.Spec.SecretRef.Namespace, objectStore.Spec.SecretRef.Name, err)
	}

	return client.quarkClient.QuarkV1alpha1().NetAppObjectStores(client.nvcNamespace).Create(
		context.TODO(), objectStore, metav1.CreateOptions{})
}

// NetAppObjectStoreDelete deletes the NetAppObjectStore CR by making a REST API call to k8s api server
func (client *NVCClient) NetAppObjectStoreDelete(
	objectStoreUUID uuid.UUID,
) error {
	return client.quarkClient.QuarkV1alpha1().NetAppObjectStores(client.nvcNamespace).Delete(
		context.TODO(), objectStoreUUID.String(), metav1.DeleteOptions{})
}

// NetAppObjectStoreGet gets the NetAppObjectStore CR by making a REST API call to k8s api server
func (client *NVCClient) NetAppObjectStoreGet(objectStoreUUID uuid.UUID) (*quarkv1alpha1.NetAppObjectStore, error) {
	return client.quarkClient.QuarkV1alpha1().NetAppObjectStores(client.nvcNamespace).Get(
		context.TODO(), objectStoreUUID.String(), metav1.GetOptions{})
}

// NetAppObjectStoreGetAll gets all the NetAppObjectStore CRs by making a REST API call to k8s api server
func (client *NVCClient) NetAppObjectStoreGetAll() ([]*quarkv1alpha1.NetAppObjectStore, error) {
	objectStoreList, err := client.quarkClient.QuarkV1alpha1().NetAppObjectStores(client.nvcNamespace).List(
		context.TODO(), metav1.ListOptions{})
	if err != nil {
		return nil, err
	}
	return netAppObjectStoreToSlice(objectStoreList), nil
}

func netAppObjectStoreToSlice(list *quarkv1alpha1.NetAppObjectStoreList) []*quarkv1alpha1.NetAppObjectStore {
	objectStores := make([]*quarkv1alpha1.NetAppObjectStore, len(list.Items))
	for i := range list.Items {
		objectStores[i] = &list.Items[i]
	}
	return objectStores
}

// netAppObjectStoreUpdate attempts to update a NetAppObjectStore CR
func (client *NVCClient) netAppObjectStoreUpdate(objectStore *quarkv1alpha1.NetAppObjectStore) (*quarkv1alpha1.NetAppObjectStore, error) {
	newObjectStore, err := client.quarkClient.QuarkV1alpha1().NetAppObjectStores(client.nvcNamespace).Update(
		context.TODO(), objectStore, metav1.UpdateOptions{})
	if err != nil {
		if strings.Contains(err.Error(), quarkconfig.OldVersionErr) {
			return nil, StaleVersionError{}
		}
		return nil, err
	}
	return newObjectStore, nil
}

// NetAppObjectStoreIncrementSecretRefCounter increments the Spec.SecretRefCounter. This signifies that the object store secret
// has been updated and quark control/data plane should update its cached secret value.
func (client *NVCClient) NetAppObjectStoreIncrementSecretRefCounter(objectStoreUUID uuid.UUID) error {
	objStore, err := client.objectStoreLister.NetAppObjectStores(client.nvcNamespace).Get(objectStoreUUID.String())
	if err != nil {
		return err
	}

	if objStore.Spec.SecretRefCounter > objStore.Status.SecretRefCounter {
		return fmt.Errorf("invalid operation, a secret update is already in progress")
	}

	objStore.Spec.SecretRefCounter++
	_, err = client.netAppObjectStoreUpdate(objStore)
	return err
}

func netAppObjectStoreSecretUpdateSynced(objectStore *quarkv1alpha1.NetAppObjectStore) bool {
	return objectStore.Spec.SecretRefCounter == objectStore.Status.SecretRefCounter
}

func netAppObjectStoreReady(objectStore *quarkv1alpha1.NetAppObjectStore) bool {
	for _, cond := range objectStore.Status.Conditions {
		if cond.Type == quarkv1alpha1.ObjectStoreReadyCondition && cond.Status == corev1.ConditionTrue {
			return true
		}
	}
	return false
}
