package v2

import (
	"context"
	"fmt"
	"time"

	log "github.com/sirupsen/logrus"
	"k8s.io/apimachinery/pkg/api/errors"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	quarkv1alpha1 "quark.netapp.io/volume-controller/pkg/apis/volumecontroller/v1alpha1"
)

// Following status codes correspond to certain StatusReason constants in the Kubernetes Go client library.
// For a complete list of StatusReason constants and their typical meaning, refer to:
// https://github.com/kubernetes/kubernetes/blob/master/staging/src/k8s.io/apimachinery/pkg/api/errors/errors.go
// https://github.com/kubernetes/apimachinery/blob/master/pkg/apis/meta/v1/types.go

var retryableErrorCodes = map[int32]bool{
	500: true, // Internal Server Error
	502: true, // Bad Gateway
	503: true, // Service Unavailable
	504: true, // Gateway Timeout
	429: true, // Too Many Requests
}

// APIFunc is a type that represents any Kubernetes API function that returns a result and an error
type APIFunc func() (interface{}, error)

// withRetry takes any Kubernetes API function and applies retry logic to it
func (client *NVCClient) withRetry(apiCall APIFunc) (interface{}, error) {
	var err error
	timeout := client.Config.K8sApiRetryInterval
	retries := client.Config.K8sApiRetryCount

	for i := 0; i < retries; i++ {
		result, err := apiCall()
		if err != nil {
			if isRetryableError(err) {
				log.Debugf("Attempt %d failed with error: %v. Retrying after %v...", i+1, err, timeout)
				time.Sleep(timeout)
				continue
			}
			log.Debugf("Attempt %d failed with non-retryable error: %v", i+1, err)
			// If the error is not retryable, return the result and the error immediately
			return result, err
		}
		// If the API call was successful, return the result
		return result, nil
	}
	log.Debugf("All %d attempts failed. Final error: %v", retries, err)
	// If the API call failed after all retry attempts, return the result and the error from the last attempt
	return nil, err
}

func isRetryableError(err error) bool {
	if serr, ok := err.(*errors.StatusError); ok {
		code := serr.Status().Code
		return retryableErrorCodes[code]
	}
	return false
}

// Create apis
func (client *NVCClient) createWithRetry(obj interface{}) (interface{}, error) {
	return client.withRetry(func() (interface{}, error) {
		switch objType := obj.(type) {
		case *quarkv1alpha1.NetAppClusterPeering:
			return client.quarkClient.QuarkV1alpha1().NetAppClusterPeerings(client.nvcNamespace).Create(
				context.TODO(), objType, metav1.CreateOptions{})
		case *quarkv1alpha1.Quark:
			return client.quarkClient.QuarkV1alpha1().Quarks(client.nvcNamespace).Create(
				context.TODO(), objType, metav1.CreateOptions{})
		case *quarkv1alpha1.NetAppStoragePool:
			return client.quarkClient.QuarkV1alpha1().NetAppStoragePools(client.nvcNamespace).Create(
				context.TODO(), objType, metav1.CreateOptions{})
		case *quarkv1alpha1.NetAppExportPolicy:
			return client.quarkClient.QuarkV1alpha1().NetAppExportPolicies(client.nvcNamespace).Create(
				context.TODO(), objType, metav1.CreateOptions{})
		case *quarkv1alpha1.NetAppSnapshotPolicy:
			return client.quarkClient.QuarkV1alpha1().NetAppSnapshotPolicies(client.nvcNamespace).Create(
				context.TODO(), objType, metav1.CreateOptions{})
		case *quarkv1alpha1.NetAppVolume:
			return client.quarkClient.QuarkV1alpha1().NetAppVolumes(client.nvcNamespace).Create(
				context.TODO(), objType, metav1.CreateOptions{})
		case *quarkv1alpha1.NetAppVolumeSnapshot:
			return client.quarkClient.QuarkV1alpha1().NetAppVolumeSnapshots(client.nvcNamespace).Create(
				context.TODO(), objType, metav1.CreateOptions{})
		// Add cases for other types as needed
		default:
			return nil, fmt.Errorf("unsupported type %T", objType)
		}
	})
}

// UpdateStatus apis
func (client *NVCClient) updateStatusWithRetry(obj interface{}) (interface{}, error) {
	return client.withRetry(func() (interface{}, error) {
		switch objType := obj.(type) {
		case *quarkv1alpha1.NetAppClusterPeering:
			return client.quarkClient.QuarkV1alpha1().NetAppClusterPeerings(client.nvcNamespace).UpdateStatus(
				context.TODO(), objType, metav1.UpdateOptions{})
		case *quarkv1alpha1.Quark:
			return client.quarkClient.QuarkV1alpha1().Quarks(client.nvcNamespace).UpdateStatus(
				context.TODO(), objType, metav1.UpdateOptions{})
		case *quarkv1alpha1.NetAppStoragePool:
			return client.quarkClient.QuarkV1alpha1().NetAppStoragePools(client.nvcNamespace).UpdateStatus(
				context.TODO(), objType, metav1.UpdateOptions{})
		case *quarkv1alpha1.NetAppExportPolicy:
			return client.quarkClient.QuarkV1alpha1().NetAppExportPolicies(client.nvcNamespace).UpdateStatus(
				context.TODO(), objType, metav1.UpdateOptions{})
		case *quarkv1alpha1.NetAppSnapshotPolicy:
			return client.quarkClient.QuarkV1alpha1().NetAppSnapshotPolicies(client.nvcNamespace).UpdateStatus(
				context.TODO(), objType, metav1.UpdateOptions{})
		case *quarkv1alpha1.NetAppVolume:
			return client.quarkClient.QuarkV1alpha1().NetAppVolumes(client.nvcNamespace).UpdateStatus(
				context.TODO(), objType, metav1.UpdateOptions{})
		// Add cases for other types as needed
		default:
			return nil, fmt.Errorf("unsupported type %T", objType)
		}
	})
}

// Update apis
func (client *NVCClient) updateWithRetry(obj interface{}) (interface{}, error) {
	return client.withRetry(func() (interface{}, error) {
		switch objType := obj.(type) {
		case *quarkv1alpha1.NetAppClusterPeering:
			return client.quarkClient.QuarkV1alpha1().NetAppClusterPeerings(client.nvcNamespace).Update(
				context.TODO(), objType, metav1.UpdateOptions{})
		case *quarkv1alpha1.Quark:
			return client.quarkClient.QuarkV1alpha1().Quarks(client.nvcNamespace).Update(
				context.TODO(), objType, metav1.UpdateOptions{})
		case *quarkv1alpha1.NetAppStoragePool:
			return client.quarkClient.QuarkV1alpha1().NetAppStoragePools(client.nvcNamespace).Update(
				context.TODO(), objType, metav1.UpdateOptions{})
		case *quarkv1alpha1.NetAppExportPolicy:
			return client.quarkClient.QuarkV1alpha1().NetAppExportPolicies(client.nvcNamespace).Update(
				context.TODO(), objType, metav1.UpdateOptions{})
		case *quarkv1alpha1.NetAppSnapshotPolicy:
			return client.quarkClient.QuarkV1alpha1().NetAppSnapshotPolicies(client.nvcNamespace).Update(
				context.TODO(), objType, metav1.UpdateOptions{})
		case *quarkv1alpha1.NetAppVolume:
			return client.quarkClient.QuarkV1alpha1().NetAppVolumes(client.nvcNamespace).Update(
				context.TODO(), objType, metav1.UpdateOptions{})
		case *quarkv1alpha1.NetAppVolumeSnapshotsList:
			return client.quarkClient.QuarkV1alpha1().NetAppVolumeSnapshotsLists(client.nvcNamespace).Update(
				context.TODO(), objType, metav1.UpdateOptions{})
		// Add cases for other types as needed
		default:
			return nil, fmt.Errorf("unsupported type %T", objType)
		}
	})
}

// GET apis
func (client *NVCClient) getWithRetry(obj interface{}) (interface{}, error) {
	return client.withRetry(func() (interface{}, error) {
		switch objType := obj.(type) {
		case *quarkv1alpha1.NetAppClusterPeering:
			return client.quarkClient.QuarkV1alpha1().NetAppClusterPeerings(client.nvcNamespace).Get(
				context.TODO(), objType.Name, metav1.GetOptions{})
		case *quarkv1alpha1.Quark:
			return client.quarkClient.QuarkV1alpha1().Quarks(client.nvcNamespace).Get(
				context.TODO(), objType.Name, metav1.GetOptions{})
		case *quarkv1alpha1.NetAppStoragePool:
			return client.quarkClient.QuarkV1alpha1().NetAppStoragePools(client.nvcNamespace).Get(
				context.TODO(), objType.Name, metav1.GetOptions{})
		case *quarkv1alpha1.NetAppExportPolicy:
			return client.quarkClient.QuarkV1alpha1().NetAppExportPolicies(client.nvcNamespace).Get(
				context.TODO(), objType.Name, metav1.GetOptions{})
		case *quarkv1alpha1.NetAppSnapshotPolicy:
			return client.quarkClient.QuarkV1alpha1().NetAppSnapshotPolicies(client.nvcNamespace).Get(
				context.TODO(), objType.Name, metav1.GetOptions{})
		case *quarkv1alpha1.NetAppVolume:
			return client.quarkClient.QuarkV1alpha1().NetAppVolumes(client.nvcNamespace).Get(
				context.TODO(), objType.Name, metav1.GetOptions{})
		case *quarkv1alpha1.NetAppVolumeSnapshot:
			return client.quarkClient.QuarkV1alpha1().NetAppVolumeSnapshots(client.nvcNamespace).Get(
				context.TODO(), objType.Name, metav1.GetOptions{})
		case *quarkv1alpha1.NetAppVolumeSnapshotsList:
			return client.quarkClient.QuarkV1alpha1().NetAppVolumeSnapshotsLists(client.nvcNamespace).Get(
				context.TODO(), objType.Name, metav1.GetOptions{})
		// Add cases for other types as needed
		default:
			return nil, fmt.Errorf("unsupported type %T", objType)
		}
	})
}

// LIST apis
func (client *NVCClient) listWithRetry(obj interface{}, opts metav1.ListOptions) (interface{}, error) {
	return client.withRetry(func() (interface{}, error) {
		switch objType := obj.(type) {
		case *quarkv1alpha1.NetAppClusterPeeringList:
			return client.quarkClient.QuarkV1alpha1().NetAppClusterPeerings(client.nvcNamespace).List(
				context.TODO(), opts)
		case *quarkv1alpha1.QuarkList:
			return client.quarkClient.QuarkV1alpha1().Quarks(client.nvcNamespace).List(
				context.TODO(), opts)
		case *quarkv1alpha1.NetAppStoragePoolList:
			return client.quarkClient.QuarkV1alpha1().NetAppStoragePools(client.nvcNamespace).List(
				context.TODO(), opts)
		case *quarkv1alpha1.NetAppVolumeList:
			return client.quarkClient.QuarkV1alpha1().NetAppVolumes(client.nvcNamespace).List(
				context.TODO(), opts)
		case *quarkv1alpha1.NetAppVolumeSnapshotList:
			return client.quarkClient.QuarkV1alpha1().NetAppVolumeSnapshots(client.nvcNamespace).List(
				context.TODO(), opts)
		// Add cases for other types as needed
		default:
			return nil, fmt.Errorf("unsupported type %T", objType)
		}
	})
}

// DELETE apis
func (client *NVCClient) deleteWithRetry(obj interface{}) error {
	_, err := client.withRetry(func() (interface{}, error) {
		switch objType := obj.(type) {
		case *quarkv1alpha1.NetAppClusterPeering:
			return nil, client.quarkClient.QuarkV1alpha1().NetAppClusterPeerings(client.nvcNamespace).Delete(
				context.TODO(), objType.ObjectMeta.Name, metav1.DeleteOptions{})
		case *quarkv1alpha1.Quark:
			return nil, client.quarkClient.QuarkV1alpha1().Quarks(client.nvcNamespace).Delete(
				context.TODO(), objType.ObjectMeta.Name, metav1.DeleteOptions{})
		case *quarkv1alpha1.NetAppStoragePool:
			return nil, client.quarkClient.QuarkV1alpha1().NetAppStoragePools(client.nvcNamespace).Delete(
				context.TODO(), objType.ObjectMeta.Name, metav1.DeleteOptions{})
		case *quarkv1alpha1.NetAppVolume:
			return nil, client.quarkClient.QuarkV1alpha1().NetAppVolumes(client.nvcNamespace).Delete(
				context.TODO(), objType.ObjectMeta.Name, metav1.DeleteOptions{})
		case *quarkv1alpha1.NetAppExportPolicy:
			return nil, client.quarkClient.QuarkV1alpha1().NetAppExportPolicies(client.nvcNamespace).Delete(
				context.TODO(), objType.ObjectMeta.Name, metav1.DeleteOptions{})
		case *quarkv1alpha1.NetAppSnapshotPolicy:
			return nil, client.quarkClient.QuarkV1alpha1().NetAppSnapshotPolicies(client.nvcNamespace).Delete(
				context.TODO(), objType.ObjectMeta.Name, metav1.DeleteOptions{})
		case *quarkv1alpha1.NetAppVolumeSnapshot:
			return nil, client.quarkClient.QuarkV1alpha1().NetAppVolumeSnapshots(client.nvcNamespace).Delete(
				context.TODO(), objType.ObjectMeta.Name, metav1.DeleteOptions{})
		default:
			return nil, fmt.Errorf("unsupported type %T", objType)
		}
	})
	return err
}

// DeleteCollection apis
func (client *NVCClient) deleteCollectionWithRetry(obj interface{}) error {
	_, err := client.withRetry(func() (interface{}, error) {
		switch objType := obj.(type) {
		case *quarkv1alpha1.NetAppVolume:
			return nil, client.quarkClient.QuarkV1alpha1().NetAppVolumes(client.nvcNamespace).DeleteCollection(
				context.TODO(), metav1.DeleteOptions{}, metav1.ListOptions{})
		// Add cases for other types as needed
		default:
			return nil, fmt.Errorf("unsupported type %T", objType)
		}
	})
	return err
}
