package v2

import (
	"fmt"

	corev1 "k8s.io/api/core/v1"
	quarkv1alpha1 "quark.netapp.io/volume-controller/pkg/apis/volumecontroller/v1alpha1"
	"quark.netapp.io/volume-controller/pkg/quarkconfig"
	"quark.netapp.io/volume-controller/pkg/util"
)

type NVCClientOnlineStatus string

const (
	// object status constants
	NVCClientOnlineStatusOnline   NVCClientOnlineStatus = "Online"
	NVCClientOnlineStatusOffline  NVCClientOnlineStatus = "Offline"
	NVCClientOnlineStatusCreating NVCClientOnlineStatus = "Creating"

	//storagepool lifecycle states
	StoragePoolLifeCycleStateAvailable = "StoragePool is available"
	StoragePoolLifeCycleStateError     = "StoragePool is in error state"
	StoragePoolLifeCycleStateCreating  = "StoragePool is being created"

	//volume lifecycle states
	VolumeLifeCycleStateAvailable = "Volume is available"
	VolumeLifeCycleStateError     = "Volume is in error state"
	VolumeLifeCycleStateCreating  = "Volume is being created"
	VolumeLifeCycleStateRestoring = "Volume is being restored"
)

type NVCLifeCycleSubsystemStatus struct {
	Subsystem     string `json:"subsystem"`
	Status        bool   `json:"status"`
	Message       string `json:"message"`
	InternalError bool   `json:"internalError"`
}

type NVCLifeCycleStatus struct {
	Message          string                        `json:"message"`
	IsOnline         bool                          `json:"online"`
	SubsystemsStatus []NVCLifeCycleSubsystemStatus `json:"subsystemsStatus"`
}

func NewNVCLifeCycleOnline(message string, subsystemStatus []NVCLifeCycleSubsystemStatus) NVCLifeCycleStatus {
	return NVCLifeCycleStatus{message, true, subsystemStatus}
}

func NewNVCLifeCycleOffline(message string) NVCLifeCycleStatus {
	return NVCLifeCycleStatus{message, false, []NVCLifeCycleSubsystemStatus{}}
}

type NetAppStoragePoolInfo struct {
	StoragePool          *quarkv1alpha1.NetAppStoragePool
	StoragePoolLifeCycle NVCLifeCycleStatus
}

type NetAppVolumeInfo struct {
	Volume          *quarkv1alpha1.NetAppVolume
	VolumeLifeCycle NVCLifeCycleStatus
}

// NetAppStoragePoolGetInfo returns storagepool CR and its lifecycle details like state, online, all subsystems status.
func (client *NVCClient) NetAppStoragePoolGetInfo(storagePoolUUID string) (*NetAppStoragePoolInfo, error) {
	isOnline, message, storagePool, err := client.NetAppStoragePoolStatus(storagePoolUUID)
	if err != nil {
		return nil, fmt.Errorf("failed to get the NetAppStoragePoolStatus due to error: %v", err)
	}

	if isOnline {
		return &NetAppStoragePoolInfo{storagePool, NewNVCLifeCycleOnline(message, getAllSubsystemStatus(storagePool))}, nil
	}

	return &NetAppStoragePoolInfo{storagePool, NewNVCLifeCycleOffline(message)}, nil
}

// NetAppVolumeGetInfo returns volume CR and its lifecycle details like state, online, protocol
// corresponding subsystem status.
func (client *NVCClient) NetAppVolumeGetInfo(volumeUUID string) (*NetAppVolumeInfo, error) {
	volume, err := client.NetAppVolumeGet(volumeUUID)
	if err != nil {
		return nil, err
	}
	volumeLifeCycle, err := client.NetAppVolumeGetLifeCycle(volume)
	if err != nil {
		return nil, err
	}

	return &NetAppVolumeInfo{volume, volumeLifeCycle}, nil
}

func (client *NVCClient) NetAppVolumeGetLifeCycle(volume *quarkv1alpha1.NetAppVolume) (NVCLifeCycleStatus, error) {
	var (
		subsystemStatus []NVCLifeCycleSubsystemStatus
	)

	isOnline, message, storagePool, err := client.NetAppVolumeStatus(volume)
	if err != nil {
		return NVCLifeCycleStatus{}, fmt.Errorf("failed to get Volume Status due to error: %v", err)
	}

	if client.EnableGranularLiveness && storagePool.Status.GranularHealth != nil {
		// subsystemReady will be false if Storagepool is offline or subsystem is not ready
		subsystemName, subsystemReady, subsystemMessage := util.GetVolumeSubsystemStatus(storagePool, volume.Spec.Protocol)
		if subsystemReady != corev1.ConditionTrue {
			message, isOnline = subsystemMessage, false
			if message == "" {
				message = fmt.Sprintf("%s subsystem is not ready", subsystemName)
			}
		} else {
			// NFS status without LDAP is same as BOOT module status and we don't have NFS entry in Storagepool GranularStatus.
			// Hence returning BOOT status as NFS status by renaming subsystemName
			if subsystemName == quarkconfig.BootSubsystem {
				subsystemName = quarkconfig.NfsSubsystem
			}
			subsystemStatus = append(subsystemStatus, NVCLifeCycleSubsystemStatus{subsystemName, subsystemReady == corev1.ConditionTrue,
				util.TrimInternalSubsystemError(subsystemMessage), util.IsInternalSubsystemError(subsystemMessage)})
		}
	}

	if isOnline {
		return NewNVCLifeCycleOnline(message, subsystemStatus), nil
	}
	return NewNVCLifeCycleOffline(message), nil
}

// volumeOnlineConditionStatus returns status based on the Online Condition status in volume CRD.
func volumeOnlineConditionStatus(volume *quarkv1alpha1.NetAppVolume) (bool, string) {
	onlineConditionState, onlineConditionMessage := false, VolumeLifeCycleStateCreating

	for _, cond := range volume.Status.Conditions {
		if cond.Type == quarkv1alpha1.NetAppVolumeOnline {
			if cond.Status == corev1.ConditionTrue {
				onlineConditionState, onlineConditionMessage = true, VolumeLifeCycleStateAvailable
			} else {
				// Propogating the actual error to SDE so that it can be logged in SDE database.
				// SDE should not Propogate this error as it is. Instead it should use the NVCLifeCycleSubsystemStatus.
				onlineConditionState, onlineConditionMessage = false, cond.Message
			}
			break
		}
	}
	return onlineConditionState, onlineConditionMessage
}

// returns storage pool subsystems which are not healthy. It skips boot status as it's already part of
// storage pool lifecylce message.
func getAllSubsystemStatus(storagePool *quarkv1alpha1.NetAppStoragePool) []NVCLifeCycleSubsystemStatus {
	var (
		lifeCycleSubsystemStatus []NVCLifeCycleSubsystemStatus
	)

	for subsystemName, subsystemStatus := range storagePool.Status.GranularHealth {
		// NFS status without LDAP is same as BOOT module status and we don't have NFS entry in Storagepool GranularStatus.
		// Hence returning BOOT status as NFS status by renaming subsystemName
		if subsystemName == quarkconfig.BootSubsystem {
			subsystemName = quarkconfig.NfsSubsystem
		}
		lifeCycleSubsystemStatus = append(lifeCycleSubsystemStatus, NVCLifeCycleSubsystemStatus{subsystemName, subsystemStatus.Status == corev1.ConditionTrue,
			util.TrimInternalSubsystemError(subsystemStatus.Message), util.IsInternalSubsystemError(subsystemStatus.Message)})
	}

	return lifeCycleSubsystemStatus
}
