package v2

import (
	"context"
	"fmt"
	"reflect"
	"strings"

	uuid "github.com/satori/go.uuid"
	log "github.com/sirupsen/logrus"
	corev1 "k8s.io/api/core/v1"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/apimachinery/pkg/labels"
	quarkv1alpha1 "quark.netapp.io/volume-controller/pkg/apis/volumecontroller/v1alpha1"
	"quark.netapp.io/volume-controller/pkg/quarkconfig"
	"quark.netapp.io/volume-controller/pkg/util"
)

// NetAppVolumeReplicationDelete is used to signal the deletion of a NetAppVolumeReplication.
type NetAppVolumeReplicationDelete struct {
	VolumeReplication *quarkv1alpha1.NetAppVolumeReplication
}

// NetAppVolumeReplicationUpdate is used to signal the most recent state of a NetAppVolumeReplicationUpdate and the
// most recent changes.
type NetAppVolumeReplicationUpdate struct {
	VolumeReplication *quarkv1alpha1.NetAppVolumeReplication
	Status            *NetAppVolumeReplicationStatusInfo
	UpdatedConditions []quarkv1alpha1.NetAppVolumeReplicationCondition
}

// NetAppVolumeReplicationStatusInfo captures the status of any pending updates
type NetAppVolumeReplicationStatusInfo struct {
	MirrorStateUpdated         bool
	ReplicationScheduleUpdated bool
	ReplicationStatusUpdated   bool
}

// ConstructNetAppVolumeReplication creates a NetAppVolumeReplication struct resource
func ConstructNetAppVolumeReplication(
	ownerID, volumeReplicationUUID, storagePoolUUID, clusterPeeringUUID, displayName, destinationUUID, sourceUUID string,
	replicationPolicy quarkv1alpha1.NetAppVolumeReplicationPolicyType,
	replicationSchedule quarkv1alpha1.NetAppVolumeReplicationScheduleType,
	mirrorState quarkv1alpha1.NetAppVolumeReplicationMirrorStateType,
) (volumeReplication *quarkv1alpha1.NetAppVolumeReplication, err error) {
	// validate volumeReplicationUUID
	_, err = uuid.FromString(volumeReplicationUUID)
	if err != nil {
		return nil, fmt.Errorf("invalid value for volumeReplicationUUID: %v", err)
	}

	_, err = uuid.FromString(storagePoolUUID)
	if err != nil {
		return nil, fmt.Errorf("invalid value for storagePoolUUID: %v", err)
	}

	_, err = uuid.FromString(clusterPeeringUUID)
	if err != nil {
		return nil, fmt.Errorf("invalid value for clusterPeeringUUID: %v", err)
	}

	_, err = uuid.FromString(destinationUUID)
	if err != nil {
		return nil, fmt.Errorf("invalid value for destinationUUID: %v", err)
	}

	_, err = uuid.FromString(sourceUUID)
	if err != nil {
		return nil, fmt.Errorf("invalid value for sourceUUID: %v", err)
	}

	if len(strings.TrimSpace(ownerID)) == 0 {
		return nil, fmt.Errorf("ownerID cannot be empty string")
	}

	if len(strings.TrimSpace(displayName)) == 0 {
		return nil, fmt.Errorf("display name cannot be empty string")
	}

	if err = util.IsValidReplicationPolicy(replicationPolicy); err != nil {
		return nil, err
	}

	if err = util.IsValidReplicationSchedule(replicationSchedule); err != nil {
		return nil, err
	}

	if mirrorState != quarkv1alpha1.MirrorStateUninitialized &&
		mirrorState != quarkv1alpha1.MirrorStateBroken {
		return nil, fmt.Errorf("invalid mirror state for create: %v. Valid mirror states for create: %v, %v",
			mirrorState, quarkv1alpha1.MirrorStateUninitialized, quarkv1alpha1.MirrorStateBroken)
	}

	volumeReplication = &quarkv1alpha1.NetAppVolumeReplication{
		TypeMeta: metav1.TypeMeta{},
		ObjectMeta: metav1.ObjectMeta{
			Name: volumeReplicationUUID,
			Labels: map[string]string{
				quarkconfig.OwnerIDLabel:        ownerID,
				quarkconfig.StoragePoolLabel:    storagePoolUUID,
				quarkconfig.ClusterPeeringLabel: clusterPeeringUUID,
			},
			Finalizers: []string{
				quarkconfig.VolumeReplicationFinalizer,
			},
		},
		Spec: quarkv1alpha1.NetAppVolumeReplicationSpec{
			DisplayName: displayName,
			DestinationVolume: quarkv1alpha1.NetAppVolumeReplicationVolume{
				VolumeUUID: destinationUUID,
			},
			SourceVolume: quarkv1alpha1.NetAppVolumeReplicationVolume{
				VolumeUUID: sourceUUID,
			},
			Destination: quarkv1alpha1.NetAppVolumeReplicationCandidate{
				UUID: destinationUUID,
			},
			Source: quarkv1alpha1.NetAppVolumeReplicationCandidate{
				UUID: sourceUUID,
			},
			MirrorState:         mirrorState,
			ReplicationPolicy:   replicationPolicy,
			ReplicationSchedule: replicationSchedule,
		},
	}
	return
}

// ConstructNetAppVolumeReplicationForBackup creates a NetAppVolumeReplication struct resource for backup usecase.
// sourceUUID is the source volume uuid and destinationUUID is the endpoint uuid. SDE will create the endpoint UUID
// on the object store. Endpoint UUID and a source volume have a 1-1 mapping. All backups of a source volume will be
// added to the endpoint UUID.
func ConstructNetAppVolumeReplicationForBackup(
	ownerID string,
	displayName string,
	volumeReplicationUUID uuid.UUID,
	storagePoolUUID uuid.UUID,
	objectStoreUUID uuid.UUID,
	sourceUUID uuid.UUID, // source volume uuid
	destinationUUID uuid.UUID, // endpointUUID
	snapshotToTransfer string,
) (volumeReplication *quarkv1alpha1.NetAppVolumeReplication, err error) {
	if len(strings.TrimSpace(ownerID)) == 0 {
		return nil, fmt.Errorf("ownerID cannot be empty string")
	}

	if len(strings.TrimSpace(displayName)) == 0 {
		return nil, fmt.Errorf("display name cannot be empty string")
	}

	volumeReplication = &quarkv1alpha1.NetAppVolumeReplication{
		TypeMeta: metav1.TypeMeta{},
		ObjectMeta: metav1.ObjectMeta{
			Name: volumeReplicationUUID.String(),
			Labels: map[string]string{
				quarkconfig.OwnerIDLabel:     ownerID,
				quarkconfig.StoragePoolLabel: storagePoolUUID.String(),
				quarkconfig.ObjectStoreLabel: objectStoreUUID.String(),
			},
		},
		Spec: quarkv1alpha1.NetAppVolumeReplicationSpec{
			DisplayName: displayName,
			Destination: quarkv1alpha1.NetAppVolumeReplicationCandidate{
				UUID: destinationUUID.String(),
			},
			Source: quarkv1alpha1.NetAppVolumeReplicationCandidate{
				UUID: sourceUUID.String(),
			},
			MirrorState:          quarkv1alpha1.MirrorStateMirrored, // state is always "Mirrored" for Backup case
			ReplicationType:      quarkv1alpha1.ReplicationTypeBackup,
			SnapshotToTransfer:   snapshotToTransfer,
			AbortTransferCounter: 0,
		},
	}
	return
}

// ConstructNetAppVolumeReplicationForRestore creates a NetAppVolumeReplication struct resource for restore usecase.
// sourceUUID is the endpoint uuid and destinationUUID is the destination volume. SDE will create the endpoint UUID
// on the object store. Endpoint UUID should already exist, and should have been created during the backup operation
// initiated by SDE.
// Endpoint UUID and a source volume have a 1-1 mapping. All restores of a destination volume will
// happen from the same endpoint UUID.
func ConstructNetAppVolumeReplicationForRestore(
	ownerID string,
	displayName string,
	volumeReplicationUUID uuid.UUID,
	storagePoolUUID uuid.UUID,
	objectStoreUUID uuid.UUID,
	sourceUUID uuid.UUID, // endpoint uuid
	destinationUUID uuid.UUID, // destination volume uuid
	snapshotToTransfer string,
) (volumeReplication *quarkv1alpha1.NetAppVolumeReplication, err error) {
	if len(strings.TrimSpace(ownerID)) == 0 {
		return nil, fmt.Errorf("ownerID cannot be empty string")
	}

	if len(strings.TrimSpace(displayName)) == 0 {
		return nil, fmt.Errorf("display name cannot be empty string")
	}

	volumeReplication = &quarkv1alpha1.NetAppVolumeReplication{
		TypeMeta: metav1.TypeMeta{},
		ObjectMeta: metav1.ObjectMeta{
			Name: volumeReplicationUUID.String(),
			Labels: map[string]string{
				quarkconfig.OwnerIDLabel:     ownerID,
				quarkconfig.StoragePoolLabel: storagePoolUUID.String(),
				quarkconfig.ObjectStoreLabel: objectStoreUUID.String(),
			},
		},
		Spec: quarkv1alpha1.NetAppVolumeReplicationSpec{
			DisplayName: displayName,
			Destination: quarkv1alpha1.NetAppVolumeReplicationCandidate{
				UUID: destinationUUID.String(),
			},
			Source: quarkv1alpha1.NetAppVolumeReplicationCandidate{
				UUID: sourceUUID.String(),
			},
			MirrorState:          quarkv1alpha1.MirrorStateRestored,
			ReplicationType:      quarkv1alpha1.ReplicationTypeRestore,
			SnapshotToTransfer:   snapshotToTransfer,
			AbortTransferCounter: 0,
		},
	}
	return
}

// NetAppVolumeReplicationCreate attempts to create a NetAppVolumeReplication CR
func (client *NVCClient) NetAppVolumeReplicationCreate(volumeReplication *quarkv1alpha1.NetAppVolumeReplication) (*quarkv1alpha1.NetAppVolumeReplication, error) {
	nvrInterface, err := NewNVRInstance(volumeReplication, client)
	if err != nil {
		return nil, err
	}

	return nvrInterface.Create()
}

// NetAppVolumeReplicationUpdate attempts to update a NetAppVolumeReplication CR
// Error returned by this function should not be overwritten as it has the logic to
// create appropriate concrete error types like StaleVersionError and
// clients (SDE) depend on those concrete error types
func (client *NVCClient) NetAppVolumeReplicationUpdate(volumeReplication *quarkv1alpha1.NetAppVolumeReplication) (*quarkv1alpha1.NetAppVolumeReplication, error) {
	nvrInterface, err := NewNVRInstance(volumeReplication, client)
	if err != nil {
		return nil, err
	}

	volRepl, err := nvrInterface.Update()
	if err != nil {
		if strings.Contains(err.Error(), quarkconfig.OldVersionErr) {
			return nil, StaleVersionError{}
		}
		return nil, err
	}
	return volRepl, nil
}

// NetAppVolumeReplicationDelete attempts to delete a NetAppVolumeReplication CR only if the relationship is broken
func (client *NVCClient) NetAppVolumeReplicationDelete(volumeReplicationUUID string) error {
	nvrInterface, err := NewNVRInstanceFromUUID(volumeReplicationUUID, client)
	if err != nil {
		return err
	}

	return nvrInterface.Delete()
}

// NetAppVolumeReplicationGet attempts to retrieve a NetAppVolumeReplication CR
func (client *NVCClient) NetAppVolumeReplicationGet(volumeReplicationUUID string) (*quarkv1alpha1.NetAppVolumeReplication, error) {
	return client.quarkClient.QuarkV1alpha1().NetAppVolumeReplications(client.nvcNamespace).Get(
		context.TODO(), volumeReplicationUUID, metav1.GetOptions{})
}

// NetAppVolumeReplicationGetAll attempts to retrieve all the NetAppVolumeReplication CRs with replication type either
// unset or set to CRR. NVRs with unset replication types are also for CRR.
// Deprecation warning: This api will be deprecated in future and clients are expected to move to using
// NetAppVolumeReplicationGetAllByReplicationType API instead.
func (client *NVCClient) NetAppVolumeReplicationGetAll() ([]*quarkv1alpha1.NetAppVolumeReplication, error) {
	volumeReplicationsList, err := client.quarkClient.QuarkV1alpha1().NetAppVolumeReplications(client.nvcNamespace).List(
		context.TODO(), metav1.ListOptions{})
	if err != nil {
		return nil, err
	}
	replSlice, err := netAppVolumeReplicationListToSlice(volumeReplicationsList), nil
	if err != nil {
		return nil, err
	}

	filteredNVRsByType := []*quarkv1alpha1.NetAppVolumeReplication{}
	for _, nvr := range replSlice {
		if nvr.Spec.ReplicationType == quarkv1alpha1.ReplicationTypeCRR ||
			nvr.Spec.ReplicationType == "" {
			filteredNVRsByType = append(filteredNVRsByType, nvr)
		}
	}
	return filteredNVRsByType, nil
}

func (client *NVCClient) NetAppVolumeReplicationGetAllByReplicationType(replType quarkv1alpha1.NetAppVolumeReplicationType) ([]*quarkv1alpha1.NetAppVolumeReplication, error) {
	volumeReplicationsList, err := client.quarkClient.QuarkV1alpha1().NetAppVolumeReplications(client.nvcNamespace).List(
		context.TODO(), metav1.ListOptions{})
	if err != nil {
		return nil, err
	}
	replSlice, err := netAppVolumeReplicationListToSlice(volumeReplicationsList), nil
	if err != nil {
		return nil, err
	}

	filteredNVRsByType := []*quarkv1alpha1.NetAppVolumeReplication{}
	for _, nvr := range replSlice {
		if nvr.Spec.ReplicationType == replType {
			filteredNVRsByType = append(filteredNVRsByType, nvr)
		}
	}
	return filteredNVRsByType, nil
}

func (client *NVCClient) NetAppVolumeReplicationFilterByClusterPeering(clusterPeeringName string) ([]*quarkv1alpha1.NetAppVolumeReplication, error) {
	label := map[string]string{quarkconfig.ClusterPeeringLabel: clusterPeeringName}
	volumeReplicationsList, err := client.NetAppVolumeReplicationFilterByLabel(label)
	if err != nil {
		return nil, err
	}
	return netAppVolumeReplicationListToSlice(volumeReplicationsList), nil
}

func (client *NVCClient) NetAppVolumeReplicationFilterByStoragePool(storagePoolName string) ([]*quarkv1alpha1.NetAppVolumeReplication, error) {
	label := map[string]string{quarkconfig.StoragePoolLabel: storagePoolName}
	volumeReplicationsList, err := client.NetAppVolumeReplicationFilterByLabel(label)
	if err != nil {
		return nil, err
	}
	return netAppVolumeReplicationListToSlice(volumeReplicationsList), nil
}

func (client *NVCClient) NetAppVolumeReplicationFilterByObjectStoreConfig(objectStoreName string) ([]*quarkv1alpha1.NetAppVolumeReplication, error) {
	label := map[string]string{quarkconfig.ObjectStoreLabel: objectStoreName}
	volumeReplicationsList, err := client.NetAppVolumeReplicationFilterByLabel(label)
	if err != nil {
		return nil, err
	}
	return netAppVolumeReplicationListToSlice(volumeReplicationsList), nil
}

func (client *NVCClient) NetAppVolumeReplicationFilterByLabel(filterLabels map[string]string) (*quarkv1alpha1.NetAppVolumeReplicationList, error) {
	labelSelector := metav1.LabelSelector{MatchLabels: filterLabels}
	return client.quarkClient.QuarkV1alpha1().NetAppVolumeReplications(client.nvcNamespace).List(
		context.TODO(), metav1.ListOptions{LabelSelector: labels.Set(labelSelector.MatchLabels).String()})
}

func netAppVolumeReplicationListToSlice(list *quarkv1alpha1.NetAppVolumeReplicationList) []*quarkv1alpha1.NetAppVolumeReplication {
	volumeReplications := make([]*quarkv1alpha1.NetAppVolumeReplication, len(list.Items))
	for i := range list.Items {
		volumeReplications[i] = &list.Items[i]
	}
	return volumeReplications
}

// NetAppVolumeReplicationMirrorState returns the MirrorState of the NetAppVolumeReplication resource
func NetAppVolumeReplicationMirrorState(volumeReplication *quarkv1alpha1.NetAppVolumeReplication) quarkv1alpha1.NetAppVolumeReplicationMirrorStateType {
	return volumeReplication.Status.MirrorState
}

// NetAppVolumeReplicationUninitialized checks if the Replication relationship is Uninitialized
func NetAppVolumeReplicationUninitialized(volumeReplication *quarkv1alpha1.NetAppVolumeReplication) bool {
	return (volumeReplication.Status.MirrorState == quarkv1alpha1.MirrorStateUninitialized)
}

// NetAppVolumeReplicationMirrored checks if the Replication relationship is "Mirrored"
func NetAppVolumeReplicationMirrored(volumeReplication *quarkv1alpha1.NetAppVolumeReplication) bool {
	return (volumeReplication.Status.MirrorState == quarkv1alpha1.MirrorStateMirrored)
}

// NetAppVolumeReplicationRestored checks if the Replication relationship is "Restored"
func NetAppVolumeReplicationRestored(volumeReplication *quarkv1alpha1.NetAppVolumeReplication) bool {
	return (volumeReplication.Status.MirrorState == quarkv1alpha1.MirrorStateRestored)
}

// NetAppVolumeReplicationBroken checks if the Replication relationship is Broken
func NetAppVolumeReplicationBroken(volumeReplication *quarkv1alpha1.NetAppVolumeReplication) bool {
	return (volumeReplication.Status.MirrorState == quarkv1alpha1.MirrorStateBroken)
}

// NetAppVolumeReplicationStatus returns the MirrorState of the NetAppVolumeReplication resource
// Expected values are one of ReplicationStatusIdle, ReplicationStatusTransferring
func NetAppVolumeReplicationStatus(volumeReplication *quarkv1alpha1.NetAppVolumeReplication) quarkv1alpha1.NetAppVolumeReplicationStatusType {
	return volumeReplication.Status.ReplicationStatus
}

// NetAppVolumeReplicationScheduleUpdated checks if the Spec and the Status ReplicationSchedule field match
// A match would indicate a successful update
func NetAppVolumeReplicationScheduleUpdated(volumeReplication *quarkv1alpha1.NetAppVolumeReplication) bool {
	return volumeReplication.Spec.ReplicationSchedule == volumeReplication.Status.ReplicationSchedule
}

// NetAppVolumeReplicationHealthy checks if the relationship is healthy
func NetAppVolumeReplicationHealthy(volumeReplication *quarkv1alpha1.NetAppVolumeReplication) bool {
	for _, cond := range volumeReplication.Status.Conditions {
		if cond.Type == quarkv1alpha1.VolumeReplicationHealthy && cond.Status == corev1.ConditionTrue {
			return true
		}
	}
	return false
}

// NetAppVolumeReplicationInResyncMode checks if relationship is in resync mode
func NetAppVolumeReplicationInResyncMode(volumeReplication *quarkv1alpha1.NetAppVolumeReplication) bool {
	return (volumeReplication.Spec.MirrorState == quarkv1alpha1.MirrorStateMirrored && volumeReplication.Status.MirrorState == quarkv1alpha1.MirrorStateBroken)
}

// NetAppVolumeReplicationTransferSnapshotError checks if the current transfer of the snapshot failed
func NetAppVolumeReplicationTransferSnapshotError(volumeReplication *quarkv1alpha1.NetAppVolumeReplication) error {
	for _, cond := range volumeReplication.Status.Conditions {
		if cond.Type == quarkv1alpha1.VolumeReplicationTransferError && cond.Status == corev1.ConditionTrue {
			return fmt.Errorf(cond.Message)
		}
	}
	return nil
}

// NetAppVolumeReplicationTransferSnapshotAborted checks if a given snapshot transfer is aborted by checking the appropriate condition
func NetAppVolumeReplicationTransferSnapshotAborted(volumeReplication *quarkv1alpha1.NetAppVolumeReplication) bool {
	for _, cond := range volumeReplication.Status.Conditions {
		if cond.Type == quarkv1alpha1.VolumeReplicationTransferAborted && cond.Status == corev1.ConditionTrue {
			return true
		}
	}
	return false
}

// NetAppVolumeReplicationInitialize is used to start the initialize operation
// Initialization is done automatically, but can also be driven by the CVS API in case it initially failed.
func (client *NVCClient) NetAppVolumeReplicationInitialize(volumeReplicationUUID, transferSnapshot string) error {
	nvrInterface, err := NewNVRInstanceFromUUID(volumeReplicationUUID, client)
	if err != nil {
		return err
	}

	return nvrInterface.Initialize(transferSnapshot)
}

// NetAppVolumeReplicationResync is used to start a resync operation
// Resync is performed when the user wants to resume mirroring data after a Break.
// Resync can only be done on a Broken relationship.
func (client *NVCClient) NetAppVolumeReplicationResync(volumeReplicationUUID string) error {
	nvrInterface, err := NewNVRInstanceFromUUID(volumeReplicationUUID, client)
	if err != nil {
		return err
	}

	return nvrInterface.Resync()
}

// NetAppVolumeReplicationBreak is used to start the break operation
// Break is performed when the user wants to read/write to the secondary volume.
// This might be done, for example, if the primary region is down. Break can only be done on a Mirrored relationship.
func (client *NVCClient) NetAppVolumeReplicationBreak(volumeReplicationUUID string) error {
	nvrInterface, err := NewNVRInstanceFromUUID(volumeReplicationUUID, client)
	if err != nil {
		return err
	}

	return nvrInterface.Break()
}

// NetAppVolumeReplicationTransferSnapshot is used to start the transfer of a snapshot identified by name.
// Transfer Snapshot is initiated internally by the Replication Scheduler in SDE.
// It will be performed every 10 minutes, hourly, or daily based on the schedule chosen by the customer.
// NOTE: Transfer snapshot is not supported for the first release of CRR.
func (client *NVCClient) NetAppVolumeReplicationTransferSnapshot(volumeReplicationUUID, snapshotName string) error {
	nvrInterface, err := NewNVRInstanceFromUUID(volumeReplicationUUID, client)
	if err != nil {
		return err
	}

	return nvrInterface.TransferSnapshot(snapshotName)
}

// NetAppVolumeReplicationTransferAbort is used to abort a transfer that is started by NetAppVolumeReplicationTransferSnapshot()
// NOTE: Aborting transfer is not supported for the first release
func (client *NVCClient) NetAppVolumeReplicationTransferAbort(volumeReplicationUUID string) error {
	nvrInterface, err := NewNVRInstanceFromUUID(volumeReplicationUUID, client)
	if err != nil {
		return err
	}

	return nvrInterface.TransferAbort()
}

// NetAppVolumeReplicationScheduleModify is used to modify the ReplicationSchedule
// Accepted values are ReplicationSchedule10Minutely, ReplicationScheduleHourly and ReplicationScheduleDaily
func (client *NVCClient) NetAppVolumeReplicationScheduleModify(volumeReplicationUUID string, schedule quarkv1alpha1.NetAppVolumeReplicationScheduleType) error {
	nvrInterface, err := NewNVRInstanceFromUUID(volumeReplicationUUID, client)
	if err != nil {
		return err
	}

	return nvrInterface.ScheduleModify(schedule)
}

// NetAppVolumeReplicationUpdated checks for specific modifications/ updates to the NetAppVolumeReplication CR
func (client *NVCClient) NetAppVolumeReplicationUpdated(volumeReplication *quarkv1alpha1.NetAppVolumeReplication) *NetAppVolumeReplicationStatusInfo {
	volumeReplicationUpdate := &NetAppVolumeReplicationStatusInfo{
		MirrorStateUpdated:         true,
		ReplicationScheduleUpdated: true,
	}

	if volumeReplication.Spec.MirrorState != volumeReplication.Status.MirrorState {
		volumeReplicationUpdate.MirrorStateUpdated = false
	}
	if volumeReplication.Spec.ReplicationSchedule != volumeReplication.Status.ReplicationSchedule {
		volumeReplicationUpdate.ReplicationScheduleUpdated = false
	}
	log.WithFields(log.Fields{
		"MirrorStateUpdated":         volumeReplicationUpdate.MirrorStateUpdated,
		"ReplicationScheduleUpdated": volumeReplicationUpdate.ReplicationScheduleUpdated,
	}).Debug("NetAppVolumeReplication state has been updated.")
	return volumeReplicationUpdate
}

// handleNetappVolumeReplicationUpdate is used to publish NetAppVolumeReplication specific events
// This is done by comparing the previous and current state of the NetAppVolumeReplication CR
func (client *NVCClient) handleNetappVolumeReplicationUpdate(
	oldVolumeReplication *quarkv1alpha1.NetAppVolumeReplication,
	newVolumeReplication *quarkv1alpha1.NetAppVolumeReplication) {
	sendEvent := false
	newVolumeReplicationUpdate := client.NetAppVolumeReplicationUpdated(newVolumeReplication)
	oldVolumeReplicationUpdate := client.NetAppVolumeReplicationUpdated(oldVolumeReplication)
	if newVolumeReplicationUpdate != oldVolumeReplicationUpdate {
		log.WithFields(log.Fields{
			"volumeReplication": newVolumeReplication.Name,
			"oldStatus":         oldVolumeReplicationUpdate,
			"newStatus":         newVolumeReplicationUpdate,
		}).Debug("NetApp Volume Replication state has been updated.")
		sendEvent = true
	}
	// We send events when the NetAppVolumeReplications's ReplicationStatus field has changed.
	if newVolumeReplication.Status.ReplicationStatus != oldVolumeReplication.Status.ReplicationStatus {
		sendEvent = true
		newVolumeReplicationUpdate.ReplicationStatusUpdated = true
		log.WithFields(log.Fields{
			"volumeReplication": newVolumeReplication.Name,
			"oldStatus":         oldVolumeReplicationUpdate,
			"newStatus":         newVolumeReplicationUpdate,
		}).Debug("NetApp Volume Replication has updated ReplicationStatus.")
	}
	// We send events when the NetAppVolumeReplications's conditions have changed.
	if !reflect.DeepEqual(oldVolumeReplication.Status.Conditions, newVolumeReplication.Status.Conditions) {
		sendEvent = true
		log.WithFields(log.Fields{
			"volumeReplication": newVolumeReplication.Name,
			"conditions":        newVolumeReplication.Status.Conditions,
		}).Debug("NetApp Volume Replication has updated conditions.")

	}
	if sendEvent {
		client.volumeReplicationUpdateChannel <- &NetAppVolumeReplicationUpdate{
			VolumeReplication: newVolumeReplication,
			Status:            newVolumeReplicationUpdate,
			UpdatedConditions: newVolumeReplication.Status.Conditions,
		}
	}
}

// handleVolumeReplicationDelete is used to terminate the watch on VolumeReplication events
func (client *NVCClient) handleVolumeReplicationDelete(volumeReplication *quarkv1alpha1.NetAppVolumeReplication) {
	client.volumeReplicationDeleteChannel <- &NetAppVolumeReplicationDelete{VolumeReplication: volumeReplication}
}
