package v2

import (
	"context"
	"fmt"

	log "github.com/sirupsen/logrus"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	quarkv1alpha1 "quark.netapp.io/volume-controller/pkg/apis/volumecontroller/v1alpha1"
	"quark.netapp.io/volume-controller/pkg/quarkconfig"
)

type nvrCrr struct {
	nvrCommonImpl
}

func (r *nvrCrr) Create() (*quarkv1alpha1.NetAppVolumeReplication, error) {
	storagePoolUUID := r.nvr.Labels[quarkconfig.StoragePoolLabel]
	_, err := r.nvcClient.NetAppStoragePoolGet(storagePoolUUID)
	if err != nil {
		return nil, fmt.Errorf("error getting storage pool %v: %v", storagePoolUUID, err)
	}

	clusterPeeringUUID := r.nvr.Labels[quarkconfig.ClusterPeeringLabel]
	_, err = r.nvcClient.NetAppClusterPeeringGet(clusterPeeringUUID)
	if err != nil {
		return nil, fmt.Errorf("error getting cluster peering %v: %v", clusterPeeringUUID, err)
	}

	dstVolUUID := r.nvr.Spec.DestinationVolume.VolumeUUID
	dstVol, err := r.nvcClient.NetAppVolumeGet(dstVolUUID)
	if err != nil {
		return nil, fmt.Errorf("error getting destination volume %v: %v", dstVolUUID, err)
	}

	// Check if destination volume belongs to the storagepool supplied in the replication cr
	if dstVol.Labels[quarkconfig.StoragePoolLabel] != storagePoolUUID {
		return nil, fmt.Errorf("destination volume storagepool (%v) not the same as storagepool supplied in cr (%v)", dstVol.Labels[quarkconfig.StoragePoolLabel], storagePoolUUID)
	}

	// dest volume should be dp enabled and not read/write if the volume replication is being created in Uninitialized state
	if r.nvr.Spec.MirrorState == quarkv1alpha1.MirrorStateUninitialized &&
		dstVol.Spec.DataProtection != quarkv1alpha1.DataProtectionEnabled {
		return nil, fmt.Errorf("volume replication mirror state cannot be set to \"Uninitialized\" since the destination volume %s is not data protection enabled", dstVolUUID)
	}

	// dest volume needs to be read/write if the volume replication is being created in Broken state i.e flip resync scenario
	if r.nvr.Spec.MirrorState == quarkv1alpha1.MirrorStateBroken &&
		dstVol.Spec.DataProtection == quarkv1alpha1.DataProtectionEnabled {
		return nil, fmt.Errorf("volume replication mirror state cannot be set to \"Broken\" since the destination volume %s has data protection enabled",
			dstVolUUID)
	}

	return r.nvcClient.quarkClient.QuarkV1alpha1().NetAppVolumeReplications(r.nvcClient.nvcNamespace).Create(
		context.TODO(), r.nvr, metav1.CreateOptions{})
}

func (r *nvrCrr) Delete() error {
	if !NetAppVolumeReplicationBroken(r.nvr) && !NetAppVolumeReplicationUninitialized(r.nvr) {
		return fmt.Errorf("unable to delete volume replication in mirror state \"%v\": "+
			"volume replication's mirror state must be one of the follow to be deleted: \"%v\"",
			r.nvr.Status.MirrorState,
			fmt.Sprintf("%v, %v", quarkv1alpha1.MirrorStateBroken, quarkv1alpha1.MirrorStateUninitialized))
	}
	return r.nvcClient.quarkClient.QuarkV1alpha1().NetAppVolumeReplications(r.nvcClient.nvcNamespace).Delete(
		context.TODO(), r.nvr.Name, metav1.DeleteOptions{})
}

func (r *nvrCrr) Initialize(transferSnapshot string) error {
	if r.nvr.Spec.MirrorState == quarkv1alpha1.MirrorStateMirrored &&
		r.nvr.Spec.TransferSnapshot == transferSnapshot {
		return nil
	}
	if r.nvr.Status.MirrorState != quarkv1alpha1.MirrorStateUninitialized {
		return fmt.Errorf("volume replication state must be \"%v\". Current state is \"%v\"",
			quarkv1alpha1.MirrorStateUninitialized, r.nvr.Status.MirrorState)
	}

	volumeReplicationCopy := r.nvr.DeepCopy()
	volumeReplicationCopy.Spec.MirrorState = quarkv1alpha1.MirrorStateMirrored
	volumeReplicationCopy.Spec.TransferSnapshot = transferSnapshot
	_, err := r.nvcClient.NetAppVolumeReplicationUpdate(volumeReplicationCopy)
	return err
}

func (r *nvrCrr) Resync() error {
	if r.nvr.Spec.MirrorState == quarkv1alpha1.MirrorStateMirrored {
		return nil
	}
	if !NetAppVolumeReplicationBroken(r.nvr) {
		return fmt.Errorf("volume replication must be %v", quarkv1alpha1.MirrorStateBroken)
	}

	// Set volume's Spec.dataProtection field to true for resync
	err := r.nvcClient.netAppVolumeDataProtectionUpdate(
		r.nvr.Spec.DestinationVolume.VolumeUUID,
		quarkv1alpha1.DataProtectionEnabled)
	if err != nil {
		// netAppVolumeDataProtectionUpdate could return StaleVersionError which should not
		// be overwritten here as clients (SDE) depend on the concrete error type.
		log.WithFields(log.Fields{
			"volume":            r.nvr.Spec.DestinationVolume.VolumeUUID,
			"volumeReplication": r.nvr.Name,
			"dataProtection":    quarkv1alpha1.DataProtectionEnabled,
			"error":             err.Error(),
		}).Error("Unable to update dataProtection for destination volume.")
		return err
	}

	volumeReplicationCopy := r.nvr.DeepCopy()
	volumeReplicationCopy.Spec.MirrorState = quarkv1alpha1.MirrorStateMirrored
	_, err = r.nvcClient.NetAppVolumeReplicationUpdate(volumeReplicationCopy)
	if err != nil {
		// Set dataProtection field to false on the volume
		// Ignore any errors since the operation to resync would likely be retried
		if volErr := r.nvcClient.netAppVolumeDataProtectionUpdate(
			r.nvr.Spec.DestinationVolume.VolumeUUID,
			quarkv1alpha1.DataProtectionDisabled); volErr != nil {
			log.WithFields(log.Fields{
				"volume":            r.nvr.Spec.DestinationVolume.VolumeUUID,
				"volumeReplication": r.nvr.Name,
				"dataProtection":    quarkv1alpha1.MirrorStateMirrored,
				"error":             volErr.Error(),
			}).Warn("Unable to update dataProtection for volume.")
		}
	}
	return err
}

func (r *nvrCrr) Break() error {
	if r.nvr.Spec.MirrorState == quarkv1alpha1.MirrorStateBroken {
		return nil
	}

	if !NetAppVolumeReplicationMirrored(r.nvr) {
		// Per SDE's request via burt-1599543: If CRR relationship is in re-sync mode for a while and SDE
		// determines transfer is stuck, they will want to abort the resync operation.
		// Resync is aborted by setting spec back to "Broken" state and ccpd will respond accordingly.
		// So we allow SDE to break the relationship while transfer is in progress.
		// Transfer in progress is ensured by SDE so we intentionally do not validate this condition
		// at the nvcclient layer before honoring this request.
		if NetAppVolumeReplicationInResyncMode(r.nvr) {
			log.WithFields(log.Fields{
				"volume":            r.nvr.Spec.DestinationVolume.VolumeUUID,
				"volumeReplication": r.nvr.Name,
			}).Info("Replication relationship in resync mode, allowing break operation.")
		} else {
			return fmt.Errorf("to break relationship, volume replication state must be \"%v\". Current state is \"%v\"",
				quarkv1alpha1.MirrorStateMirrored, r.nvr.Status.MirrorState)
		}
	}

	// Set volume's Spec.dataProtection field to false
	err := r.nvcClient.netAppVolumeDataProtectionUpdate(
		r.nvr.Spec.DestinationVolume.VolumeUUID,
		quarkv1alpha1.DataProtectionDisabled)
	if err != nil {
		// netAppVolumeDataProtectionUpdate could return StaleVersionError which should not
		// be overwritten here as clients (SDE) depend on the concrete error type.
		log.WithFields(log.Fields{
			"volume":            r.nvr.Spec.DestinationVolume.VolumeUUID,
			"volumeReplication": r.nvr.Name,
			"dataProtection":    quarkv1alpha1.DataProtectionDisabled,
			"error":             err.Error(),
		}).Error("Unable to update dataProtection for destination volume.")
		return err
	}

	volumeReplicationCopy := r.nvr.DeepCopy()
	volumeReplicationCopy.Spec.MirrorState = quarkv1alpha1.MirrorStateBroken
	_, err = r.nvcClient.NetAppVolumeReplicationUpdate(volumeReplicationCopy)
	if err != nil {
		// Set dataProtection field to true on the volume
		// Ignore any errors since the break operation would likely be retried
		if volErr := r.nvcClient.netAppVolumeDataProtectionUpdate(
			r.nvr.Spec.DestinationVolume.VolumeUUID,
			quarkv1alpha1.DataProtectionEnabled); volErr != nil {
			log.WithFields(log.Fields{
				"volume":            r.nvr.Spec.DestinationVolume.VolumeUUID,
				"volumeReplication": r.nvr.Name,
				"error":             volErr.Error(),
				"dataProtection":    quarkv1alpha1.MirrorStateBroken,
			}).Warn("Unable to update dataProtection for volume.")
		}
	}
	return err
}

// Not used in CRR case, as we rely on schedule within dmap to transfer snapshots
func (r *nvrCrr) TransferSnapshot(snapshotName string) error {
	log.WithFields(log.Fields{
		"volume":            r.nvr.Spec.DestinationVolume.VolumeUUID,
		"volumeReplication": r.nvr.Name,
	}).Warn("API is a no-op for CRR case currently. Snapshots tranferred using dmap schedule")
	return nil
}

func (r *nvrCrr) TransferAbort() error {
	if r.nvr.Spec.TransferSnapshot == "" {
		return nil
	}
	volumeReplicationCopy := r.nvr.DeepCopy()
	volumeReplicationCopy.Spec.TransferSnapshot = ""
	_, err := r.nvcClient.NetAppVolumeReplicationUpdate(volumeReplicationCopy)
	return err
}

func (r *nvrCrr) ScheduleModify(schedule quarkv1alpha1.NetAppVolumeReplicationScheduleType) error {
	if r.nvr.Spec.ReplicationSchedule == schedule {
		return nil
	}
	volumeReplicationCopy := r.nvr.DeepCopy()
	volumeReplicationCopy.Spec.ReplicationSchedule = schedule
	_, err := r.nvcClient.NetAppVolumeReplicationUpdate(volumeReplicationCopy)
	return err
}
