package v2

import (
	"context"
	"fmt"
	"strings"

	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	quarkv1alpha1 "quark.netapp.io/volume-controller/pkg/apis/volumecontroller/v1alpha1"
	"quark.netapp.io/volume-controller/pkg/quarkconfig"
)

// Interface to abstract the functionalities supported for the three NVR usecases: CRR, Backup and Restore.
type nvrInterface interface {
	Create() (*quarkv1alpha1.NetAppVolumeReplication, error)
	Update() (*quarkv1alpha1.NetAppVolumeReplication, error)
	Delete() error
	Get() (*quarkv1alpha1.NetAppVolumeReplication, error)
	Initialize(transferSnapshot string) error
	Resync() error
	Break() error
	TransferSnapshot(snapshotName string) error
	TransferAbort() error
	ScheduleModify(schedule quarkv1alpha1.NetAppVolumeReplicationScheduleType) error
}

type nvrCommonImpl struct {
	nvr       *quarkv1alpha1.NetAppVolumeReplication
	nvcClient *NVCClient
}

// Factory function that creates a concrete instance for nvr interface based on the type field in volumeReplication CR.
func NewNVRInstance(volumeReplication *quarkv1alpha1.NetAppVolumeReplication, client *NVCClient) (nvrInterface, error) {
	if volumeReplication == nil {
		return nil, fmt.Errorf("invalid input, volumeReplication cannot be nil")
	}

	return nvrInstanceBasedOnType(volumeReplication, client)

}

// Factory function that creates a concrete instance for nvr interface based on the type of the volumeReplication UUID.
func NewNVRInstanceFromUUID(volumeReplicationUUID string, client *NVCClient) (nvrInterface, error) {
	volumeReplication, err := client.quarkClient.QuarkV1alpha1().NetAppVolumeReplications(client.nvcNamespace).Get(
		context.TODO(), volumeReplicationUUID, metav1.GetOptions{})
	if err != nil {
		return nil, err
	}

	return nvrInstanceBasedOnType(volumeReplication, client)
}

func nvrInstanceBasedOnType(volumeReplication *quarkv1alpha1.NetAppVolumeReplication, client *NVCClient) (nvrInterface, error) {
	switch volumeReplication.Spec.ReplicationType {
	case quarkv1alpha1.ReplicationTypeBackup:
		return &nvrBackup{
			nvrCommonImpl: nvrCommonImpl{
				nvr:       volumeReplication,
				nvcClient: client,
			},
		}, nil
	case quarkv1alpha1.ReplicationTypeRestore:
		return &nvrRestore{
			nvrCommonImpl: nvrCommonImpl{
				nvr:       volumeReplication,
				nvcClient: client,
			},
		}, nil
	default:
		// default case would be crr, existing nvr cr's created before upgrade would not have ReplicationType set
		return &nvrCrr{
			nvrCommonImpl: nvrCommonImpl{
				nvr:       volumeReplication,
				nvcClient: client,
			},
		}, nil
	}
}

func (r *nvrCommonImpl) Get() (*quarkv1alpha1.NetAppVolumeReplication, error) {
	return r.nvr, nil
}

func (r *nvrCommonImpl) Update() (*quarkv1alpha1.NetAppVolumeReplication, error) {
	newVolumeRepl, err := r.nvcClient.quarkClient.QuarkV1alpha1().NetAppVolumeReplications(r.nvcClient.nvcNamespace).Update(
		context.TODO(), r.nvr, metav1.UpdateOptions{})
	if err != nil {
		if strings.Contains(err.Error(), quarkconfig.OldVersionErr) {
			return nil, StaleVersionError{}
		}
		return nil, err
	}
	return newVolumeRepl, nil
}
