// Package quarkconfig contains constants are used across more than one quark package
// ** Please do not add to this file if your variables are scoped to a single package ***
package quarkconfig

import (
	"fmt"
	"time"

	corev1 "k8s.io/api/core/v1"
	rbachelpers "k8s.io/api/rbac/v1"
	"k8s.io/apimachinery/pkg/api/resource"

	quarkv1alpha1 "quark.netapp.io/volume-controller/pkg/apis/volumecontroller/v1alpha1"
)

// ServiceLevel is the storage service level
type ServiceLevel string

// Quark Container Name & Image Info
type ContainerImage struct {
	Name  string
	Image string
}

// Quark Pod and Container Info
type PodInfo struct {
	Namespace        string
	Replica          int
	PodLabelSelector string
	Container        map[string]ContainerImage
}

type AzureSecretInfo struct {
	ClientID       string
	TenantID       string
	ClientSecret   string
	SubscriptionID string
}

const (
	// Well-known Object Names
	AzureSecretName          = "azure-conf"
	NvcSecretName            = "nvc-sa-key"
	SvcMeshCtrlSecretName    = "asmc-sa-key"
	SupportabilitySecretName = "supportability-sa-key"
	RootCertName             = "root-certs"
	EtcdClientSecretName     = "etcd-client-tls-cert"
	NhcSecretName            = "nhc-sa-key"
	NHCPDBName               = "nhc-pdb"
	ETCDPDBName              = "etcd-pdb"

	// Apps
	QuarkAppName = "quark"

	// Deployments
	QuarkOperatorDeploymentName         = "quark-operator"
	NVCDeploymentName                   = "netapp-volume-controller"
	SvcMeshControllerDeploymentName     = "netapp-svcmesh-controller"
	HASpareDeploymentName               = "ha-spare"
	MigrationSpare                      = "migration-spare"
	NHCStatefulSetName                  = "netapp-ha-controller"
	NodeMonitorDaemonSetName            = "node-monitor"
	IPCDeploymentName                   = "image-prepull-controller"
	CSCDeploymentName                   = "cloud-snapshot-controller"
	NSCDaemonSetName                    = "ntapvol-daemonset"
	QuarkOperatorName                   = "quark-operator"
	ImagePrepullDaemonsetServiceAccount = "image-prepull-daemonset"
	MigrationSpareServiceAccount        = "netapp-migration-spare-account"
	LowPriorityClass                    = "low-priority"
	NegativePriorityClass               = "negative-priority"
	KubeStateMetricsDeploymentName      = "quark-kube-state-metrics"
	QuarkPrometheusServerDeploymentName = "quark-prometheus-server"
	QuarkPrometheusServerPVCName        = "quark-prometheus-server-fa"
	QuarkOtelDeploymentName             = "quark-otel-collector"
	QuarkOtelPVCName                    = "quark-otel-collector-fa"
	QuarkOtelServiceAccount             = "netapp-otel-account"
	EtcdOperatorDeploymentName          = "etcd-operator"

	// We're currently just using the systemwide default (which equates to 0),
	// but leaving this constant here makes it easier to change later.
	QuarkPodPriorityClass      = "default-priority"
	DefaultPrepullDaemonSet    = "image-prepull"
	FluentBitDaemonsetName     = "fluent-bit"
	SvcMeshCtrlClusterRoleName = "netapp-asmc-crole"

	// Well-Known Namespaces
	// This namespace hosts all the core quark compoenents
	QuarkSystemNamespace = "quark-system"
	// This namespace hosts all the core kube compoenents
	KubeSystemNamespace = "kube-system"
	// This namespace is where the IPC looks for ConfigMaps
	// with image data to create prepull DaemonSet resources.
	// TODO: This should potentially go away once quark operator can
	// create prepull DaemonSets.
	ImagePrepullNamespace = "image-prepull"

	// Nodevol
	QuarkNodevolSetName   = "quark-nodevol"
	NodevolControllerName = "nodevol-controller"
	NodevolPVCSize        = 64 << 30 //64GB
	NodevolNamespace      = QuarkSystemNamespace
	NodevolHostpath       = "/mnt/disks/nodevol"
	NodevolPath           = "/mroot"

	// Various Variants for Quark CR
	ProductionVariant = "production"
	InternalVariant   = "internal"
	AlphaVariant      = "alpha"
	OnPremVariant     = "onprem"
	MinVariant        = "min"

	// NFS Volume default permission
	DefaultUnixPermission        = "0777"
	StandardUnixPermissionLength = 4

	CIFSProtocol        = "cifs"
	SMBProtocol         = "smb"
	NFS3Protocol        = "nfs3"
	NFS4Protocol        = "nfs4"
	NFS40Protocol       = "nfs4.0"
	NFS41Protocol       = "nfs4.1"
	NFSV4Protocol       = "nfsv4"
	NFSV40Protocol      = "nfsv4.0"
	NFSV41Protocol      = "nfsv4.1"
	DefaultNFS3Protocol = NFS3Protocol
	DefaultNFS4Protocol = NFS41Protocol

	// Default AnonymousUser in export policy rule
	DefaultAnonymousUser = "65534"
	SecurityTypeKerberos = "krb"

	// Annotations
	MetadataPrefix                      = "quark.netapp.io"
	SNATAnnotation                      = MetadataPrefix + "/ntapvol-snat" // deprecated 2023.12
	SNATPortRangeAnnotation             = MetadataPrefix + "/ntapvol-snat-portrange"
	AppLabel                            = MetadataPrefix + "/app"
	VolumeLabel                         = MetadataPrefix + "/volume"
	TypeLabel                           = MetadataPrefix + "/type"
	NodeLabel                           = MetadataPrefix + "/quark-node"
	OwnerIDLabel                        = MetadataPrefix + "/owner"
	StoragePoolLabel                    = MetadataPrefix + "/storagePool"
	RemoteStoragePoolLabel              = MetadataPrefix + "/remoteStoragePool"
	RemoteLocationLabel                 = MetadataPrefix + "/remoteLocation"
	RemoteClusterNameLabel              = MetadataPrefix + "/remoteClusterName"
	ClusterPeeringLabel                 = MetadataPrefix + "/clusterPeering"
	VolumeCloneParentUUUIDLabel         = MetadataPrefix + "/cloneParentVolumeUUID"
	CMEKConfigMapLabel                  = MetadataPrefix + "/cmekConfig"
	StoragePoolEncryptionLabel          = MetadataPrefix + "/storagePoolEncryption"
	VolumeEncryptionLabel               = MetadataPrefix + "/volumeEncryption"
	ForwardingRuleLabelStatusAnnotation = MetadataPrefix + "/forwarding-rule-label-status"

	// Max Volumes encrypted in parallel
	MaxVolumesEncryptionInParallel = 1

	// CMEK Migration Workflow Polling Interval
	EncryptionCheckIntervalInMinutes = 1
	ObjectStoreLabel                 = MetadataPrefix + "/objectStore"
	VolumeReplicationLabel           = MetadataPrefix + "/volumeReplication"
	EncryptionCheckIntervalInSeconds = 60
	EncryptionRetryIntervalInSeconds = 60
	EncryptionRetryCount             = 5

	// SDE volume and SP labels
	SDELabelPrefix = "sde.netapp.io"
	SDEVolumeUUID  = SDELabelPrefix + "/volUUID"
	SDESpUUID      = SDELabelPrefix + "/spUUID"

	// Finalizers
	AddressFinalizer                    = "address.quark.netapp.io"
	ADClientAddressFinalizer            = "ad-client-address.quark.netapp.io"
	CloudSnapshotFinalizer              = "cloudsnapshot.quark.netapp.io"
	CloudSnapshotRestoreFinalizerSuffix = ".restore.quark.netapp.io"
	QuarkFinalizer                      = "quarks.quark.netapp.io"
	VolumeFinalizer                     = "volume.quark.netapp.io"
	ClusterPeeringFinalizer             = "peering.quark.netapp.io"
	ServiceMeshFinalizer                = "servicemesh.quark.netapp.io"
	VolumeReplicationFinalizer          = "replication.quark.netapp.io"
	EtcdKeyFinalizer                    = "etcdkey.quark.netapp.io"
	ObjectStoreFinalizer                = "objectstore.quark.netapp.io"

	//Storage pool subsystems
	BootSubsystem        = "BOOT"
	NfsSubsystem         = "NFS"
	NfsKerberosSubSystem = "NFS_KERBEROS"
	NfsLdapSubsystem     = "NFS_LDAP"
	SmbSubsystem         = "CIFS"
	ReplicationSubsystem = "C2C_REPLICATION"

	//Storage pool subsystem status
	SubsystemReady    = "ready"
	SubsystemNotReady = "notready"

	// CMEK constants
	CMEKDefaultTimeout           = "25" // secs (TODO-cmek double check with dmap team)
	CMEKConfigMapPrefix          = "cmek-"
	CMEKProtocolTag              = "cmek_protocol"
	CMEKProjectIDTag             = "project_id"
	CMEKKeyRingTag               = "key_ring"
	CMEKKeyRingLocationTag       = "key_ring_location"
	CMEKKeyNameTag               = "key_name"
	CMEKTimeoutTag               = "cmek_timeout"
	CMEKRootCACertMountPathTag   = "root_ca_cert_path"
	CMEKEmulatorOauthTag         = "cmek_oauth"
	CMEKEmulatorVerifyIpHostTag  = "cmek_verify_ip_host"
	CMEKLogLevelTag              = "cmek_log_level"
	CMEKMigration                = "cmek_migration"
	CMEKKeyPathPrefixETCD        = "/v" + QuarkEtcdKeyVersion + "/StoragePool/"
	CMEKKeyPathSuffixETCD        = "/key_mgmt_service/"
	CMEKStateETCD                = "CMEKState"
	CMEKFirstBootKey             = "cmek_first_boot_key"
	CMEKConfiguredEnvVariableKey = "QUARK_CMEK_CONFIGURED"

	RSIDLabel           = MetadataPrefix + "/RSID"
	MaxSnapshotListTask = 5
	// Upgrade Specific Constants
	// Volume configMap Key which has the new QuarkPod deployment definition
	QuarkPodDefinition = "quark-pod-definition"
	QuarkVersion       = "quark-version"
	// Upgrade Monitor file to exchange info between Podrick & DMAP
	// Caution: DO NOT Change this Arbitrarily as It will break Upgrade
	QuarkUpgradeMonitorFile = "/mroot/upgrade"
	// Upgrade ETCD key to exchange info between Podrick & DMAP
	// Caution: DO NOT Change this Arbitrarily as It will break Upgrade
	QuarkUpgradeEtcdKey = "status"
	// ClusterUpgradeTimeout is used by ha tool to set default timeout for master and node pools upgrade scenarios
	ClusterUpgradeTimeout = 45
	// Quark ETCD cluster's fully qualified domain name endpoint that all clients will talk to
	QuarkEtcdClusterEndpoint = "quark-etcd-cluster-client.quark-system.svc.cluster.local:2379"
	// Retry etcd call count
	EtcdRetries = 5
	// ETCD timeout error
	EtcdTimedout = "unableToReachEtcd"
	// ETCD client dial timeout
	EtcdDailTimeout = 5 * time.Second
	// DMAP flag to enable writing bootargs to etcd
	QuarkUseEtcdBootargs = "true"
	// DMAP env variable value for etcd key format version
	QuarkEtcdKeyVersion = "1"
	// DMAP env variable value for disk count per RG for Azure live-resize enablement
	AzureDiskCountPerRG = "5"

	// Azure ILB Annotations
	AzureILBTypeAnnotation    = "service.beta.kubernetes.io/azure-load-balancer-internal"
	AzureILBSubnetAnnotation  = "service.beta.kubernetes.io/azure-load-balancer-internal-subnet"
	AzureILBHAPortsAnnotation = "service.beta.kubernetes.io/azure-load-balancer-enable-high-availability-ports"

	// GCP ILB Annotations
	LegacyILBTypeAnnotation        = "cloud.google.com/load-balancer-type"
	ILBTypeAnnotation              = "networking.gke.io/load-balancer-type"
	ILBPUPIAnnotation              = "networking.gke.io/internal-load-balancer-subnet"
	ILBGlobalAccessAnnotation      = "networking.gke.io/internal-load-balancer-allow-global-access"
	GCPILBBackendService           = "service.kubernetes.io/backend-service"
	ILBTCPForwardingRuleAnnotation = "service.kubernetes.io/tcp-forwarding-rule"
	ILBUDPForwardingRuleAnnotation = "service.kubernetes.io/udp-forwarding-rule"

	ILBInternal = "Internal"

	SNATActiveOpen        = "activeopen"
	SNATActiveOpenUDP     = "activeopenu"
	SNATActiveOpenTCP1    = "secdactopen1"
	SNATActiveOpenTCP2    = "secdactopen2"
	SNATActiveOpenTCP3    = "secdactopen3"
	SNATActiveOpenTCP4    = "secdactopen4"
	SNATActiveOpenTCPNSU1 = "secdactopen5"
	SNATActiveOpenTCPNSU2 = "secdactopen6"
	SNATActiveOpenTCPNSU3 = "secdactopen7"
	SNATActiveOpenTCPNSU4 = "secdactopen8"
	SNATActiveOpenUDP1    = "secdactopen1u"
	SNATActiveOpenUDP2    = "secdactopen2u"
	SNATActiveOpenUDP3    = "secdactopen3u"
	SNATActiveOpenUDP4    = "secdactopen4u"
	SNATActiveOpenUDPNSU1 = "secdactopen5u"
	SNATActiveOpenUDPNSU2 = "secdactopen6u"
	SNATActiveOpenUDPNSU3 = "secdactopen7u"
	SNATActiveOpenUDPNSU4 = "secdactopen8u"
	SNATActiveOpenUDPDIG1 = "secdactopen9u"
	SNATActiveOpenUDPDIG2 = "secdactopen10u"

	// TODO: Combine the nfs and CIFS port ranges into one variable after legacy volumes are updated
	SNATActiveOpenNFSTCPPortRange  = "tcp:61000"
	SNATActiveOpenCIFSTCPPortRange = "tcp:61001:61500"
	SNATActiveOpenCIFSUDPPortRange = "udp:61001:61500"
	SNATActiveOpenNFSUDPPortRange  = "udp:61000"

	//ConnectionTrackingPolicyNeverPersist : GCP ILB connection tracking policy value
	ConnectionTrackingPolicyNeverPersist = "NEVER_PERSIST"

	// Cloud Types supported
	CloudGCP     = "gcp"
	CloudAzure   = "azure"
	CloudCTL     = "ctl"
	CloudNone    = "none"
	CloudUnknown = "unknown"

	ServiceLevelStandard         ServiceLevel = "standard"
	ServiceLevelRegionalStandard ServiceLevel = "regional-standard"
	DefaultScheduler             string       = "default-scheduler"
	MaxVol0Size                               = 20 << 30 // 20GiB
	MaxPVCSize                                = "70368744177664"
	MinPVCSize                                = "1073741824"
	NvramSize                                 = "322122547200"
	QuarkVolumeSize                           = "68719476736"
	DefaultPrometheusSCAzure                  = "md-standard"
	DefaultPrometheusSCGCP                    = "standard"
	MaxPVCSizeAzure                           = "35183298347008"
	MinPVCSizeAzure                           = "4294967296"
	MinDataPVCsGCP                            = "1"
	MinDataPVCsAzure                          = "5"
	AvailableZonesGCP                         = 2
	AvailableZonesGCP1P                       = 1
	AvailableZonesAzure                       = 3
	//configuration names to look for in the env
	SupportabilityAssociationConfig = "supportability-config"
	QuarkExporterConfig             = "exporter-config"
	FluentbitSidecarConfig          = "fluentbit-sidecar-config"

	// CertsPath certificates mount path
	CertsPath                   = "/etc/ssl/quark-certs/"
	OSDefaultTrustedCACertsPath = "/etc/ssl/certs/"
	QuarkTrustedCACertsPath     = "/etc/ssl/quark-ca-certs/"
	QuarkTrustedCAPodVolumeName = "quark-trusted-ca-cert-vol"

	// TLS certificate files
	TLSCertFileName           = "tls.crt"
	TLSCertPrivateKeyFileName = "tls.key"
	CACertificateFileName     = "ca-certificates.crt"

	// Taint used by NHC
	NodeUnreachable               = MetadataPrefix + "/unreachable"
	ZoneFailure                   = MetadataPrefix + "/zone-failure"
	NHCHeartbeatPort              = 2443
	NHCHeadlessService            = "netapp-ha-controller-svc"
	NHCServiceDomain              = ".svc.cluster.local"
	HeartbeatCertificatePath      = "/var/secrets/heartbeat-tls/"
	HeartbeatCertificate          = HeartbeatCertificatePath + TLSCertFileName
	HeartbeatPrivateKey           = HeartbeatCertificatePath + TLSCertPrivateKeyFileName
	HeartbeatCACertificate        = QuarkTrustedCACertsPath + CACertificateFileName
	HBClientCommonName            = "HA HB Client"
	CertificateExpiryNotifyPeriod = 4320           // approximately 6 months
	CertificateVerifyTickerTime   = 24 * time.Hour // 24 hours or one day

	RootCACertMountPath = "/usr/local/share/root-certs"

	// Regional HA configuration constants
	RegionalHAConfigMapName = "regional-ha-config"
	// Indicates that a zone is healthy
	RegionalHAActiveZone = "active"
	// Used to indicate that a failure has been detected, but failover is
	// not yet complete
	RegionalHAFailingZone = "failing"
	// Used to indicate that a zone has fully failed
	RegionalHAFailedZone         = "failed"
	RecoveryDeploymentName       = "zone-recovery-test"
	ZoneFailureThresholdFraction = 0.4
	ZoneMigrationCauseUser       = "User"
	ZoneMigrationCauseFailure    = "ZoneFailure"
	StorageClassBaseName         = MetadataPrefix + "/base-name"
	ZoneTopologyKey              = "topology.kubernetes.io/zone"
	OldZoneTopologyKey           = "failure-domain.beta.kubernetes.io/zone"
	CSIZoneTopologyKey           = "topology.gke.io/zone"
	DiskCSIAzureTopologyKey      = "topology.disk.csi.azure.com/zone"
	InstanceTypeKey              = "node.kubernetes.io/instance-type"
	// this is deprecated but used to avoid problems during upgrades
	OldInstanceTypeKey          = "beta.kubernetes.io/instance-type"
	CAOkTotalUnreadyCount       = 3
	CAMaxTotalUnreadyPercentage = .45
	GkeNodePoolKey              = "cloud.google.com/gke-nodepool"
	AzureNodePoolKey            = "agentpool"
	// MaxNodeStartupTime is the maximum time from the moment the node is registered to the time the node is ready.
	MaxNodeStartupTime = 15 * time.Minute
	// MaxStatusSettingDelayAfterCreation is the maximum time for node to set its initial status after the
	// node is registered.
	MaxStatusSettingDelayAfterCreation = 2 * time.Minute

	// PrometheusPortAnnotation and PrometheusScrapeAnnotation to
	// scrape quark pod by Prometheus server
	PrometheusPortAnnotation   = "prometheus.io/port"
	PrometheusScrapeAnnotation = "prometheus.io/scrape"

	// SMB DNS ConfigMap name
	SMBDNSConfigName = "secd-dns-config"

	// Message Constants from NVC
	// SuccessSynced is used as part of the Event 'reason' when a NetAppVolume is synced
	SuccessSynced = "Synced"
	// ErrResourceExists is used as part of the Event 'reason' when a NetAppVolume fails
	// to sync due to a Deployment of the same name already existing.
	ErrResourceExists = "ErrResourceExists"
	// NetAppVolumeOnlined is used as part of the Event 'reason' when a NetAppVolume is online
	// (Only used in Podrick)
	NetAppVolumeOnlined = "NetAppVolumeOnline"
	// NetAppVolumeOfflined is used as part of the Event 'reason' when a NetAppVolume is offline
	NetAppVolumeOfflined = "NetAppVolumeOffline"
	// NetAppVolumeReady is used as part of the NetAppVolumeOnline's condition 'reason'
	// when a NetAppVolume transitions to Online. (Only used in Podrick)
	NetAppVolumeReady = "VolumeReady"
	// NetAppVolumeNotReady is used as part of the NetAppVolumeOnline's condition 'reason'
	// when a NetAppVolume transitions to Offline.
	NetAppVolumeNotReady = "VolumeNotReady"
	// NetAppVolumeUpgrade is used as part of NetAppVolume's upgrade workflow
	NetAppVolumeUpgrade = "NetAppVolumeUpgrade"
	// NetAppStoragePoolUpgrade is used as part of NetAppStoragePool's upgrade workflow
	NetAppStoragePoolUpgrade = "NetAppStoragePoolUpgrade"
	// NetAppStoragePoolOnlined is used as part of the Event 'reason' when a NetAppStoragePool is online
	// (Only used in Podrick)
	NetAppStoragePoolOnlined = "NetAppStoragePoolOnline"
	// NetAppStoragePool Granular health status is changed.
	NetAppStoragePoolSubsystemStatusChanged = "NetAppStoragePoolSubsystemStatusChange"
	// NetAppStoragePoolOfflined is used as part of the Event 'reason' when a NetAppStoragePool is offline
	NetAppStoragePoolOfflined = "NetAppStoragePoolOffline"
	// NetAppStoragePoolReady is used as part of the NetAppStoragePoolOnline's condition 'reason'
	// when a NetAppStoragePool transitions to Online. (Only used in Podrick)
	NetAppStoragePoolReady = "StoragePoolReady"
	// NetAppStoragePoolNotReady is used as part of the NetAppStoragePoolOnline's condition 'reason'
	// when a NetAppStoragePool transitions to Offline.
	NetAppStoragePoolNotReady = "StoragePoolNotReady"
	// NetAppStoragePoolDisable is used as part of the NetAppStoragePoolDisable's condition 'reason'
	// when a NetAppStoragePool transitions to disable.
	NetAppStoragePoolDisable = "StoragePoolDisable"
	// StoragePoolEnable is used as part of the StoragePoolEnable's condition 'reason'
	// when a NetAppStoragePool transitions to enable.
	NetAppStoragePoolEnable = "StoragePoolEnable"

	// MessageDiskAdd is the message used for events and conditions when a disk is being added.
	MessageDiskAdd = "Pod is getting restarted due to disk add."
	// MessageDNSUpdate is the message used for events and conditions when a DNS is being updated.
	MessageDNSUpdate = "Pod is getting restarted due to DNS update."
	// MessageQuarkDeploymentCreated is the message used for events and conditions when the Quark
	// Deployment has been created and the volume is coming online for the first time.
	MessageQuarkDeploymentCreated = "Quark deployment created. Volume is coming online."
	// MessagePodShutdownUnknown is the message used for events and conditions when a Quark pod
	// has been/is being shutdown. The reason for this shutdown is unknown.
	MessagePodShutdownUnknown = "Pod is getting shutdown for reasons unknown."
	// MessageResourceExists is the message used for Events when a resource
	// fails to sync due to a Deployment already existing
	MessageResourceExists = "Resource %q already exists and is not managed by NetAppVolume"
	// MessageResourceSynced is the message used for an Event fired when a NetAppVolume
	// is synced successfully
	MessageResourceSynced = "NetAppVolume synced successfully"

	MessageResourceSyncedSP = "NetAppStoragePool synced successfully"

	MessageResourceGlobalAccessEnabled  = "NetAppStoragePool global access enabled"
	MessageResourceGlobalAccessDisabled = "NetAppStoragePool global access disabled"

	MessageResourceSyncedExportPolicy   = "NetAppVolumeExportPolicy synced successfully"
	MessageResourceModifySnapshotPolicy = "NetAppVolumeSnapshotPolicy modified successfully"

	// Podrick Constants
	MessageVolumeStateChanged      = "Volume state changed"
	MessageVolumeOnline            = "Volume is online."
	MessageVolumeOffline           = "Volume is offline."
	MessageStoragePoolStateChanged = "StoragePool state changed"
	MessageSnapshotSynced          = "NetAppVolumeSnapshot synced successfully"
	MessageSnapshotsListSynced     = "NetAppVolumeSnapshotsList synced successfully"
	//Upgrade Messages
	MessageUpgradeInProgress = "NetAppVolume upgrade is in progress"
	MessageUpgradeFailed     = "NetAppVolume upgrade failed"

	// AD related error messages
	MessageADServerError        = "AD Server Error"
	MessageADServerSiteNotFound = "Site not found"

	// KDC related error messages
	MessageKDCUnreachable = "KDC Unreachable"

	// MessagePodShutdownUnknown is the message used for events and conditions when a Quark pod
	// has been/is being shutdown. The reason for this shutdown is unknown.
	MessagePodShutdownStoragePoolDisabled = "Pod is getting shutdown due to disabled storage pool."

	// SECD/DMAP Error codes for SMB operations in REST API Payload
	// Currently we are relying on the REST API command output to capture this Error code details.
	// We don't have direct error mapping in dmap code.
	// The error code are defined in https://opengrok-prd.eng.netapp.com/source/xref/dev/apps/lib/liblogerror/result_errors.h
	// Then it is converted to specific UINT code, like the error code mentioned below. This conversion is spread across multiple places.
	//    https://opengrok-prd.eng.netapp.com/source/xref/dev/onnix_mgmt/dmap_mgmt/src/dmap/cifs_support/cifs_utils.cc : ResultErrToHttpErr()
	//    https://opengrok-prd.eng.netapp.com/source/xref/dev/apps/lib/libpistacherest/src/errMap.cc: build_http_error_response()
	//    https://opengrok-prd.eng.netapp.com/source/xref/dev/message_catalogs/src/msg_catalog.h:  MSGCODE_TO_UINT()
	DMAPErrSiteNotFound       = 1841992 // RESULT_ERROR_SECD_DEFAULT_SITE_NOT_FOUND
	DMAPErrDnsIpNotReachable  = 1841811 // RESULT_ERROR_DNS_NAME_LOOKUP
	DMAPErrDomainNotReachable = 655366  // RESULT_ERROR_SECD_NO_SERVER_AVAILABLE
	DMAPErrUnknownUser        = 656442  // RESULT_ERROR_KERBEROS_C_PRINCIPAL_UNKNOWN
	DMAPErrInvalidPassword    = 656441  // RESULT_ERROR_KERBEROS_BAD_INTEGRITY, RESULT_ERROR_KERBEROS_PREAUTH_FAILED

	SMBFinalizer         = "smb.quark.netapp.io"
	SMBSecretDeleteLabel = "smbDelete"

	CcpdStoragePoolFinalizer = "ccpd.quark.netapp.io"

	// Cluster Peering Reasons
	EnsuringServiceMeshControlPlane = "EnsuringServiceMeshControlPlane"
	ServiceMeshDataPlaneActive      = "ServiceMeshDataPlaneActive"
	ErrorEnablingServieMesh         = "ErrorEnablingServieMesh"
	ErrorInjectingProxy             = "ErrorInjectingProxy"

	// Maximum snapshots reached error code
	ENOMORESNAPSCODE = 135

	// Stale version error message
	OldVersionErr = "the object has been modified; please apply your changes to the latest version and try again"

	// Size constants
	// MiB is 2^20
	MiB = 1 << 20
	// GiB is 2^30
	GiB = 1 << 30
	// TiB is 2^40
	TiB = 1 << 40

	// Min size to allow for QoS throttling
	QoSMinSize = 100 * GiB

	//Size of EtcD volume
	EtcDVolSize = 4 * GiB
	// Maximum number of operations permitted in a transaction
	EtcdMaxTxnOps = "1024"
	// ETCDCTL_API version
	EtcdCtlAPI = "3"
	// Quark ETCD cluster name
	QuarkEtcdClusterName = "quark-etcd-cluster"
	// NVC metric exporter port
	NvcMetricsPort = "9805"
	// Podrick server port
	PodrickServerPort = 8083
	//Podrick metrics Port
	PodrickMetricsPort = "9808"
	//HA metrics port
	HAMetricsPort = "9807"

	// may change in future as compaction need changes
	//EtcdAutoCompactionTime Auto compaction period for etcd
	EtcdAutoCompactionTime = "10h"

	//This environment variable will be added to etcd pods cr when etcd cost reduction
	//feature will be enabled during quark upgrade. It will make sure to restart all
	//the etcd pods to remove nodeselector.
	//In case of fresh Install it will be added to etcd pods cr when the etcd cost
	//reduction feature is enabled.
	EtcdNodeSelectorRemoval = "true"

	//AGGR_TP_FACTOR for thin-provisioning
	AggrTpFactor = 1

	//Ceil round operation
	CeilRoundFunction = "ceil"
	//Floor round operation
	FloorRoundFunction = "floor"

	//Server Error
	ErrorBadRequest       = 400
	ErrorNotFound         = 404
	ErrorMethodNotAllowed = 405
	ErrorConflict         = 409
	ErrorInternalError    = 500

	// Service Account Names
	NetAppPodrickSA    = "netapp-podrick-controller-account"
	NetAppPodrickSAOld = "podrick-account"
	NetAppQuarkSA      = "netapp-quark-account"
	NetAppNodevolSA    = "netapp-nodevol-controller-account"

	// Roles
	NetAppPodrickRole    = "netapp-podrick-controller-role"
	NetAppPodrickRoleOld = "podrick-role"
	NetAppQuarkRole      = "netapp-quark-role"

	// Role Bindings
	NetAppPodrickRoleBinding    = "netapp-podrick-controller-role-binding"
	NetAppPodrickRoleBindingOld = "podrick-binding"
	NetAppQuarkRoleBinding      = "netapp-quark-role-binding"

	// Cluster Role
	NetAppPodrickClusterRole    = "netapp-podrick-controller-cluster-role"
	NetAppPodrickClusterRoleOld = "podrick-cluster-role"
	NetAppQuarkClusterRole      = "netapp-quark-cluster-role"

	// Cluster Role Bindings
	NetAppPodrickClusterRoleBinding    = "netapp-podrick-controller-cluster-role-binding"
	NetAppPodrickClusterRoleBindingOld = "podrick-cluster-binding"
	NetAppQuarkClusterRoleBinding      = "netapp-quark-cluster-role-binding"

	// K8s Secret names
	SecretNameQuarkTrustedCA  = "quark-trusted-ca"
	SecretNameEtcdServerTls   = "etcd-server-tls-cert"
	SecretNameEtcdClientTls   = "etcd-client-tls-cert"
	SecretNameEtcdPeerTls     = "etcd-peer-tls-cert"
	SecretNameEtcdOperatorTls = "etcd-operator-tls-cert"
	SecretNameHaHbClientTls   = "ha-heartbeat-client-tls-cert"
	SecretNameHaHbServerTls   = "ha-heartbeat-server-tls-cert"

	BackupPrepullDaemonSet = "image-prepull-backup"
	QuarkDigest            = "sha256"

	// Service mesh shared constants
	IstioSystemNamespace = "istio-system"
	ASMSystemNamespace   = "asm-system"
	RevisionLabel        = "istio.io/rev"

	// Otel Config Generation Label.
	OtelCfgLabel = "quark.netapp.io/otel-cfg-generation"

	MaxRetries = 5

	EncryptionVolumeIsEncrypted    = 7210010
	EncryptionAlreadyInProgress    = 7209893
	EncryptionNotInProgress        = 7209894
	EncryptionAlreadyInPausedState = 7209902
	EncryptionNotInPausedState     = 7209914
	VolumeNotFound                 = 13040
	// The below ones are common case for any encryption actions
	EncryptionVolumeRehost      = 7209906
	EncryptionVolumeInMove      = 7209984
	EncryptionVolumeCloneSplit  = 7209324
	EncryptionConversionPending = 7209950

	// Fatal errors
	EncryptedAggregatesNotSupported    = 7209748
	EncryptionNotAllowedReadOnlyVolume = 7209282
	IllegalSnapLockOption              = 7209089

	// Volume encryption status from DMAP
	EncryptionPausedFromDMAP          = "Paused"
	EncryptionStateConvertingFromDMAP = "converting_to_encrypted"
)

var TLSCertificates = []string{
	TLSCertFileName,
	TLSCertPrivateKeyFileName,
}

// Below are netapp-volume/quarkpod ns -> secret-names
var NVCNamespaceSecrets = []string{
	SecretNameEtcdClientTls,
	SecretNameQuarkTrustedCA,
	SecretNameHaHbServerTls,
}

// Below are quark-system ns -> secret-names
var QuarkSystemNamespaceSecrets = []string{
	SecretNameEtcdServerTls,
	SecretNameEtcdOperatorTls,
	SecretNameEtcdPeerTls,
	SecretNameHaHbClientTls,
	SecretNameQuarkTrustedCA,
}

var PurgeOldSecretsNSDefault = []string{
	"ha-heartbeat-server-tls",
	"etcd-client-tls",
}

var PurgeOldSecretsNSQuarkSystem = []string{
	"ha-heartbeat-client-tls",
	"etcd-operator-tls",
	"etcd-peer-tls",
	"etcd-server-tls",
}

type subnetIPExhaustionError struct {
	baseError error
}

func (e subnetIPExhaustionError) Error() string {
	return fmt.Sprintf("Addresses exhausted for all the given Subnets: %v",
		e.baseError)
}

func NewSubnetsIPExhaustionError(err error) subnetIPExhaustionError {
	return subnetIPExhaustionError{baseError: err}
}

func IsSubnetsIPExhaustionError(err error) bool {
	_, ok := err.(subnetIPExhaustionError)
	return ok
}

// These rules are required by the service-mesh controller (currently Anthos).
// Sourced from https://confluence.ngage.netapp.com/display/CVO/Anthos+Service+Mesh+Spec#AnthosServiceMeshSpec-RBAC
var ASMCRuleSet = []rbachelpers.PolicyRule{
	{
		APIGroups: []string{""},
		Resources: []string{"events"},
		Verbs:     []string{"*"},
	},
	{
		APIGroups: []string{""},
		Resources: []string{"services"},
		Verbs:     []string{"get", "create", "delete"},
	},
	{
		APIGroups: []string{"quark.netapp.io"},
		Resources: []string{"netappclusterpeerings", "netappclusterpeerings/status"},
		Verbs:     []string{"*"},
	},
	{
		APIGroups: []string{"quark.netapp.io"},
		Resources: []string{"netappstoragepools"},
		Verbs:     []string{"get"},
	},
	{
		APIGroups: []string{"apps"},
		Resources: []string{"deployments"},
		Verbs:     []string{"get", "list", "watch", "update", "patch"},
	},
	{
		APIGroups: []string{"mesh.cloud.google.com"},
		Resources: []string{"controlplanerevisions"},
		Verbs:     []string{"get", "list"},
	},
	{
		APIGroups: []string{"security.istio.io"},
		Resources: []string{"peerauthentications"},
		Verbs:     []string{"create"},
	},
}

// Generate the name of the Cluster Peering service for a Storage Pool
func GetClusterPeeringServiceName(storagePoolUUID string) string {
	return fmt.Sprintf("cluster-peering-%s", storagePoolUUID)
}

// CRRports defines the ports used for CRR Services
var CRRPorts = []corev1.ServicePort{
	{
		Name:     "dsmf",
		Protocol: "TCP",
		Port:     11104,
	},
	{
		Name:     "csm",
		Protocol: "TCP",
		Port:     11105,
	},
	{
		Name:     "csm2",
		Protocol: "TCP",
		Port:     7700,
	},
}

type QuarkContainer = string

type AzureClientENVNotSetError struct {
	Message string
}

func (e AzureClientENVNotSetError) Error() string {
	return e.Message
}

type StoragePoolShrinkExceedTotalVolCapacityError struct {
	RequestedSize  resource.Quantity
	MinAllowedSize resource.Quantity
}

func (e StoragePoolShrinkExceedTotalVolCapacityError) Error() string {
	return fmt.Sprintf("SP shrink size: %s is less than total capacity of all volumes in the storage pool: %s", e.RequestedSize.String(), e.MinAllowedSize.String())
}

// Below error is raised when the requested size is less than total capacity of all persistent disks (PD)
type StoragePoolShrinkExceedTotalPDCapacityError struct {
	MinAllowedSize resource.Quantity
}

func (e StoragePoolShrinkExceedTotalPDCapacityError) Error() string {
	return fmt.Sprintf("Unable to shrink the storage pool below its current minimum size (%s) due to underlying infrastructure constraints", e.MinAllowedSize.String())
}

const (
	QuarkContainerNVC               QuarkContainer = "netapp-volume-controller"
	QuarkContainerSvcMeshCtrl       QuarkContainer = "svcmesh-controller"
	QuarkContainerNSC               QuarkContainer = "ntapvol-snat-controller"
	QuarkContainerCSC               QuarkContainer = "cloud-snapshot-controller"
	QuarkContainerADC               QuarkContainer = "adc"
	QuarkContainerNHC               QuarkContainer = "netapp-ha-controller"
	QuarkContainerNM                QuarkContainer = "node-monitor"
	QuarkContainerQuarkOp           QuarkContainer = "quark-operator"
	QuarkContainerNodeVolSupport    QuarkContainer = "nodevol-supportability"
	QuarkContainerNodeVol           QuarkContainer = "nodevol-controller"
	QuarkContainerFluentBitDS       QuarkContainer = "fluent-bit-daemonset"
	QuarkContainerPromDS            QuarkContainer = "prometheus-node-exporter"
	QuarkContainerPromServer        QuarkContainer = "prometheus-server"
	QuarkContainerPromReload        QuarkContainer = "prometheus-server-configmap-reload"
	QuarkContainerPromInit          QuarkContainer = "init-chown-data"
	QuarkContainerKSM               QuarkContainer = "kube-state-metrics"
	QuarkContainerEtcd              QuarkContainer = "etcd"
	QuarkContainerEtcdCostReduction QuarkContainer = "etcdCostReduction"
	QuarkContainerEtcdOp            QuarkContainer = "etcd-operator"
	QuarkContainerNoResources       QuarkContainer = "no-resources"
	QuarkContainerBusybox           QuarkContainer = "busybox"
	QuarkContainerEtcdSpare         QuarkContainer = "etcdspare"
	QuarkContainerOtel              QuarkContainer = "otel-collector"
	QuarkK8sResMetircs              QuarkContainer = "k8s-res-metrics"
	// Quark Pod Containers
	QuarkContainerDMAP            QuarkContainer = "dmap"
	QuarkContainerPodrick         QuarkContainer = "podrick"
	QuarkContainerSECD            QuarkContainer = "secd"
	QuarkContainerLogging         QuarkContainer = "fluent-bit"
	QuarkContainerSupportability  QuarkContainer = "supportability"
	QuarkContainerCCPD            QuarkContainer = "ccpd"
	QuarkContainerIstioValidation QuarkContainer = "istio-validation"
	QuarkContainerIstioProxy      QuarkContainer = "istio-proxy"
	QuarkContainerInit            QuarkContainer = "init-container"
)

// This is the current default service mesh memory requests and half the current
// service mesh memory limits.
// NOTE: Currently these are generic in nature hence 'SvcMesh*' instead of 'ASM*'
// Assuming different cloud mesh's use the same underlying tech and act similar, we shouldnt
// need to dissern from different cloud mesh resource requests/limits
const (
	SvcMeshMemoryRequests = "128Mi"
	SvcMeshMemoryLimits   = "500Mi"
	SvcMeshCPURequests    = "100m"
	SvcMeshCPULimits      = "200m"
)

var KnownQuarkComponentImages = map[string]bool{
	"DMAP":              true,
	"Supportability":    true,
	"Secd":              true,
	"NVC":               true,
	"SvcMeshCtrl":       true,
	"Podrick":           true,
	"NSC":               true,
	"CSC":               true,
	"ADC":               true,
	"NHC":               true,
	"CCPD":              true,
	"NodevolWorker":     true,
	"NodevolController": true,
}

var AzureDiskIOPSThroughput = map[int][2]int{
	//Size 	IOPS, ThroughPut(MB/s)
	//4:     [2]int{120, 25},
	//8:     [2]int{120, 25},
	16:    [2]int{120, 25},
	32:    [2]int{120, 25},
	64:    [2]int{240, 50},
	128:   [2]int{500, 100},
	256:   [2]int{1100, 125},
	512:   [2]int{2300, 150},
	1024:  [2]int{5000, 200},
	2048:  [2]int{7500, 250},
	4096:  [2]int{7500, 250},
	8192:  [2]int{16000, 500},
	16384: [2]int{18000, 750},
	32767: [2]int{20000, 900},
}

// Helper function to make this code easier to read and maintain. It should be
// easy to see which pods are getting what resources by looking at this code.
func getQuarkInfraContainerResourcesInternal(container QuarkContainer) (cpuReq, cpuLim, memReq, memLim string) {
	resources := map[QuarkContainer][]string{
		// container                     cpuReq, cpuLim, memReq, memLim
		QuarkContainerNVC:               []string{"100m", "2", "30Mi", ""},
		QuarkContainerSvcMeshCtrl:       []string{"100m", "2", "50Mi", "200Mi"},
		QuarkContainerNSC:               []string{"50m", "200m", "15Mi", ""},
		QuarkContainerCSC:               []string{"10m", "200m", "15Mi", ""},
		QuarkContainerADC:               []string{"10m", "1", "15Mi", ""},
		QuarkContainerNHC:               []string{"10m", "200m", "15Mi", ""},
		QuarkContainerNM:                []string{"5m", "100m", "10Mi", "20Mi"},
		QuarkContainerQuarkOp:           []string{"10m", "200m", "15Mi", ""},
		QuarkContainerNodeVolSupport:    []string{"25m", "200m", "100Mi", "500Mi"},
		QuarkContainerNodeVol:           []string{"25m", "200m", "100Mi", "500Mi"},
		QuarkContainerFluentBitDS:       []string{"100m", "200m", "200Mi", "400Mi"},
		QuarkContainerPromDS:            []string{"10m", "100m", "10Mi", ""},
		QuarkContainerPromServer:        []string{"20m", "200m", "200Mi", "5000Mi"},
		QuarkContainerPromReload:        []string{"5m", "100m", "10Mi", "50Mi"},
		QuarkContainerKSM:               []string{"20m", "200m", "200Mi", ""},
		QuarkContainerEtcd:              []string{"400m", "2", "100Mi", ""},
		QuarkContainerEtcdCostReduction: []string{"100m", "2", "100Mi", ""},
		QuarkContainerEtcdOp:            []string{"10m", "100m", "10Mi", ""},
		QuarkContainerNoResources:       []string{"", "", "", ""},
		QuarkContainerEtcdSpare:         []string{"5m", "100m", "1Mi", "400Mi"},
		QuarkContainerOtel:              []string{"25m", "200m", "200Mi", "950Mi"},
		QuarkK8sResMetircs:              []string{"5m", "10m", "10Mi", "20Mi"},
	}

	if val, ok := resources[container]; ok {
		return val[0], val[1], val[2], val[3]
	} else {
		panic(fmt.Errorf("unknown container '%s'", container))
	}
}

func GetQuarkInfraContainerResources(container QuarkContainer, cloud string) corev1.ResourceRequirements {
	cpuReq, cpuLim, memReq, memLim := getQuarkInfraContainerResourcesInternal(container)

	// CTL has very limited resources so we cannot make requests
	if cloud == CloudCTL {
		cpuReq, cpuLim, memReq, memLim = getQuarkInfraContainerResourcesInternal(QuarkContainerNoResources)
	}

	requests := corev1.ResourceList{}
	limits := corev1.ResourceList{}

	if cpuReq != "" {
		requests[corev1.ResourceCPU] = resource.MustParse(cpuReq)
	}
	if cpuLim != "" {
		limits[corev1.ResourceCPU] = resource.MustParse(cpuLim)
	}
	if memReq != "" {
		requests[corev1.ResourceMemory] = resource.MustParse(memReq)
	}
	if memLim != "" {
		limits[corev1.ResourceMemory] = resource.MustParse(memLim)
	}

	return corev1.ResourceRequirements{
		Requests: requests,
		Limits:   limits,
	}
}

func GetInstanceType(quark *quarkv1alpha1.Quark, sizeType string) string {
	cloudInstanceTypes := map[string]map[string]string{
		CloudGCP: {
			"tiny":   "n1-standard-2",
			"small":  "n1-standard-4",
			"medium": "n1-standard-8",
			"large":  "n1-standard-16",
			"xlarge": "n1-standard-32",
		},
		CloudAzure: {
			"small":  "Standard_DS3_v2",
			"medium": "Standard_DS4_v2",
			"large":  "Standard_DS5_v2",
		},
	}
	if quark.Spec.EnableGcp1p {
		cloudInstanceTypes[CloudGCP]["tiny"] = "n2-standard-2"
		cloudInstanceTypes[CloudGCP]["small"] = "n2-standard-4"
		cloudInstanceTypes[CloudGCP]["medium"] = "n2-standard-8"
		cloudInstanceTypes[CloudGCP]["large"] = "n2-standard-16"
		cloudInstanceTypes[CloudGCP]["xlarge"] = "n2-standard-32"
	}
	return cloudInstanceTypes[quark.Spec.Cloud][sizeType]
}

// Helper function to create service level buckets. Container Resources are appended to the buckets in the main function.
func getServiceLevelBucketsInternal(quark *quarkv1alpha1.Quark, productionVariant bool) (buckets map[int][]string) {
	nvramSize := "8Gi"
	if quark.Spec.DisableDiskPacking == "true" {
		// If disk packing is specifically disabled, use 4Gi nvram
		nvramSize = "4Gi"
	}

	switch cloud := quark.Spec.Cloud; cloud {
	case CloudGCP:
		/* NOTE: buckets are same for production and internal configs in 1.2.0 release
		Make bucket changes here if it differs in future releases */
		buckets = map[int][]string{
			// We use bucket number in order to access previous bucket's max size during iteration
			// Bucket no.: name, nvramSize, maxVolSize, dmsCPUs, maxIOPS
			// For 3P maxIOPS is calculated Ti * IOPSPerTiB.  For 1P is set here
			1: []string{"small", nvramSize, "1Ti", "4", "0"},
			2: []string{"medium", nvramSize, "3Ti", "8", "0"},
			3: []string{"large", nvramSize, "5Ti", "14", "0"},
			4: []string{"xlarge", nvramSize, "200Ti", "20", "0"},
		}
		if quark.Spec.EnableGcp1p {
			buckets = map[int][]string{
				1: []string{"small", nvramSize, "3Ti", "4", "3072"},
				2: []string{"medium", nvramSize, "8Ti", "8", "8192"},
				3: []string{"large", nvramSize, "14Ti", "14", "14336"},
				4: []string{"xlarge", nvramSize, "200Ti", "28", "61440"},
			}
		}
	case CloudAzure:
		// same for both production and internal, change here if configs are different
		buckets = map[int][]string{
			1: []string{"small", nvramSize, "2Ti", "4", "0"},
			2: []string{"medium", nvramSize, "4Ti", "8", "0"},
			3: []string{"large", nvramSize, "128Ti", "16", "0"},
		}
	case CloudCTL:
		// will be used for CTL
		buckets = map[int][]string{
			// Bucket    name, nvramSize, maxVolSize
			1: []string{"small", nvramSize, "1Ti", "4", "0"},
			2: []string{"medium", nvramSize, "3Ti", "8", "0"},
			3: []string{"large", nvramSize, "5Ti", "14", "0"},
			4: []string{"xlarge", nvramSize, "200Ti", "20", "0"},
		}
		if quark.Spec.EnableGcp1p {
			buckets = map[int][]string{
				1: []string{"small", nvramSize, "3Ti", "4", "3072"},
				2: []string{"medium", nvramSize, "8Ti", "8", "8192"},
				3: []string{"large", nvramSize, "14Ti", "14", "14336"},
				4: []string{"xlarge", nvramSize, "200Ti", "28", "61440"},
			}
		}
	}
	return buckets
}

// returns node selector for determining the node pool when VMRS is enabled
func MakeNodeSelector(rightSize bool, vmSize string) string {
	if rightSize {
		return InstanceTypeKey + "=" + vmSize
	}
	return ""
}

// Builds resource requirements for each quark container, to be added to quark pod buckets
func getContainerResources(bucketSize string, productionVariant bool, quark *quarkv1alpha1.Quark) []quarkv1alpha1.ContainerResource {
	resources := map[QuarkContainer][]string{
		// container                           cpuReq, cpuLim, memReq, memLim
		QuarkContainerPodrick:        []string{"25m", "200m", "250Mi", "250Mi"},
		QuarkContainerSECD:           []string{"25m", "200m", "250Mi", "250Mi"},
		QuarkContainerLogging:        []string{"25m", "200m", "50Mi", "200Mi"},
		QuarkContainerSupportability: []string{"25m", "200m", "100Mi", "500Mi"},
	}
	if quark.Spec.Nodevol == "true" {
		resources[QuarkContainerSupportability] = []string{"10m", "40m", "30Mi", "60Mi"}
	}
	resources[QuarkContainerPodrick] = []string{"25m", "200m", "500Mi", "500Mi"}
	resources[QuarkContainerSupportability] = []string{"10m", "40m", "200Mi", "200Mi"}

	if quark.Spec.EnableCcpd {
		resources[QuarkContainerCCPD] = []string{"25m", "200m", "500Mi", "500Mi"}
	}
	if quark.Spec.EnableServiceMesh {
		resources[QuarkContainerIstioValidation] = []string{
			SvcMeshCPURequests,
			SvcMeshCPULimits,
			SvcMeshMemoryRequests,
			SvcMeshMemoryLimits,
		}
		resources[QuarkContainerIstioProxy] = []string{
			SvcMeshCPURequests,
			SvcMeshCPULimits,
			SvcMeshMemoryRequests,
			SvcMeshMemoryLimits,
		}
	}

	dmapMediumMemReq := "23.6Gi"
	dmapSmallMemReq := "9.3Gi"
	if quark.Spec.Cloud == CloudAzure {
		dmapMediumMemReq = "21.6Gi"
		dmapSmallMemReq = "8.3Gi"
	}

	/* NOTE: resource allocation is same for production and internal configs in 1.2.0 release
	Make resource allocation changes here if it differs in future releases */
	switch bucketSize {
	case "small":
		resources[QuarkContainerDMAP] = []string{"2", "4", dmapSmallMemReq, "9.3Gi"}
	case "medium":
		resources[QuarkContainerDMAP] = []string{"6", "8", dmapMediumMemReq, "23.6Gi"}
	case "large":
		resources[QuarkContainerDMAP] = []string{"14", "15", "28Gi", "28Gi"}
	case "xlarge":
		resources[QuarkContainerDMAP] = []string{"22", "22", "28Gi", "28Gi"}
		if quark.Spec.EnableGcp1p {
			resources[QuarkContainerDMAP] = []string{"30", "30", "28Gi", "28Gi"}
		}
	}
	return parseContainerResources(resources)
}

// helper function that parses string values of resource requirements into type pertaining to ContainerResource struct
func parseContainerResources(resources map[QuarkContainer][]string) (containerResources []quarkv1alpha1.ContainerResource) {
	for k, v := range resources {
		cr := quarkv1alpha1.ContainerResource{
			ContainerName: k,
			CPUsRequest:   resource.MustParse(v[0]),
			CPUsLimit:     resource.MustParse(v[1]),
			MemRequest:    resource.MustParse(v[2]),
			MemLimit:      resource.MustParse(v[3]),
		}
		containerResources = append(containerResources, cr)
	}
	return
}

func GetQOSlimits(quark *quarkv1alpha1.Quark, fieldName string) string {
	QOSValues := map[string]map[string]string{
		CloudGCP: {
			"DiskIOPSPerGiB": "30",
			"TpAmpFactor":    "2.5",
			"IOPSPerTiB":     "4096",
			"MiBsPerTiB":     "128",
			// TODO: deprecate after storage pool is enabled by default
			"CapacityLimit": "14Ti",
			"IOPSLimit":     "60000",
			"TputLimit":     "1152",
		},
		CloudAzure: {
			"DiskIOPSPerGiB": "30",
			"TpAmpFactor":    "1.9",
			"IOPSPerTiB":     "1024",
			"MiBsPerTiB":     "64",
			// TODO: deprecate after storage pool is enabled by default
			"CapacityLimit": "14Ti",
			// TODO: tentative, change these based on Perf
			"IOPSLimit": "14000",
			"TputLimit": "640",
		},
		CloudCTL: {
			"DiskIOPSPerGiB": "30",
			"TpAmpFactor":    "2.5",
			"IOPSPerTiB":     "4096",
			"MiBsPerTiB":     "128",
			// TODO: deprecate after storage pool is enabled by default
			"CapacityLimit": "14Ti",
			"IOPSLimit":     "60000",
			"TputLimit":     "1152",
		},
	}
	if quark.Spec.EnableGcp1p {
		// For 1P, it's 16MiBps/TiB
		QOSValues[CloudGCP]["MiBsPerTiB"] = "16"
		QOSValues[CloudCTL]["MiBsPerTiB"] = "16"
		QOSValues[CloudCTL]["IOPSLimit"] = "61440"
	}
	return QOSValues[quark.Spec.Cloud][fieldName]
}

// GetQuarkPodBuckets This function parses string values of service level bucket fields into type pertaining to
// QuarkPodBucket struct
func GetQuarkPodBuckets(quark *quarkv1alpha1.Quark, productionVariant, rightSize bool) (qpb []quarkv1alpha1.QuarkPodBucket) {
	var (
		minVolSize string
	)
	buckets := getServiceLevelBucketsInternal(quark, productionVariant)
	for k, v := range buckets {
		if k == 1 {
			// first bucket
			minVolSize = "0Gi"
		} else {
			// this field value will be the value of MaxVolumeSize of previous bucket
			minVolSize = buckets[k-1][2]
		}
		bucket := quarkv1alpha1.QuarkPodBucket{
			Name:               v[0],
			NvramSize:          resource.MustParse(v[1]),
			NodeSelector:       MakeNodeSelector(rightSize, GetInstanceType(quark, v[0])),
			MinVolumeSize:      resource.MustParse(minVolSize),
			MaxVolumeSize:      resource.MustParse(v[2]),
			DmsCPUs:            resource.MustParse(v[3]),
			MaxIOPS:            resource.MustParse(v[4]),
			ContainerResources: getContainerResources(v[0], productionVariant, quark),
		}
		qpb = append(qpb, bucket)
	}
	return
}

// StoragePoolEncryptionState type is used to set StoragePoolEncryptionLabel with respective state values
type StoragePoolEncryptionState string

const (
	StoragePoolEncryptionPause  StoragePoolEncryptionState = "pause"
	StoragePoolEncryptionResume StoragePoolEncryptionState = "resume"
	StoragePoolEncryptionStart  StoragePoolEncryptionState = "start"
)

func (s StoragePoolEncryptionState) String() string {
	return string(s)
}

var CMEK_KEYS = []string{"CMEK_KEY_RING",
	"CMEK_KEY_NAME",
	"CMEK_TIMEOUT",
	"CMEK_PROTOCOL",
	"CMEK_PROJECT_ID",
	"CMEK_KEY_RING_LOCATION",
	"CMEK_TRUSTED_FILE_PATH"}
