package helm

import (
	"bytes"
	"encoding/base64"
	"fmt"
	"strings"
	"time"

	log "github.com/sirupsen/logrus"

	"github.com/davecgh/go-spew/spew"
	"helm.sh/helm/v3/pkg/action"
	"helm.sh/helm/v3/pkg/chart"
	"helm.sh/helm/v3/pkg/chart/loader"
	"helm.sh/helm/v3/pkg/release"
	"k8s.io/client-go/rest"
)

// Retries for conflict errors (i.e the object has been modified; please apply your changes to the latest version and try again)
const maxRetries = 5
const retryInterval = 1

type HelmReconciler struct {
	Kubeconfig         *rest.Config
	Base64EncodedChart string
	Values             map[string]interface{}
	ReleaseName        QuarkHelmChart
	Namespace          string
	actionConfig       *action.Configuration
	helmRestConfig     *CustomRESTClientGetter
	chartArchive       []byte
	chart              *chart.Chart
	initalized         bool
}

func NewHelmReconciler(kubeconfig *rest.Config, base64EncodedChart string, values map[string]interface{}, releaseName QuarkHelmChart, namespace string) *HelmReconciler {
	return &HelmReconciler{
		Kubeconfig:         kubeconfig,
		Base64EncodedChart: base64EncodedChart,
		Values:             values,
		ReleaseName:        releaseName,
		Namespace:          namespace,
	}
}

func (h *HelmReconciler) Reconcile() (err error) {
	if err := h.init(); err != nil {
		return err
	}

	installedRelease, err := h.GetInstalledRelease()
	if err != nil {
		return err
	}
	if installedRelease == nil {
		err = h.migrateAndInstall()
	} else {
		err = h.upgrade()
	}
	return err
}

// GetQuarkCRName returns the resulting name of a quark cr from a collection of helm values
func GetQuarkCRName(values map[string]interface{}, namespace string) string {
	if name, ok := values["overwriteQuarkCRName"].(string); ok && name != "" {
		return name
	}
	return namespace
}

func (h *HelmReconciler) GetInstalledRelease() (foundRelease *release.Release, err error) {
	if err = h.init(); err != nil {
		return nil, err
	}
	get := action.NewGet(h.actionConfig)
	foundRelease, err = get.Run(string(h.ReleaseName))
	if err != nil {
		if strings.Contains(err.Error(), "not found") {
			log.WithField("ReleaseName", h.ReleaseName).Debug("Release not installed")
			return nil, nil
		}
		return nil, err
	}
	log.WithField("ReleaseName", h.ReleaseName).Debug("Release already installed")
	return foundRelease, nil
}

func (h *HelmReconciler) migrateAndInstall() (err error) {
	log.Debugln("Chart not installed, starting install workflow")
	// Quark cluster is not running on helm, migrate + install + upgrade to reconcile
	err = h.migrateQuarkResourcesToHelm()
	if err != nil {
		return fmt.Errorf("error migrating release: %v", err)
	}
	err = h.install()
	if err != nil {
		return fmt.Errorf("error installing release: %v", err)
	}
	// Release must be upgraded after install to reconcile spec changes
	err = h.upgrade()
	return err
}

func (h *HelmReconciler) install() error {
	log.Debugln("Starting helm chart install")
	install := action.NewInstall(h.actionConfig)
	install.ReleaseName = string(h.ReleaseName)
	install.Namespace = h.Namespace
	install.CreateNamespace = true
	install.Replace = true
	return h.retryOnConflict(func() (*release.Release, error) {
		return install.Run(h.chart, h.Values)
	})
}

func (h *HelmReconciler) upgrade() error {
	log.Debugln("Starting helm chart upgrade")
	upgrade := action.NewUpgrade(h.actionConfig)
	upgrade.Force = true
	upgrade.Namespace = h.Namespace
	// https://shipmight.com/blog/understanding-helm-upgrade-reset-reuse-values
	upgrade.ResetValues = true
	return h.retryOnConflict(func() (*release.Release, error) {
		return upgrade.Run(string(h.ReleaseName), h.chart, h.Values)
	})
}

func (h *HelmReconciler) retryOnConflict(operation func() (*release.Release, error)) error {
	var lastError error
	for i := 0; i < maxRetries; i++ {
		_, err := operation()
		if err != nil {
			if strings.Contains(err.Error(), "the object has been modified") {
				log.Debugln("Operation resulted in conflict, retrying...")
				time.Sleep(retryInterval)
				lastError = err
				continue
			} else {
				return err
			}
		} else {
			return nil
		}
	}
	return lastError
}

func loadChart(chartArchive []byte) (*chart.Chart, error) {
	chartReader := bytes.NewReader(chartArchive)
	chart, err := loader.LoadArchive(chartReader)
	if err != nil {
		return nil, fmt.Errorf("error loading chart from archive: %v", err)
	}
	return chart, nil
}

func (h *HelmReconciler) dumpDebugValues() {
	log.Debug("Chart values")
	log.Debugf("Override values %s", spew.Sdump(h.Values))
}

func (h *HelmReconciler) init() (err error) {
	if h.initalized {
		return nil
	}
	if err := h.initNetwork(); err != nil {
		return err
	}
	if h.Base64EncodedChart != "" {
		if err := h.initChart(); err != nil {
			return err
		}
	}
	if h.Namespace == "" {
		return fmt.Errorf("no namespace set for HelmReconciler")
	}
	h.initalized = true
	return nil
}

func (h *HelmReconciler) initNetwork() (err error) {
	h.actionConfig = new(action.Configuration)
	h.helmRestConfig = NewCustomRESTClientGetter(h.Kubeconfig, h.Namespace)
	err = h.actionConfig.Init(h.helmRestConfig, h.Namespace, "", log.Debugf)
	if err != nil {
		return fmt.Errorf("error initializing action configuration: %v", err)
	}
	return nil
}

func (h *HelmReconciler) initChart() (err error) {
	h.chartArchive, err = base64.StdEncoding.DecodeString(h.Base64EncodedChart)
	if err != nil {
		return fmt.Errorf("error decoding base64-encoded chart archive: %v", err)
	}
	h.chart, err = loadChart(h.chartArchive)
	if err != nil {
		return fmt.Errorf("error loading chart: %v", err)
	}
	// Automatically apply the correct values.yaml files depending on env, cloud, etc
	h.Values, err = applyDefaultEnvironmentBasedValues(h.Base64EncodedChart, h.Values)
	if err != nil {
		return err
	}
	h.dumpDebugValues()
	return nil
}
